/* appicon.c- icon for applications (not mini-window)
 *
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "window.h"
#include "pixmap.h"
#include "icon.h"
#include "appicon.h"
#include "actions.h"
#include "resource.h"
#include "stacking.h"
#include "dock.h"
#include "funcs.h"

/*
 * TODO:
 *  how to find out which running instance was launched from where
 * (command line, dock etc.)  _GNUSTEP_WM_PID ??
 * 
 * icon_file for the dock is got from the preferences file by
 * using the classname/instancename
 */

/**** Global variables ****/
extern Cursor wCursor[WCUR_LAST];
extern WPreferences wPreferences;

static void iconMouseDown(WObjDescriptor *desc, XEvent *event);
static void iconDblClick(WObjDescriptor *desc, XEvent *event);

WMethodTable wAppIconMethods = {
    (WEventHandler)iconMouseDown,
      NULL,
      (WEventHandler)iconDblClick,
      NULL
};

#if 0
/*
 *---------------------------------------------------------------------- 
 * wAppIconCreateFromIcon--
 * 	Create an icon for use in the Dock, by "converting" a normal
 * icon to a dock icon. Tries to guess how the app is launched and
 * if it can't, fail.
 * 
 * Returns:
 * 	A WAppIcon structure or NULL, if the icon can't be docked.
 * 
 * Side effects:
 *	The dock_icon flag in the icon is set.
 * 
 * Notes:
 * 	WIcons with the dock_icon flag set should not be destroyed
 * by other routines.
 * 	Only GNUstep apps can be hidden into the Dock icon, since
 * we can have multiple instances of X apps running.
 *---------------------------------------------------------------------- 
 */
WAppIcon*
wAppIconCreateFromIcon(WIcon *icon)
{
    WAppIcon *dicon;
    Window main_win;
    char **argv;
    int argc;

    dicon = wmalloc(sizeof(WAppIcon));
    memset(dicon, 0, sizeof(WAppIcon));
    dicon->icon = icon;
    /* we've got to find out how the app is invoked:
     * by getting the path for itself that's passed
     * by the app, using the icon in the .App wrapper
     * or
     * with WM_COMMAND, using icon_file as it's icon
     * or
     * if none of the above is possible, return NULL */
    
    
    /* check WM_COMMAND */
    if (!wApplicationOf(icon->owner->main_window)) {
	main_win = icon->owner->client_win;
    }  else {
	main_win = icon->owner->main_window;
    }
    if (!XGetCommand(dpy, main_win, &argv, &argc)) {
	wfree(dicon);
	return NULL;
    }
    dicon->icon_file = icon->owner->icon_file;
    /* TODO: finish argv -> command conversion */
    dicon->command = strdup(argv[0]);

    XFreeStringList(argv);

    icon->dock_icon = True;
    return dicon;
}

#endif


WAppIcon*
wAppIconCreateForCommand(WScreen *scr, char **argv, int argc, char *wm_class,
			 char *wm_instance)
{
    WAppIcon *dicon;
    WResource *res;
    char *image, *mask;
    
    dicon = wmalloc(sizeof(WAppIcon));
    memset(dicon, 0, sizeof(WAppIcon));

    if (argv) {
	dicon->argv = argv;
	dicon->argc = argc;
    }
    res = wResourceGet(NULL, wm_class, wm_instance);
    if (!wm_class && !wm_instance) {
	image = wPreferences.logo_file;
	mask = NULL;
    } else if (!res) {
	image = NULL;
	mask = NULL;
    } else {
	image = res->icon_file;
	mask = res->icon_mask_file;	
    }
    dicon->wm_class = wm_class;
    dicon->wm_instance = wm_instance;

    dicon->icon = wIconCreateWithImage(scr, image, mask);
    /* will be overriden by dock */
    dicon->icon->core->descriptor.methods = &wAppIconMethods;
    dicon->icon->core->descriptor.parent_type = WCLASS_ICON;
    dicon->icon->core->descriptor.parent = dicon;
    AddToStackList(dicon->icon->core);
    
    return dicon;
}



WAppIcon*
wAppIconCreate(WWindow *leader_win)
{
    WAppIcon *aicon;

    aicon = wmalloc(sizeof(WAppIcon));
    memset(aicon, 0, sizeof(WAppIcon));

    aicon->icon = wIconCreate(leader_win);
    /* will be overriden if docked */
    aicon->icon->core->descriptor.methods = &wAppIconMethods;
    aicon->icon->core->descriptor.parent_type = WCLASS_ICON;
    aicon->icon->core->descriptor.parent = aicon;
    AddToStackList(aicon->icon->core);
    aicon->icon->show_title = 0;
    wIconUpdate(aicon->icon);

    return aicon;
}


void
wAppIconDestroy(WAppIcon *aicon)
{
    RemoveFromStackList(aicon->icon->core);
    wIconDestroy(aicon->icon);
    if (aicon->argv)
      XFreeStringList(aicon->argv);

    wfree(aicon);
}



void
wAppIconPaint(WAppIcon *aicon)
{
    WScreen *scr = aicon->icon->core->screen_ptr;

    wIconPaint(aicon->icon);

    if (scr->dock_dots && aicon->docked && !aicon->running) {
	XSetClipMask(dpy, scr->copy_gc, scr->dock_dots->mask);
	XSetClipOrigin(dpy, scr->copy_gc, 0, 0);
	XCopyArea(dpy, scr->dock_dots->image, aicon->icon->core->window, 
		  scr->copy_gc, 0, 0, scr->dock_dots->width,
		  scr->dock_dots->height, 0, 0);
	XSetClipMask(dpy, scr->copy_gc, None);
    }
    if (aicon->launching) {
	XFillRectangle(dpy, aicon->icon->core->window, scr->stipple_gc, 
		       0, 0, ICON_WIDTH, ICON_HEIGHT);
    }

}


static int
canBeDocked(WWindow *wwin)
{
    if (wwin->wm_class || wwin->wm_instance) {
	char **argv;
	int argc;
	  
	if (!XGetCommand(dpy, wwin->client_win, &argv, &argc)) {
	    return False;
	}
	XFreeStringList(argv);
	return True;
    }
    return False;
}


/******************************************************************/
static void 
iconDblClick(WObjDescriptor *desc, XEvent *event)
{
    WAppIcon *aicon = desc->parent;
    WApplication *wapp;

    if (event->xbutton.button == Button3) {
	WWindow *wwin = aicon->icon->owner;
	if (wwin->wm_instance && wwin->wm_class) {
	    printf("%s.%s\n", wwin->wm_instance, wwin->wm_class);
	} else if (wwin->wm_class) {
	    printf("%s\n", wwin->wm_class);
	} else if (wwin->wm_instance) {
	    printf("%s\n", wwin->wm_instance);
	} else {
	    printf("instance/class name not available for window %x\n",
		   (unsigned)wwin->client_win);
	}
    } else {
	wapp = wApplicationOf(aicon->icon->owner->main_window);
#ifdef DEBUG0
	if (!wapp) {
	    wWarning("could not find application descriptor for app icon!!");
	    return;
	}
#endif
	wUnhideApplication(wapp);
    }
}


static void 
iconMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WAppIcon *aicon = desc->parent;
    WIcon *icon = aicon->icon;
    WScreen *scr = icon->core->screen_ptr;
    XEvent ev;
    int x=aicon->x_pos, y=aicon->y_pos;
    int dx=event->xbutton.x, dy=event->xbutton.y;
    int grabbed=0;
    int shad_x, shad_y, docking=0, dockable;

#ifdef DEBUG
    puts("Moving icon");
#endif
    if (event->xbutton.state & Mod1Mask)
      wLowerFrame(icon->core);
    else
      wRaiseFrame(icon->core);

    if (XGrabPointer(dpy, icon->core->window, False, ButtonMotionMask
		     |ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, 
		     None, CurrentTime) !=GrabSuccess) {
	wWarning("pointer grab failed for icon move");
    }
    dockable = canBeDocked(icon->owner);
    while(1) {
	XMaskEvent(dpy, PointerMotionMask|ButtonReleaseMask|ButtonMotionMask
		   |ExposureMask, &ev);
	switch (ev.type) {
	 case Expose:
	    DispatchEvent(&ev);
	    break;

	 case MotionNotify:
	    if (!grabbed) {
		if (abs(dx-ev.xmotion.x)>=MOVE_THRESHOLD
		    || abs(dy-ev.xmotion.y)>=MOVE_THRESHOLD) {
		    XChangeActivePointerGrab(dpy, ButtonMotionMask
					     |ButtonReleaseMask, 
					     wCursor[WCUR_MOVE], CurrentTime);
		    grabbed=1;
		} else {
		    break;
		}
	    }
	    x = ev.xmotion.x_root - dx;
	    y = ev.xmotion.y_root - dy;
	    XMoveWindow(dpy, icon->core->window, x, y);
	    if (dockable && x > scr->scr_width - ICON_WIDTH*5/2) {
		if (wDockSnapIcon(scr->dock, aicon, x, y, &shad_x, &shad_y)) {
		    XMoveWindow(dpy, scr->dock_shadow, shad_x, shad_y);
		    if (!docking) {
			Window wins[2];

			wins[0] = icon->core->window;
			wins[1] = scr->dock_shadow;
			XRestackWindows(dpy, wins, 2);
			XMapWindow(dpy, scr->dock_shadow);
		    }
		    docking = 1;
		}
	    } else if (docking) {
		XUnmapWindow(dpy, scr->dock_shadow);
		docking = 0;
	    }
	    break;

	 case ButtonRelease:
	    if (docking) {
		/* icon was docked */
		SlideWindow(icon->core->window, x, y, shad_x, shad_y,
			    DOCK_SPRING_STEPS, DOCK_SPRING_DELAY);
		XUnmapWindow(dpy, scr->dock_shadow);
		aicon->x_pos = shad_x;
		aicon->y_pos = shad_y;
		
		wDockAttachIcon(scr->dock, aicon);
	    } else {
		XMoveWindow(dpy, icon->core->window, x, y);
		aicon->x_pos = x;
		aicon->y_pos = y;
	    }
#ifdef DEBUG
	    puts("End icon move");
#endif
	    XUngrabPointer(dpy, CurrentTime);
	    return;
	}
    }
}
    
