/* appmenu.c- application defined menu
 *
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "button.h"
#include "menu.h"
#include "actions.h"
#include "funcs.h"



/******** Global Variables **********/
extern Atom _XA_WINDOWMAKER_MENU;
extern Atom _XA_WINDOWMAKER_WM_PROTOCOLS;




enum {
    WMCreateMenu = 1,
      WMInsertItem = 2,
      WMSetSubmenu = 3,
};



static void 
notifyClient(WMenu *menu, WMenuEntry *entry)
{
    
}



static WMenu*
findMenu(LinkedList *list, int code)
{
    WMenu *menu;
    
    while (list) {
	menu = list->head;
	if (menu->code==code)
	  return menu;
	list = list->tail;
    }
    return NULL;
}

static WMenu*
parseMenuCommand(WScreen *scr, Window window, LinkedList **menu_list, 
		 char *command)
{
    WMenu *menu, *main_menu;
    char *rest;
    char title[MAX_MENU_TEXT_LENGTH+32];
    int cmd, menu_code;
    
    rest = strchr(command, ' ');
    if (!rest) {
#ifdef DEBUG0
	wWarning("appmenu:invalid menu command: %s", command);
#endif
	return NULL;
    }
    *rest = 0;
    if (sscanf(command, "%i", &cmd)!=1 && (cmd < 1 || cmd > 3)) {
#ifdef DEBUG0
	wWarning("appmenu:bad menu command code: %s", command);
#endif
	return NULL;
    }
    main_menu = NULL;
    switch (cmd) {
     case WMCreateMenu: {
	 if (sscanf(rest, "%i %s", &menu_code, title)!=2) {
#ifdef DEBUG0
	     wWarning("appmenu:bad CreateMenu parameter format: %s", rest);
#endif
	     return NULL;
	 }
	 menu = wMenuCreate(scr, title, False);
	 if (!menu) {
	     wWarning("appmenu:could not create menu");
	     return NULL;
	 }
	 menu->code = menu_code;
	 menu->flags.app_menu = 1;
	 *menu_list = list_cons(menu, *menu_list);
     }
	break;

     case WMInsertItem: {
	 int tag, idx, enabled;
	 KeySym ksym;
	 WMenuEntry *entry;

	 if (sscanf(rest, "%i %i %i %i %ui %s", &menu_code, &tag, &idx, 
		    &enabled, (unsigned int*)&ksym, title)!=6) {
#ifdef DEBUG0
	     wWarning("appmenu:bad InsertItem parameter format: %s", rest);
#endif
	     return NULL;
	 }
	 menu = findMenu(*menu_list, menu_code);
	 if (!menu) {
#ifdef DEBUG0
	     wWarning("appmenu:bad menu code in InsertItem: %s", rest);
#endif
	     return NULL;
	 }
	 entry = wMenuInsertCallback(menu, idx, title, notifyClient, 
			(void*)window);
	 if (!entry) {
	     wWarning("appmenu:could not create menu entry");
	     return NULL;
	 }
	 entry->code = tag;
	 entry->rtext = strdup(XKeysymToString(ksym));
     }
	break;

     case WMSetSubmenu: {
	 int tag, submenu_code;
	 WMenu *submenu;
	 WMenuEntry *entry;

	 if (sscanf(rest, "%i %i %i", &menu_code, &tag, &submenu_code)!=3) {
#ifdef DEBUG0
	     wWarning("appmenu:bad SetSubmenu parameter format: %s", rest);
#endif
	     return NULL;
	 }
	 submenu = findMenu(*menu_list, submenu_code);
	 if (!submenu) {
#ifdef DEBUG0
	     wWarning("appmenu:bad submenu code in SetSubmenu: %s", rest);
#endif
	     return NULL;
	 }
	 if (menu_code==0) {
	     main_menu = submenu;
	     submenu->parent = NULL;
	 } else {
	     int i;
	     menu = findMenu(*menu_list, menu_code);
	     if (!menu) {
#ifdef DEBUG0
		 wWarning("appmenu:bad menu code in SetSubmenu: %s", rest);
#endif
		 return NULL;
	     }
	     entry = NULL;
	     for (i=0; i < menu->entry_no; i++) {
		 if (menu->entries[i]->code==tag) {
		     entry = menu->entries[i];
		 }
	     }
	     if (!entry) {
#ifdef DEBUG0
		 wWarning("appmenu:bad item tag in SetSubmenu: %s", rest);
#endif
		 return NULL;
	     }
	     entry->cascade = submenu;
	     submenu->parent = menu;
	 }
     }
	break;
    }
    return main_menu;
}


WMenu*
wAppMenuGet(WScreen *scr, Window window, LinkedList **menu_list)
{
    XTextProperty prop;
    WMenu *main_menu, *new_main_menu;
    char **list;
    int count, i;


    if (!XGetTextProperty(dpy, window, &prop, _XA_WINDOWMAKER_MENU)) {
	return NULL;
    }
    if (!XTextPropertyToStringList(&prop, &list, &count)) {
	return NULL;
    }
    new_main_menu = NULL;
    main_menu = NULL;
    for (i = 0; i < count; i++) {
	new_main_menu = parseMenuCommand(scr, window, menu_list, list[i]);
	if (new_main_menu)
	  main_menu = new_main_menu;
    }
    XFreeStringList(list);
    XDeleteProperty(dpy, window, _XA_WINDOWMAKER_MENU);

    return main_menu;
}

void
wAppMenuDestroy(WApplication *wapp) 
{
    LinkedList *list;
    WMenu *menu;
    
    list = wapp->menu_list;
    while (list) {
	menu = list->head;
	wMenuDestroy(menu, False);
	list = list->tail;
    }
    wapp->menu = NULL;
}
