/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "window.h"
#include "properties.h"
#include "actions.h"
#include "icon.h"
#include "client.h"
#include "funcs.h"
#include "stacking.h"

/****** Global Variables ******/

/* contexts */
extern XContext wWinContext;

extern Atom _XA_WM_STATE;
extern Atom _XA_WM_PROTOCOLS;
extern Atom _XA_WM_COLORMAP_WINDOWS;


/*
 *--------------------------------------------------------------------
 * wClientRestore--
 * 	Reparent the window back to the root window.
 * 
 *--------------------------------------------------------------------
 */
void
wClientRestore(WWindow *wwin)
{
    XUnmapWindow(dpy, wwin->client_win);
    XSetWindowBorderWidth(dpy, wwin->client_win, wwin->old_border_width);    
    XReparentWindow(dpy, wwin->client_win, wwin->screen_ptr->root_win,
		    wwin->frame_x/*-FRAME_BORDER_WIDTH+wwin->old_border_width*/, 
		    wwin->frame_y/*-FRAME_BORDER_WIDTH+wwin->old_border_width*/);
    if (wwin->flags.mapped)
      XMapWindow(dpy, wwin->client_win);
    
    /* don't let the window get iconified after restart */
    if (wwin->flags.shaded)
      wClientSetState(wwin, NormalState, None);
}


/*
 *---------------------------------------------------------------------- 
 * wClientSetState--
 * 	Set the state of the client window to one of the window
 * states defined in ICCCM (Iconic, Withdrawn, Normal)
 * 
 * Side effects:
 * 	The WM_STATE property of the window is updated as well as the
 * WWindow.state variable.
 *---------------------------------------------------------------------- 
 */
void
wClientSetState(WWindow *wwin, int state, Window icon_win)
{
    unsigned long data[2];
    
    wwin->state = state;

    data[0] = (unsigned long) state;
    data[1] = (unsigned long) icon_win;

    XChangeProperty(dpy, wwin->client_win, _XA_WM_STATE, _XA_WM_STATE, 32, 
		    PropModeReplace, (unsigned char *) data, 2);
}


void
wClientGetGravityOffsets(WWindow *wwin, int *ofs_x, int *ofs_y)
{
    switch (wwin->normal_hints->win_gravity) {	
     case ForgetGravity:
     case CenterGravity:
     case StaticGravity:
	*ofs_x = 0;
	*ofs_y = 0;
	break;
     case NorthWestGravity:
	*ofs_x = -1;
	*ofs_y = -1;
	break;
     case NorthGravity:
	*ofs_x = 0;
	*ofs_y = -1;
	break;
     case NorthEastGravity:
	*ofs_x = 1;
	*ofs_y = -1;
	break;
     case WestGravity:
	*ofs_x = -1;
	*ofs_y = 0;
	break;
     case EastGravity:
	*ofs_x = 1;
	*ofs_y = 0;
	break;
     case SouthWestGravity:
	*ofs_x = -1;
	*ofs_y = 1;
	break;
     case SouthGravity:
	*ofs_x = 0;
	*ofs_y = 1;
	break;
     case SouthEastGravity:
	*ofs_x = 1;
	*ofs_y = 1;
	break;
    }
}



void
wClientConfigure(WWindow *wwin, XConfigureRequestEvent *xcre)
{
    XWindowChanges xwc;
    int nx, ny, nwidth, nheight;
    int ofs_x, ofs_y;

    if (wwin==NULL) {
	/* 
	 * configure a window that was not mapped by us 
	 */
	xwc.x = xcre->x;
	xwc.y = xcre->y;
	xwc.width = xcre->width;
	xwc.height = xcre->height;
	xwc.border_width = xcre->border_width;
	xwc.stack_mode = xcre->detail;
	xwc.sibling = xcre->above;	
	XConfigureWindow(dpy, xcre->window, xcre->value_mask, &xwc);
	return;
    }
#ifdef SHAPE
    {
	int junk;
	unsigned int ujunk;
	int b_shaped;
	
	XShapeSelectInput(dpy, wwin->client_win, ShapeNotifyMask);
	XShapeQueryExtents(dpy, wwin->client_win, &b_shaped, &junk, &junk, 
			   &ujunk, &ujunk, &junk, &junk, &junk, &ujunk, 
			   &ujunk);
	wwin->flags.shaped = b_shaped;
    }
#endif
    if (xcre->value_mask & CWStackMode) {
	WObjDescriptor *desc;
	WWindow *sibling;

	if ((xcre->value_mask & CWSibling) &&
	    (XFindContext(dpy, xcre->above, wWinContext, 
			  (XPointer *)&desc) == XCSUCCESS)
	    && (desc->parent_type==WCLASS_WINDOW)) {
	    sibling=desc->parent;
	    xwc.sibling = sibling->frame->window;
	} else {
	    xwc.sibling = xcre->above;
	}
	xwc.stack_mode = xcre->detail;
	XConfigureWindow(dpy, wwin->frame->window,
			 xcre->value_mask & (CWSibling | CWStackMode), &xwc);
	/* fix stacking order */
	RemakeStackList(wwin->screen_ptr);
    }

    wClientGetGravityOffsets(wwin, &ofs_x, &ofs_y);
    
    if (xcre->value_mask & CWBorderWidth) {
	wwin->old_border_width = xcre->border_width;
    }

    
    if (xcre->value_mask & CWX)
      nx = xcre->x - FRAME_BORDER_WIDTH;
    else
      nx = wwin->frame_x;
    
    if (xcre->value_mask & CWY)
      ny = xcre->y - FRAME_BORDER_WIDTH - ((ofs_y < 0) ? 0 : wwin->top_width);
    else
      ny = wwin->frame_y;

    if (xcre->value_mask & CWWidth)
      nwidth = xcre->width;
    else
      nwidth = wwin->frame->width;
    
    if (xcre->value_mask & CWHeight) 
      nheight = xcre->height + wwin->top_width + wwin->bottom_width;
    else
      nheight = wwin->frame->height;

    wWindowConfigure(wwin, nx, ny, nwidth, nheight);
}


void
wClientSendProtocol(WWindow *wwin, Atom protocol, Time time)
{
    XEvent event;

    event.xclient.type = ClientMessage;
    event.xclient.message_type = _XA_WM_PROTOCOLS;
    event.xclient.format = 32;
    event.xclient.display = dpy;
    event.xclient.window = wwin->client_win;
    event.xclient.data.l[0] = protocol;
    event.xclient.data.l[1] = time;
    event.xclient.data.l[2] = 0;
    event.xclient.data.l[3] = 0;
    XSendEvent(dpy, wwin->client_win, False, NoEventMask, &event);
    XFlush(dpy);
}

void
wClientSendProtocolMessage(WWindow *wwin, Atom protocol, Time time, 
			   int data2, int data3)
{
    XEvent event;

    event.xclient.type = ClientMessage;
    event.xclient.message_type = _XA_WM_PROTOCOLS;
    event.xclient.format = 32;
    event.xclient.display = dpy;
    event.xclient.window = wwin->client_win;
    event.xclient.data.l[0] = protocol;
    event.xclient.data.l[1] = time;
    event.xclient.data.l[2] = data2;
    event.xclient.data.l[3] = data3;
    XSendEvent(dpy, wwin->client_win, False, NoEventMask, &event);
    XFlush(dpy);
}


void
wClientKill(WWindow *wwin)
{
    XKillClient(dpy, wwin->client_win);
    XFlush(dpy);
}

/*
 *---------------------------------------------------------------------- 
 * wClientCheckProperty--
 * 	Handles PropertyNotify'es, verifying which property was
 * changed and updating internal state according to that, like redrawing
 * the icon title when it is changed.
 * 
 * Side effects:
 * 	Depends on the changed property.
 * 
 * TODO: CLIENT_LEADER, _GNUSTEP_WM_ATTR
 *---------------------------------------------------------------------- 
 */
void
wClientCheckProperty(WWindow *wwin, XPropertyEvent *event) 
{
    XWindowAttributes attribs;
#ifdef I18N
    XTextProperty text_prop;
    char **list;
    int num;
#endif
    
    switch (event->atom) {
     case XA_WM_NAME:
	/* window title was changed */
        if (wwin->title!=NULL)
	  XFree(wwin->title);
#ifndef I18N
	if (!XFetchName(dpy, wwin->client_win, &(wwin->title))) {
	  wwin->title = NULL;
	  wwin->title_width = XTextWidth(wwin->screen_ptr->title_font.font,
					 DEF_WINDOW_TITLE,
					 strlen(DEF_WINDOW_TITLE));
	} else {
	  wwin->title_width = XTextWidth(wwin->screen_ptr->title_font.font, 
					 wwin->title, strlen(wwin->title));
	}
#else
	if (XGetWMName(dpy, wwin->client_win, &text_prop) != 0) {
	  if (text_prop.value)
	    text_prop.nitems = strlen(text_prop.value);
	  if (XmbTextPropertyToTextList(dpy, &text_prop, &list, &num) >= 
	      Success && num > 0 && *list) {
	    wwin->title = *list;
	    wwin->title_width = XmbTextEscapement(wwin->screen_ptr->title_font.font,
						  wwin->title,
						  strlen(wwin->title));
	  } else {
	    wwin->title = NULL;
	    wwin->title_width = XmbTextEscapement(wwin->screen_ptr->title_font.font,
						  DEF_WINDOW_TITLE,
						  strlen(DEF_WINDOW_TITLE));
	  }
	} else {
	  wwin->title = NULL;
	  wwin->title_width = XmbTextEscapement(wwin->screen_ptr->title_font.font,
						DEF_WINDOW_TITLE,
						strlen(DEF_WINDOW_TITLE));
	}
#endif
	wWindowPaint(wwin, BORDER_TOP, False);
	break;
	
     case XA_WM_ICON_NAME:
	if (!wwin->icon) break;

	/* icon title was changed */
	wIconUpdate(wwin->icon);
	break;
	
     case XA_WM_HINTS:
	/* WM_HINTS */

	if (wwin->wm_hints)
	  XFree(wwin->wm_hints);
	wwin->wm_hints = XGetWMHints(dpy, wwin->client_win);

	if (wwin->wm_hints)  {
#if 1 /* TODO: fix window_group change */
	    Window old_mainwin = wwin->main_window;

	    if (wwin->wm_hints->flags & WindowGroupHint) {
		wwin->group_id = wwin->wm_hints->window_group;
		wwin->main_window = wwin->group_id;
	    } else {
		if (wwin->client_leader == None) {
		    wwin->main_window = wwin->client_leader;
		} else {
		    wwin->main_window = None;
		}
		wwin->group_id = None;
	    }
	    if (wwin->transient_for == None ||
		wwin->transient_for == wwin->screen_ptr->root_win) {
		if ((wwin->main_window==None || old_mainwin!=wwin->main_window)
		    && old_mainwin!=None) {
		    WApplication *wapp;

		    puts("change main window");
		    wapp = wApplicationOf(old_mainwin);
		    if (wapp)
		      wApplicationChange(wapp, wwin->main_window);
		}
		if (wwin->main_window!=None
		    && !wApplicationOf(wwin->main_window)) {
		    wApplicationCreate(wwin->screen_ptr, wwin->main_window);
		}
	    }
#endif
	    /* update icon */
	    if ((wwin->wm_hints->flags & IconPixmapHint)
		|| (wwin->wm_hints->flags & IconWindowHint)) {
		if (wwin->flags.miniaturized) {
		    wIconUpdate(wwin->icon);
		} else if (wwin->flags.hidden) {
		    WApplication *wapp;
		    wapp = wApplicationOf(wwin->main_window);
		    if (wapp) {
			wIconUpdate(wapp->main_window_desc->icon);
		    }
		}
	    }

	    if (wwin->wm_hints->flags & UrgencyHint)
	      wwin->flags.urgent = 1;
	    else
	      wwin->flags.urgent = 0;
	} else {
	    wwin->group_id = None;
	}
	break;
	
     case XA_WM_NORMAL_HINTS:
	/* normal (geometry) hints */
	
	XGetWindowAttributes(dpy, wwin->client_win, &attribs);
	GetNormalHints(wwin, &attribs, False);
	break;
	
     case XA_WM_TRANSIENT_FOR:
	{
	    Window new_owner;
	    WWindow *owner;
	    
	    if (!PropGetTransientForHint(wwin->client_win, &new_owner)) {
		new_owner = None;
	    } else {
		if (new_owner==0 || new_owner == wwin->client_win) {
		    new_owner = wwin->screen_ptr->root_win;
		}
	    }
	    if (new_owner!=wwin->transient_for) {
		owner = wWindowFor(wwin->transient_for);
		if (owner) {
		    if (owner->flags.semi_focused) {
			owner->flags.semi_focused = 0;
			if (owner->flags.mapped || owner->flags.shaded)
			  wWindowPaint(owner, BORDER_ALL, False);
		    }
		}
		owner = wWindowFor(new_owner);
		if (owner) {
		    if (!owner->flags.semi_focused) {
			owner->flags.semi_focused = 1;
			if (owner->flags.mapped || owner->flags.shaded)
			  wWindowPaint(owner, BORDER_ALL, False);
		    }
		}
		wwin->transient_for = new_owner;
		if (new_owner==None) {
		    if (!wwin->window_flags.miniaturizable) {
			wwin->window_flags.miniaturizable = 1;
			wWindowConfigureBorders(wwin);
		    }
		} else if (wwin->window_flags.miniaturizable) {
		    wwin->window_flags.miniaturizable = 0;
		    wWindowConfigureBorders(wwin);
		}
	    }
	}
	break;
	
     default:
	if (event->atom==_XA_WM_PROTOCOLS) {
	    PropGetProtocols(wwin->client_win, &wwin->protocols);
	    if (wwin->protocols.DELETE_WINDOW) {
		wwin->window_flags.broken_close = 0;
	    } else {
		wwin->window_flags.broken_close = 1;		
	    }
	    wWindowUpdateButtonImages(wwin);

	    wWindowPaint(wwin, BORDER_ALL, False);
	} else if (event->atom==_XA_WM_COLORMAP_WINDOWS) {
	    GetColormapWindows(wwin);
	    ReInstallActiveColormap();
	}
    }
}


/*
 *---------------------------------------------------------------------- 
 * GetNormalHints--
 * 	Get size (normal) hints and a default geometry for the client 
 * window. The hints are also checked for inconsistency.
 * 
 * Side effects:
 * 	normal_hints is filled with valid data and geometry information 
 * about the client is also changed.
 *---------------------------------------------------------------------- 
 */
void
GetNormalHints(WWindow *wwin, XWindowAttributes *wattribs, int geometry)
{
    int pre_icccm=0;
    
    /* find a position for the window */
    wwin->normal_hints = XAllocSizeHints();
    if (!PropGetNormalHints(wwin->client_win, wwin->normal_hints, &pre_icccm)) {
	wwin->normal_hints->flags = 0;
    }
    wwin->client.x = wattribs->x;
    wwin->client.y = wattribs->y;
    
    wwin->client.width = wattribs->width;
    wwin->client.height = wattribs->height;

    if (!(wwin->normal_hints->flags & PWinGravity)) {
	wwin->normal_hints->win_gravity = NorthWestGravity;
    }
    if (!(wwin->normal_hints->flags & PMinSize)) {
	wwin->normal_hints->min_width = MIN_WSIZE;
	wwin->normal_hints->min_height = MIN_WSIZE;
    }
    if (!(wwin->normal_hints->flags & PBaseSize)) {
	wwin->normal_hints->base_width = wwin->normal_hints->min_width;
	wwin->normal_hints->base_height = wwin->normal_hints->min_height;
    }    
    if (!(wwin->normal_hints->flags & PMaxSize)) {
	wwin->normal_hints->max_width = wwin->screen_ptr->scr_width*2;
	wwin->normal_hints->max_height = wwin->screen_ptr->scr_height*2;
    }
    
    if (!(wwin->normal_hints->flags & PResizeInc)) {
	wwin->normal_hints->width_inc = 1;
	wwin->normal_hints->height_inc = 1;
    } else {	
	if (wwin->normal_hints->width_inc <= 0)
	  wwin->normal_hints->width_inc = 1;
	if (wwin->normal_hints->height_inc <= 0)
	  wwin->normal_hints->height_inc = 1;
    }

    if (wwin->normal_hints->min_width <= 0) 
      wwin->normal_hints->min_width = MIN_WSIZE;
    if (wwin->normal_hints->min_height <= 0)
      wwin->normal_hints->min_height = MIN_WSIZE;

    if (wwin->normal_hints->min_height > wwin->normal_hints->max_height) {
	wwin->normal_hints->min_height = wwin->normal_hints->max_height;
    }
    if (wwin->normal_hints->min_width > wwin->normal_hints->max_width) {
	wwin->normal_hints->min_width = wwin->normal_hints->max_width;
    }

    /* pre ICCCM (old) client */
    if (pre_icccm && !wwin->screen_ptr->flags.startup && geometry) {
#ifdef DEBUG
	printf("PRE ICCCM\n");	
#endif	
	if (wwin->normal_hints->flags & (USPosition|PPosition)) {
	    wwin->client.x = wwin->normal_hints->x;
	    wwin->client.y = wwin->normal_hints->y;
	}
	if (wwin->normal_hints->flags & (USSize|PSize)) {
	    wwin->client.width = wwin->normal_hints->width;
	    wwin->client.height = wwin->normal_hints->height;
	}
    }
}


void 
GetColormapWindows(WWindow *wwin)
{
    if (wwin->cmap_windows) {
	XFree(wwin->cmap_windows);
    }
    if (!XGetWMColormapWindows(dpy, wwin->client_win, &(wwin->cmap_windows),
			       &(wwin->cmap_window_no))) {
	wwin->cmap_window_no = 0;
	wwin->cmap_windows = NULL;
    }
}


