/* dialog.c - dialog windows for internal use
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>


#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "wcore.h"
#include "button.h"
#include "wconfig.h"
#include "actions.h"
#include "window.h"
#include "texture.h"
#include "funcs.h"
#include "dialog.h"
#include "pixmap.h"
#include "dlg_pixmaps.h"
#include "properties.h"



extern XContext wWinContext;



static void button1Click(WObjDescriptor *obj, XEvent *event);
static void button2Click(WObjDescriptor *obj, XEvent *event);


static WMethodTable button1Methods;
static WMethodTable button2Methods;


void
wDialogInitialize()
{
    memcpy(&button1Methods, &wButtonMethods, sizeof(WMethodTable));
    memcpy(&button2Methods, &wButtonMethods, sizeof(WMethodTable));
    button1Methods.click = button1Click;
    button2Methods.click = button2Click;
}


void 
wDialogDestroy(WDialog *dlg)
{
    wButtonDestroy(dlg->but1);
    if (dlg->but2)
      wButtonDestroy(dlg->but2);
    wCoreDestroy(dlg->core);
    wUnmanageWindow(dlg->wwin, False);
    if (dlg->pixmap)
      wPixmapDestroy(dlg->pixmap);
    XFreeGC(dpy, dlg->gc);
    wfree(dlg);
}


void
wDialogPaint(WDialog *dlg)
{
    WScreen *scr = dlg->core->screen_ptr;
    Window win = dlg->core->window;
    char *msg;
    int i, y=0;
    
    XDrawLine(dpy, win, 
	      scr->normal_button_texture->solid.dim_gc, 0, 80, 400, 80);
    XDrawLine(dpy, win, 
	      scr->normal_button_texture->solid.light_gc, 0, 81, 400, 81);
#ifndef I18N    
    XSetFont(dpy, dlg->gc, scr->panel_title_font.font->fid);
    XDrawString(dpy, win, dlg->gc, 80, 40, dlg->title, strlen(dlg->title));
    XSetFont(dpy, dlg->gc, scr->panel_text_font.font->fid);    
#else
    XmbDrawString(dpy, win, scr->panel_title_font.font, dlg->gc, 80, 40, 
		  dlg->title, strlen(dlg->title));
#endif

    /* paint multi-line message */
    msg = dlg->message;
    while (*msg!=0) {
	i=0;
	while (msg[i]!='\n' && msg[i]!=0) i++;
#ifndef I18N
	XDrawString(dpy, win, dlg->gc, 20, 100+y, msg, i);
#else
	XmbDrawString(dpy, win, scr->panel_text_font.font, dlg->gc, 20, 100+y, 
		      msg, i);
#endif
	msg += i;
	y += dlg->text_h;
	if (*msg==0) break;
	msg++;
    }
    wButtonPaint(dlg->but1, False);
    if (dlg->but2)
      wButtonPaint(dlg->but2, False);

    if (dlg->pixmap) {
	XSetClipMask(dpy, dlg->gc, dlg->pixmap->mask);
	if (dlg->pixmap->depth==1)
	  XCopyPlane(dpy, dlg->pixmap->image, dlg->core->window, 
		     dlg->gc, 0, 0, dlg->pixmap->width,
		     dlg->pixmap->height, 15, 15, 1);
	else
	  XCopyArea(dpy, dlg->pixmap->image, dlg->core->window, 
		    dlg->gc, 0, 0, dlg->pixmap->width,
		    dlg->pixmap->height, 15, 15);
	XSetClipMask(dpy, dlg->gc, None);
    }
}


int
wMessageDialog(WScreen *scr, char *title, char *message, int type)
{
    XEvent event;
    WDialog *dlg;
    XGCValues gcv;
    int selected;
    GNUstepWMAttributes gattr;
    XClassHint *chint;
    KeyCode kcode;
    XWindowAttributes attr;

    dlg = wmalloc(sizeof(WDialog));
    dlg->message = message;
    dlg->title = title;
    dlg->selected = 0;
    dlg->type = type;
    dlg->core = wCoreCreateTopLevel(scr, (scr->scr_width-400)/2,
				    (scr->scr_height-180)/2, 400, 180, 0,
				    True);
    XSetWindowBackground(dpy, dlg->core->window,
			 scr->normal_button_texture->any.color.pixel);

    gattr.flags = (GSWindowStyleAttr|GSWindowLevelAttr);
    gattr.window_style = (NSTitledWindowMask);
    gattr.window_level = NSFloatingWindowLevel;
    PropWriteGNUstepWMAttr(dlg->core->window, &gattr);

    dlg->core->descriptor.self = dlg;
    dlg->core->descriptor.parent_type = WCLASS_DIALOG;
    dlg->core->descriptor.parent = dlg;

    chint = XAllocClassHint();
    chint->res_name = "";
    chint->res_class = "WindowMaker";
    XSetClassHint(dpy, dlg->core->window, chint);
    XFree(chint);

    gcv.foreground = scr->black_pixel;
    dlg->gc = XCreateGC(dpy, dlg->core->window, GCForeground, &gcv);
    XSetClipOrigin(dpy, dlg->gc, 15, 15);
    XStoreName(dpy, dlg->core->window, "");
#ifdef XPM
    dlg->pixmap = wPixmapCreateFromXPMData(scr, GNUSTEP_XPM);
#else
    dlg->pixmap =
      wPixmapCreateFromXBMData(scr, GNUSTEP_XBM_DATA,GNUSTEP_MASK_XBM_DATA,
			       GNUSTEP_XBM_WIDTH, GNUSTEP_XBM_HEIGHT,
			       scr->black_pixel, scr->white_pixel);
#endif /* XPM */
    switch (type) {
     case WD_INFORMATION:
	dlg->but1 = wButtonCreate(dlg->core, 320, 145, 70, 24);
	dlg->but1->label = "OK ";
	dlg->but1->label_gc = dlg->gc;
	dlg->but1->image = scr->button_arrow;
	dlg->but1->core->descriptor.methods = &button1Methods;
	dlg->but1->core->descriptor.parent_type = WCLASS_DIALOG;
	dlg->but1->core->descriptor.parent = dlg;
	dlg->but2=NULL;
	break;

     case WD_EXIT_CONFIRM:
	dlg->but1 = wButtonCreate(dlg->core, 240, 145, 70, 24);
	dlg->but2 = wButtonCreate(dlg->core, 320, 145, 70, 24);
	dlg->but1->label = "Exit";
	dlg->but1->label_gc = dlg->gc;
	dlg->but2->label = "Cancel";
	dlg->but2->label_gc = dlg->gc;
	dlg->but2->image = scr->button_arrow;
	dlg->but1->core->descriptor.methods = &button1Methods;
	dlg->but1->core->descriptor.parent_type = WCLASS_DIALOG;
	dlg->but1->core->descriptor.parent = dlg;
	dlg->but2->core->descriptor.methods = &button2Methods;
	dlg->but2->core->descriptor.parent_type = WCLASS_DIALOG;
	dlg->but2->core->descriptor.parent = dlg;
	break;
    }
    dlg->wwin = wManageWindow(scr, dlg->core->window);
    XMapSubwindows(dpy, dlg->core->window);
    XMapWindow(dpy, dlg->core->window);
    XFlush(dpy); 
    dlg->text_h = scr->panel_text_font.height;

    XGetWindowAttributes(dpy, dlg->core->window, &attr);    
    XSelectInput(dpy, dlg->core->window, attr.your_event_mask|KeyPressMask
		 |KeyReleaseMask);

    kcode = XKeysymToKeycode(dpy, XK_Return);
    dlg->selected = -1;
    while (dlg->selected < 0) {
/*
	WObjDescriptor *desc;
*/	
	XNextEvent(dpy, &event);
	/*
	if (event.type == KeyPress
	    && event.xkey.window == dlg->core->window) {
	    if (event.xkey.keycode == kcode) {
		if (XFindContext(dpy, event.xkey.window, wWinContext, 
				 (XPointer *)&desc)!=XCNOENT) {
		    if (desc->methods && desc->methods->key_press!=NULL) {
			(*desc->methods->key_press)(desc, &event);
			continue;
		    }
		}
	    }
	}
	 */
	DispatchEvent(&event);
    }
    selected = dlg->selected;
    wDialogDestroy(dlg);
    return selected;
}


static void 
button1Click(WObjDescriptor *obj, XEvent *event)
{
    WDialog *dlg = obj->parent;

    switch (dlg->type) {
     case WD_INFORMATION:
	dlg->selected=WDB_OK;
	break;
     case WD_EXIT_CONFIRM:
	dlg->selected=WDB_EXIT;
	break;
     default:
	dlg->selected=-1;
    }
    
}


static void 
button2Click(WObjDescriptor *obj, XEvent *event)
{
    WDialog *dlg = obj->parent;
    
    switch (dlg->type) {
     case WD_EXIT_CONFIRM:
	dlg->selected=WDB_CANCEL;
	break;
     default:
	dlg->selected=-1;
    }
}

