/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the wfree Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the wfree Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef XPM
#include <X11/xpm.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "raster.h"
#include "screen.h"
#include "pixmap.h"

extern WPreferences wPreferences;

#ifdef DITHERING
#ifdef XPM
/*
 *----------------------------------------------------------------------
 * loadXPMImage--
 * 	Loads a XPM file into a WImage with alpha channel.
 * 
 * Returns:
 * 	The loaded image.
 *---------------------------------------------------------------------- 
 */
static WImage *
loadXPMImage(WScreen *scr, char *file)
{
    WImage *img;
    XpmImage image;
    unsigned char *color_table[3];
    unsigned int *p;
    int i, ofs, x, y;
    int transp=-1;
    
    if (XpmReadFileToXpmImage(file, &image, (XpmInfo *)NULL)!=XpmSuccess) {
	return NULL;
    }
    if (image.height<2) {
	wWarning("can't load pixmap file \"%s\". Size too weird", file);
	XpmFreeXpmImage(&image);
	return NULL;
    }

    if (image.colorTable==NULL) {
	XpmFreeXpmImage(&image);
	return NULL;
    }
    img = malloc(sizeof(WImage));
    if (img==NULL) {
	goto out_of_mem;
    }
    img->data = malloc(image.width*image.height*3);
    if (img->data==NULL) {
	wfree(img);

	out_of_mem:
	wWarning("virtual memory exhausted during pixmap load. Not fatal...");
	XpmFreeXpmImage(&image);
	return NULL;
    }
    img->width = image.width;
    img->height = image.height;
    
    /* make color table */
    for (i=0; i<3; i++) {
	color_table[i] = malloc(image.ncolors*sizeof(char));
	if (!color_table[i]) {
	    for (i=i-1;i>=0; i--) {
		if (color_table[i])
		  free(color_table[i]);
	    }
	    wfree(img->data);
	    wfree(img);	    
	    wWarning("virtual memory exhausted. Continuing...");
	    XpmFreeXpmImage(&image);
	    return NULL;
	}
    }

    for (i=0; i<image.ncolors; i++) {
	XColor xcolor;
	
	if (strncmp(image.colorTable[i].c_color,"None",4)==0) {
	    color_table[0][i]=0xbe;
	    color_table[1][i]=0xbe;
	    color_table[2][i]=0xbe;
	    transp = i;
	    continue;
	}
	if (XParseColor(dpy, scr->w_colormap, image.colorTable[i].c_color, 
			&xcolor)) {
	    color_table[0][i] = xcolor.red>>8;
	    color_table[1][i] = xcolor.green>>8;
	    color_table[2][i] = xcolor.blue>>8;
	} else {
	    color_table[0][i]=0xbe;
	    color_table[1][i]=0xbe;
	    color_table[2][i]=0xbe;
	}
    }

/*    
    img->alpha = malloc(image.width*image.height);
    if (img->alpha) {
	memset(img->alpha, 0xff, image.width*image.height);	
    } else {
	transp = -1;
    }
 */
    /* convert image to WImage */    
    p = image.data;
    ofs = 0;
    for (y=0; y<image.height; y++) {
	for (x=0; x<image.width; x++) {
	    /*
	    if (*p==transp)
	      img->alpha[ofs]=0;
	     */
	    img->data[ofs*3]=color_table[0][*p];
	    img->data[ofs*3+1]=color_table[1][*p];
	    img->data[ofs*3+2]=color_table[2][*p];
	    ofs++;
	    p++;
	}
    }
    for(i=0; i<3; i++) {
	free(color_table[i]);
    }
    XpmFreeXpmImage(&image);
    return img;
}
#endif /* XPM */



/*
 *----------------------------------------------------------------------
 * LoadImageFile--
 * 	Guesses the file type and loads an image file.
 * 
 * Returns:
 * 	WImage or NULL on error.
 *---------------------------------------------------------------------- 
 */
WImage *
LoadImageFile(WScreen *scr, char *path)
{
    WImage *img;
 
    if (scr->visual->class!=TrueColor 
	&& !scr->rscreen->flags.dither) return NULL;
    switch (IdentPixmapFile(path)) {
     case PIX_XPM:
#ifdef XPM
	img = loadXPMImage(scr, path);
	if (!img)
	  wWarning("could not load XPM file \"%s\"", path);	
	return img;
#else
	wWarning("cannot load \"%s\": XPM format not supported", path);
	return NULL;
#endif /* XPM */

     case PIX_XBM:
	  wWarning("can't convert XBM file \"%s\" to internal format", path);
	return NULL;

     case PIX_INVALID:
	wSysError("can't identify image file \"%s\"", path);
	break;
	
     default:
	wWarning("unknown image format in \"%s\"", path);
    }
    return NULL;
}



/*
 *----------------------------------------------------------------------
 * wDitherImage_pseudo8--
 * 	Dithers and converts a 24bit WImage into a 8bpp Pixmap.
 * 
 * Returns:
 * 	A 8bpp pixmap of the same size as the input image.
 *---------------------------------------------------------------------- 
 */
Pixmap
wDitherImage_pseudo8(WScreen *scr, WImage *image)
{
    WScreenDescriptor *ws = scr->rscreen;
    XImage *ximg;
    unsigned char *dith;
    unsigned char *error, *nerror, *tmp;
    int x, y, r, g, b, rd, gd, bd, rerr, gerr, berr, ofs, i;
    Pixmap pix;

    dith = wmalloc(image->width*image->height);
    nerror = wmalloc((image->width+1)*3);
    error = wmalloc((image->width+1)*3);

    tmp = error;
    for (i=0; i<image->width*3; i++) {
	*tmp = image->data[i];
	tmp++;
    }
    /* convert and dither the image to WImage */
    rd = 0xff/(ws->r_colors);
    gd = 0xff/(ws->g_colors);
    bd = 0xff/(ws->b_colors);
    ofs = 0;
    for (y=0; y<image->height; y++) {
	if (y<image->height-1) {
	    tmp = nerror;
	    for (i=ofs+(image->width)*3; i<ofs+(image->width)*3*2; i++) {
		*tmp = image->data[i];
		tmp++;
	    }
	    *tmp = *(tmp-3);
	    tmp++;
	    *tmp = *(tmp-3);
	    tmp++;
	    *tmp = *(tmp-3);
	    tmp++;
	}
	for (x=0; x<image->width; x++) {
	    /* reduce pixel */
	    r = (error[x*3] * ws->r_colors)/0xff;
	    g = (error[x*3+1] * ws->g_colors)/0xff;
	    b = (error[x*3+2] * ws->b_colors)/0xff;
	    if (r<0) r=0; else if (r>=ws->r_colors) r=ws->r_colors-1;
	    if (g<0) g=0; else if (g>=ws->g_colors) g=ws->g_colors-1;
	    if (b<0) b=0; else if (b>=ws->b_colors) b=ws->b_colors-1;

	    i = r*ws->g_colors*ws->b_colors + g*ws->b_colors + b;
	    dith[y*image->width+x]= ws->colors[i].pixel;
	    /* calc error */
	    rerr = error[x*3] - r*rd;
	    gerr = error[x*3+1] - g*gd;
	    berr = error[x*3+2] - b*bd;

	    /* distribute error */
	    /* x+1, y */
	    error[(x+1)*3]+=(rerr*3)/8;
	    error[(x+1)*3+1]+=(gerr*3)/8;
	    error[(x+1)*3+2]+=(berr*3)/8;
	    /* x, y+1 */
	    nerror[x*3]+=(rerr*3)/8;
	    nerror[x*3+1]+=(gerr*3)/8;
	    nerror[x*3+2]+=(berr*3)/8;
	    /* x+1, y+1 */
	    nerror[(x+1)*3]+=(rerr*2)/8;
	    nerror[(x+1)*3+1]+=(gerr*2)/8;
	    nerror[(x+1)*3+2]+=(berr*2)/8;
	    ofs+=3;
	}
	/* skip to next line */
	tmp = error;
	error = nerror;
	nerror = tmp;
    }
    wfree(error);
    wfree(nerror);
    ximg = XCreateImage(dpy, scr->w_visual, 8, ZPixmap, 0, (char*)dith, 
			image->width, image->height, XBitmapPad(dpy), 0);
    XSetClipMask(dpy, scr->copy_gc, None);
    pix = XCreatePixmap(dpy, scr->w_win, image->width, image->height, 8);
    XPutImage(dpy, pix, scr->copy_gc, ximg, 0, 0, 0, 0, 
	      image->width, image->height);
    XDestroyImage(ximg);
    return pix;
}


Pixmap 
wDitherImage_truecolor(WScreen *scr, WImage *image)
{
    WScreenDescriptor *ws = scr->rscreen;
    XImage *ximg;
    int x, y, i;
    int r, g, b, rw, gw, bw, rm, gm, bm;
    unsigned long pixel;
    Pixmap pixmap;
    unsigned char *data;

    pixmap = XCreatePixmap(dpy, scr->w_win, image->width, image->height,
			   scr->w_depth);

    ximg = XCreateImage(dpy, scr->w_visual, scr->w_depth, ZPixmap, 0, NULL,
			image->width, image->height, XBitmapPad(dpy), 0);

    data = image->data;
    ximg->data = malloc(ximg->bytes_per_line*ximg->height);
    if (!ximg->data) {
	XDestroyImage(ximg);
	XFreePixmap(dpy, pixmap);
	return None;
    }
#define RED(i) (data[(i)*3]<<8)
#define GREEN(i) (data[(i)*3+1]<<8)
#define BLUE(i)	(data[(i)*3+2]<<8)
    rw = ws->r_shift;
    gw = ws->g_shift;
    bw = ws->b_shift;
    rm = scr->w_visual->red_mask;
    gm = scr->w_visual->green_mask;
    bm = scr->w_visual->blue_mask;
    if (scr->w_depth >= 24 || !wPreferences.use_dithering) {
	i=0;
	for (y=0; y<image->height; y++) {
	    for (x=0; x<image->width; x++) {
		r = (rw<0 ? RED(i) << (-rw) : RED(i) >> rw);
		g = (gw<0 ? GREEN(i) << (-gw) : GREEN(i) >> gw);
		b = (bw<0 ? BLUE(i) << (-bw) : BLUE(i) >> bw); 
		pixel = ((r & rm) | (g & gm) | (b & bm));
		XPutPixel(ximg, x, y, pixel);
		i++;
	    }
	}
    } else {
	int rerr, gerr, berr, ofs;
	short *nerror, *error, *tmp;
	nerror = wmalloc((image->width+2)*3*sizeof(short));
	error = wmalloc((image->width+2)*3*sizeof(short));

	tmp = error;
	for (i=0; i<image->width*3; i++) {
	    *tmp = image->data[i];
	    tmp++;
	}
	i = 0;
	for (y=0; y<image->height; y++) {
	    if (y<image->height-1) {
		tmp = nerror;
	      	for (i=(y+1)*image->width*3; i<(y+2)*image->width*3; i++) {
		    *tmp = image->data[i];
		    tmp++;
		}
	    }
	    ofs = 0;
	    for (x=0; x<image->width; x++) {
		/* reduce pixel */
		if (error[ofs]>255) error[ofs]=255;
		if (error[ofs+1]>255) error[ofs+1]=255;
		if (error[ofs+2]>255) error[ofs+2]=255;
		r = (rw<0 ? error[ofs] << (-rw+8) : (error[ofs]<<8) >> rw)&rm;
		g = (gw<0 ? error[ofs+1] << (-gw+8) : (error[ofs+1]<<8) >> gw)&gm;
		b = (bw<0 ? error[ofs+2] << (-bw+8) : (error[ofs+2]<<8) >> bw)&bm;
		pixel = r|g|b;
		XPutPixel(ximg, x, y, pixel);
		
		/* calc error */
		rerr = error[ofs] - ((rw<0 ? r >> (-rw) : r << rw)>>8);
		gerr = error[ofs+1] - ((gw<0 ? g >> (-gw) : g << gw)>>8);
		berr = error[ofs+2] - ((bw<0 ? b >> (-bw) : b << bw)>>8);
		/* distribute error */
		/* x+1, y */
		error[(x+1)*3]+=(rerr*3)/8;
		error[(x+1)*3+1]+=(gerr*3)/8;
		error[(x+1)*3+2]+=(berr*3)/8;
		/* x, y+1 */
		nerror[ofs]+=(rerr*3)/8;
		nerror[ofs+1]+=(gerr*3)/8;
		nerror[ofs+2]+=(berr*3)/8;
		/* x+1, y+1 */
		nerror[(x+1)*3]+=(rerr*2)/8;
		nerror[(x+1)*3+1]+=(gerr*2)/8;
		nerror[(x+1)*3+2]+=(berr*2)/8;
		ofs+=3;
	    }
	    /* skip to next line */
	    tmp = error;
	    error = nerror;
	    nerror = tmp;
	}
	wfree(nerror);
	wfree(error);
    }
#undef RED
#undef GREEN
#undef BLUE
    XSetClipMask(dpy, scr->copy_gc, None);
    XPutImage(dpy, pixmap, scr->copy_gc, ximg, 0, 0, 0, 0,
	      image->width, image->height);
    XDestroyImage(ximg);
    return pixmap;
}



#if notdef
WImage*
wDitherImage_wimage(WScreen *scr, WImage *image)
{
    WScreenDescriptor *ws = scr->rscreen;
    XImage *ximg;
    unsigned char *dith;
    unsigned char *error, *nerror;
    int x, y, rd, gd, bd, rerr, gerr, berr;
    
    dith = wmalloc(image->width*image->height);
    nerror = wmalloc((image->width+2)*3);
    error = wmalloc((image->width+2)*3);
    
    memcpy(error, image->data, image->width);

    /* divide each color by these values, so they'll be mapped to
     * the available colors */
    /* convert and dither the image to WImage */    
    ofs = 0;
    for (c=0; c<3; c++) { /* for each channel...*/
	switch (c) {
	 case 0:
	    max = ws->r_colors;
	    break;
	 case 1:
	    max = ws->g_colors;
	    break;
	 case 2:
	    max = ws->b_colors;
	    break;
	}
	d = 0xffff/max;
	for (y=0; y<image->height; y++) { /* for each line...*/
	    ofs = y*image->width;
	    for (x=c; x<image->width*3; x+=3) { /* each pixel of the channel */
		i = image->data[ofs+x];
		/* copy image and reduce */
		i = error[x]/d;
		if (i<0) i=0; else if (i>=max) r=max-1;
		dith[ofs+x]=i;
		/* calc error */
		err = error[x] - i*max;
		
		/* distribute error */
		/* x+1, y */
		error[x+1]+=(err*3)/8;
		/* x, y+1 */
		nerror[x]+=(err*3)/8;
		/* x+1, y+1 */
		nerror[x+1]+=(err*2)/8;
		ofs++;
	    }
	    /* skip to next line */
	    tmp = nerror;
	    nerror = error;
	    error = tmp;
	}
    }
}
#endif /* notdef */

#ifdef XPM
static int
getXPMImage(WScreen *scr, char *file, XpmImage *image,
	    unsigned short *color_table[3], int *transp)
{
    int i;
    
    *transp = -1;
    
    if (XpmReadFileToXpmImage(file, image, (XpmInfo *)NULL)!=XpmSuccess) {
	return 0;
    }
    if (image->height<2) {
	wWarning("can't load pixmap file \"%s\". Size too weird", file);
	XpmFreeXpmImage(image);
	return 0;
    }

    if (image->colorTable==NULL) {
	XpmFreeXpmImage(image);
	return 0;
    }
    for (i=0; i<3; i++) {
	color_table[i] = malloc(image->ncolors*sizeof(unsigned short));
	if (!color_table[i]) {
	    for (i=i-1;i>=0; i--) {
		if (color_table[i])
		  free(color_table[i]);
	    }
	    wWarning("virtual memory exhausted. Continuing...");
	    XpmFreeXpmImage(image);
	    return 0;
	}
    }

    for (i=0; i<image->ncolors; i++) {
	XColor xcolor;
	
	if (strncmp(image->colorTable[i].c_color,"None",4)==0) {
	    color_table[0][i]=0xbefb;
	    color_table[1][i]=0xbefb;
	    color_table[2][i]=0xbefb;
	    *transp = i;
	    continue;
	}
	if (XParseColor(dpy, scr->w_colormap, image->colorTable[i].c_color, 
			&xcolor)){
	    color_table[0][i] = xcolor.red;
	    color_table[1][i] = xcolor.green;
	    color_table[2][i] = xcolor.blue;
	} else {
	    color_table[0][i]=0xbefb;
	    color_table[1][i]=0xbefb;
	    color_table[2][i]=0xbefb;
	}
    }
    return 1;
}


static WImage *
loadXPMIcon(WScreen *scr, WImage *back, int titleh, char *file)
{
    WScreenDescriptor *ws=scr->rscreen;
    WImage *img;
    XpmImage image;
    unsigned short *color_table[3];
    int *error[3];
    int *error_next[3], *tmp;
    int i, ofs, rd, gd, bd, x, y, r, g, b, ix, iy, fx, fy;
    register int err, erg, erb, pos;
    int transp=-1;
    
    img = malloc(sizeof(WImage));
    if (!img) {
	wWarning("virtual memory exhausted during XPM load.");
	return NULL;
    }
    img->data = malloc(ICON_WIDTH*ICON_HEIGHT*3);
    if (!img->data) {
	free(img);
	wWarning("virtual memory exhausted during XPM load.");
	return NULL;
    }
    img->width = ICON_WIDTH;
    img->height = ICON_HEIGHT;
    
    /* make error table */
    for (i=0; i<3; i++) {
	error[i] = malloc((img->width+2)*sizeof(int));
	error_next[i] = malloc((img->width+2)*sizeof(int));
	if (!error[i] || !error_next[i]) {
	    for (i=i-1;i>=0; i--) {
		if (error[i])
		  free(error[i]);
		if (error_next[i])
		  free(error_next[i]);
	    }
	    free(img->data);
	    free(img);
	    wWarning("virtual memory exhausted during XPM load.");
	    return NULL;
	}
    }

    img->alpha = NULL;
    
    if (!getXPMImage(scr, file, &image, color_table, &transp)) {
	free(img->data);
	free(img);
	img = NULL;
	goto end;
    }
   
    ix = (img->width - image.width)/2;
    iy = titleh + (img->height - image.height - titleh - 1)/2;
    fx = img->width - ix;
    fy = iy + image.height-1;
    /* init error table for the first line */
    i = 0;
    for (x=0; x<img->width; x++) {
	if (iy==0 && x>=ix && x<=fx) {
	    error[0][x] = color_table[0][image.data[i]];
	    error[1][x] = color_table[1][image.data[i]];
	    error[2][x] = color_table[2][image.data[i]];
	    i++;
	} else {
	    error[0][x] = back->data[x*3]<<8;
	    error[1][x] = back->data[x*3+1]<<8;
	    error[2][x] = back->data[x*3+2]<<8;
	}
    }

    /* divide each color by these values, so they'll be mapped to
     * the available colors */
    rd = 0xffff/(ws->r_colors-1);
    gd = 0xffff/(ws->g_colors-1);
    bd = 0xffff/(ws->b_colors-1);
    /* convert and dither the image to WImage */    
    ofs = 0;
    
    for (y=0; y<img->height; y++) {
	unsigned int *p;
	if (y!=img->height-1) {
	    p = image.data + (y-iy)*image.width;
	    i = (y+1)*back->width*3;
	    for (x=0; x<img->width; x++) {
		if (y>=iy && y<fy && x>=ix && x<fx) {
		    if (*p==transp) {
			error_next[0][x]=back->data[i]<<8;
			error_next[1][x]=back->data[i+1]<<8;
			error_next[2][x]=back->data[i+2]<<8;
		    } else {
			error_next[0][x]=color_table[0][*p];
			error_next[1][x]=color_table[1][*p];
			error_next[2][x]=color_table[2][*p];
		    }
		    p++;
		} else {
		    error_next[0][x]=back->data[i]<<8;
		    error_next[1][x]=back->data[i+1]<<8;
		    error_next[2][x]=back->data[i+2]<<8;
		}
		i+=3;
	    }
	}
	for (x=0; x<img->width; x++) {
	    i = image.data[ofs];
	    pos = ofs*3;
	    /* copy image and reduce */
	    r = error[0][x]/rd;
	    g = error[1][x]/gd;
	    b = error[2][x]/bd;
	    if (r<0) r=0; else if (r>=ws->r_colors) r=ws->r_colors-1;
	    if (g<0) g=0; else if (g>=ws->g_colors) g=ws->g_colors-1;
	    if (b<0) b=0; else if (b>=ws->b_colors) b=ws->b_colors-1;
	    img->data[pos]=r; 
	    img->data[pos+1]=g;
	    img->data[pos+2]=b;
	    /* calc error */
	    err = error[0][x] - r*rd;
	    erg = error[1][x] - g*gd;
	    erb = error[2][x] - b*bd;

	    /* distribute error */
	    /* x+1, y */
	    error[0][x+1]+=(err*3)/8;
	    error[1][x+1]+=(erg*3)/8;
	    error[2][x+1]+=(erb*3)/8;
	    /* x, y+1 */
	    error_next[0][x]+=(err*3)/8;
	    error_next[1][x]+=(erg*3)/8;
	    error_next[2][x]+=(erb*3)/8;
	    /* x+1, y+1 */
	    error_next[0][x+1]+=(err*2)/8;
	    error_next[1][x+1]+=(erg*2)/8;
	    error_next[2][x+1]+=(erb*2)/8;
	    ofs++;
	}
	/* skip to next line */
	tmp = error_next[0];
	error_next[0] = error[0];
	error[0] = tmp;
	tmp = error_next[1];
	error_next[1] = error[1];
	error[1] = tmp;
	tmp = error_next[2];
	error_next[2] = error[2];
	error[2] = tmp;
    }
    XpmFreeXpmImage(&image);
    end:
    for(i=0; i<3; i++) {
	free(color_table[i]);
	free(error[i]);
	free(error_next[i]);
    }
    return img;
}


static Pixmap 
loadXPMIconPixmap(WScreen *scr, WImage *back, int titleh, char *file)
{
    WImage *img;
    Pixmap pixmap=None;
    
    img = loadXPMIcon(scr, back, titleh, file);
    if (!img)
      return None;
    if (!DitherImage(scr, img, &pixmap)) {
	wWarning("could not convert image \"%s\" to internal format", file);
	pixmap = None;
    }
    free(img->data);
    if (img->alpha)
      free(img->alpha);
    free(img);
    return pixmap;
}


Pixmap
LoadDitheredXPM(WScreen *scr, char *file, int *w_ret, int *h_ret)
{
    WScreenDescriptor *ws = scr->rscreen;
    WImage img;
    unsigned char *data;
    Pixmap pixmap=None;
    XpmImage image;
    unsigned short *color_table[3];
    int *error[3];
    int *error_next[3], *tmp;
    int i, ofs, rd, gd, bd, x, y, r, g, b;
    register int err, erg, erb, pos;
    int transp=-1;


    if (!getXPMImage(scr, file, &image, color_table, &transp)) {
	pixmap=None;
	return None;
    }

    data = malloc(image.width*image.height*3);
    if (data==NULL) {
	wWarning("virtual memory exhausted during pixmap load. Not fatal...");
	XpmFreeXpmImage(&image);    
	for(i=0; i<3; i++) {
	    free(color_table[i]);
	}
	free(data);
	return None;
    }
    *w_ret = image.width;
    *h_ret = image.height;
    /* make error table */
    for (i=0; i<3; i++) {
	error[i] = malloc((*w_ret+2)*sizeof(int));
	error_next[i] = malloc((*w_ret+2)*sizeof(int));
	if (!error[i] || !error_next[i]) {
	    for (i=i-1;i>=0; i--) {
		if (error[i])
		  free(error[i]);
		if (error_next[i])
		  free(error_next[i]);
	    }
	    free(data);
	    wWarning("virtual memory exhausted during pixmap load. Not fatal...");
	    for(i=0; i<3; i++) {
		free(color_table[i]);
	    }
	    XpmFreeXpmImage(&image);
	    return None;
	}
    }

    /* init error table for the first line */
    i = 0;
    for (x=0; x<*w_ret; x++) {
	error[0][x] = color_table[0][image.data[i]];
	error[1][x] = color_table[1][image.data[i]];
	error[2][x] = color_table[2][image.data[i]];
	i++;
    }

    /* divide each color by these values, so they'll be mapped to
     * the available colors */
    rd = 0xffff/(ws->r_colors-1);
    gd = 0xffff/(ws->g_colors-1);
    bd = 0xffff/(ws->b_colors-1);
    /* convert and dither the image to WImage */    
    ofs = 0;
    
    for (y=0; y<*h_ret; y++) {
	unsigned int *p;
	if (y!=*h_ret-1) {
	    p = image.data + (y+1)*image.width;
	    for (x=0; x<*w_ret; x++) {
		error_next[0][x]=color_table[0][*p];
		error_next[1][x]=color_table[1][*p];
		error_next[2][x]=color_table[2][*p];
		p++;
	    }
	}
	for (x=0; x<*w_ret; x++) {
	    i = image.data[ofs];
	    pos = ofs*3;
	    /* copy image and reduce */
	    r = error[0][x]/rd;
	    g = error[1][x]/gd;
	    b = error[2][x]/bd;
	    if (r<0) r=0; else if (r>=ws->r_colors) r=ws->r_colors-1;
	    if (g<0) g=0; else if (g>=ws->g_colors) g=ws->g_colors-1;
	    if (b<0) b=0; else if (b>=ws->b_colors) b=ws->b_colors-1;
	    data[pos]=r; 
	    data[pos+1]=g;
	    data[pos+2]=b;
	    /* calc error */
	    err = error[0][x] - r*rd;
	    erg = error[1][x] - g*gd;
	    erb = error[2][x] - b*bd;

	    /* distribute error */
	    /* x+1, y */
	    error[0][x+1]+=(err*3)/8;
	    error[1][x+1]+=(erg*3)/8;
	    error[2][x+1]+=(erb*3)/8;
	    /* x, y+1 */
	    error_next[0][x]+=(err*3)/8;
	    error_next[1][x]+=(erg*3)/8;
	    error_next[2][x]+=(erb*3)/8;
	    /* x+1, y+1 */
	    error_next[0][x+1]+=(err*2)/8;
	    error_next[1][x+1]+=(erg*2)/8;
	    error_next[2][x+1]+=(erb*2)/8;
	    ofs++;
	}
	/* skip to next line */
	tmp = error_next[0];
	error_next[0] = error[0];
	error[0] = tmp;
	tmp = error_next[1];
	error_next[1] = error[1];
	error[1] = tmp;
	tmp = error_next[2];
	error_next[2] = error[2];
	error[2] = tmp;
    }
    
    img.width = *w_ret;
    img.height = *h_ret;
    img.alpha = NULL;
    img.data = data;
    if (!DitherImage(scr, &img, &pixmap)) {
	wWarning("could not convert image \"%s\" to internal format", file);
	pixmap = None;
    }
    
    XpmFreeXpmImage(&image);    
    for(i=0; i<3; i++) {
	free(color_table[i]);
	free(error[i]);
	free(error_next[i]);
    }
    free(data);
    return pixmap;
}
#endif /* XPM */
#endif /* DITHERING */


Pixmap 
LoadIcon(WScreen *scr, char *path, char *mask, int title_height)
{
#ifdef DITHERING    
    WScreenDescriptor *ws = scr->rscreen;
#endif    
    WPixmap *wpix;
    Pixmap pix=None;

#ifdef DITHERING    
    if (ws && wPreferences.use_dithering) {
	switch (IdentPixmapFile(path)) {	    
	 case PIX_XPM:
#ifdef XPM
	    pix = loadXPMIconPixmap(scr, scr->icon_back_image, title_height, 
				    path);
	    if (!pix)
	      wWarning("could not load XPM file \"%s\"", path);
#else
	    wWarning("can't load \"%s\": XPM support not compiled in", path);
#endif /* DITHERING */
	    break;
	    
	 case PIX_XBM:
	    goto no_dithering;
	    break;
	    
	 default:
	    wWarning("can't load \"%s\": unknown image format", path);
	    pix = None;
	}
    } else 
#endif /* DITHERING */    
    {
	int x, y;

	no_dithering:
	wpix = wPixmapLoadFile(scr, path, mask);
	pix = XCreatePixmap(dpy, scr->w_win, ICON_WIDTH, ICON_HEIGHT, 
			    scr->w_depth);
	XSetClipMask(dpy, scr->copy_gc, None);
	XCopyArea(dpy, scr->icon_back_pixmap, pix, scr->copy_gc, 0, 0, 
		  ICON_WIDTH, ICON_HEIGHT, 0, 0);

	x = (ICON_WIDTH - wpix->width)/2;
	y = title_height + (ICON_HEIGHT - wpix->height - title_height - 1)/2;

	XSetClipOrigin(dpy, scr->copy_gc, x, y);
	XSetClipMask(dpy, scr->copy_gc, wpix->mask);

	if (wpix->depth==1)
	  XCopyPlane(dpy, wpix->image, pix, scr->copy_gc, 0, 0, wpix->width, 
		     wpix->height, x, y, 1);
	else
	  XCopyArea(dpy, wpix->image, pix, scr->copy_gc, 0, 0, wpix->width, 
		    wpix->height, x, y);
	wPixmapDestroy(wpix);
    }
    
    if (pix && title_height>0) {
	XFillRectangle(dpy, pix, scr->icon_title_texture->normal_gc, 0, 0, 
		       ICON_WIDTH, title_height+1);
	XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		  ICON_WIDTH, 0);
	XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		  0, title_height+1);
	XDrawLine(dpy, pix, scr->icon_title_texture->dim_gc, 
		  ICON_WIDTH-1, 0, ICON_WIDTH-1, title_height+1);
    }
    return pix;
}


