/* menu.c- generic menu class, used for root menu and application menus
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "button.h"
#include "menu.h"
#include "actions.h"
#include "funcs.h"
#include "stacking.h"

/****** Global Variables ******/

extern Cursor wCursor[WCUR_LAST];

extern XContext wWinContext;

extern WPreferences wPreferences;


static void menuMouseDown(WObjDescriptor *desc, XButtonEvent *bev);
static void menuTitleMouseDown(WObjDescriptor *desc, XButtonEvent *bev);
static void menuCloseClick(WObjDescriptor *desc, XEvent *unused);

static WMethodTable wMenuMethods = {
    (WEventHandler)menuMouseDown,
      NULL,
      NULL,
      NULL
};

static WMethodTable wMenuTitlebarMethods = {
    (WEventHandler)menuTitleMouseDown,
      NULL,
      NULL,
      NULL
};


static WMethodTable wMenuButtonMethods;

void
wMenuInitialize()
{
    memcpy(&wMenuButtonMethods, &wButtonMethods, sizeof(WMethodTable));
    wMenuButtonMethods.click = menuCloseClick;
}


WMenu *
wMenuCreate(WScreen *screen, char *title, int main_menu)
{
    WMenu *menu;
    WButton *wbut;
    int title_h;
    int width, height;
    int bsize;
    
    menu = wmalloc(sizeof(WMenu));

    menu->parent = NULL;
    memset(&(menu->flags), 0, sizeof(menu->flags));
    menu->code = 0;
    menu->title = strdup(title);
#ifdef I18N
    menu->title_width = XmbTextEscapement(screen->menu_title_font.font, title,
				   strlen(title));
#else
    menu->title_width = XTextWidth(screen->menu_title_font.font, title,
				   strlen(title));
#endif
    menu->entry_no = 0;
    menu->alloced_entries = 8;
    menu->selected_entry = -1;
    menu->entries = wmalloc(sizeof(WMenuEntry)*menu->alloced_entries);
    
    title_h = MIN_TITLEFONT_HEIGHT(screen->menu_title_font.height) + 9;
    
    menu->screen_ptr = screen;

    width = menu->title_width + title_h + 7;

    height = title_h*2;
    menu->frame = wCoreCreateTopLevel(screen, 8, 2, width, height, 0, False);
    menu->frame_x = screen->app_menu_x;
    menu->frame_y = screen->app_menu_y;
    menu->frame->descriptor.parent = menu;
    menu->frame->descriptor.parent_type = WCLASS_MENU;
    menu->frame->descriptor.methods = &wMenuMethods;
    
    if (!wPreferences.obscurable_menus) {
	/* setup stacking information */
	menu->frame->stacking = wmalloc(sizeof(WStacking));
	menu->frame->stacking->above = NULL;
	menu->frame->stacking->under = NULL;
	menu->frame->stacking->child_of = NULL;
	menu->frame->stacking->window_level =
	  (main_menu ? NSMainMenuWindowLevel : NSSubmenuWindowLevel);
    }
    /* create borders */
    menu->titlebar = wCoreCreate(menu->frame, 0, 0, menu->frame->width,
				title_h);
    menu->titlebar->descriptor.parent = menu;
    menu->titlebar->descriptor.parent_type = WCLASS_MENU;
    menu->titlebar->descriptor.methods = &wMenuTitlebarMethods;

    menu->menu = wCoreCreate(menu->frame, 0, title_h, width, height-title_h);

    menu->menu->descriptor.parent = menu;
    menu->menu->descriptor.parent_type = WCLASS_MENU;
    menu->menu->descriptor.methods = &wMenuMethods;
    
    /* create close button */
    bsize = (title_h - 8);
    wbut = wButtonCreate(menu->titlebar, width - 4 - bsize, 
			 (title_h + 1 - bsize)/2, bsize, bsize);
    /* setup object descriptor */
    wbut->core->descriptor.parent = menu;
    wbut->core->descriptor.parent_type = WCLASS_MENU;
    wbut->core->descriptor.methods = &wMenuButtonMethods;
    wbut->image = screen->b_pixmaps[WBUT_CLOSE];
    menu->close_button = wbut;
    
    menu->menu_texture_data = 0;
    menu->titlebar_texture_data = 0;

    XMapSubwindows(dpy, menu->frame->window);

    if (!wPreferences.obscurable_menus)
      AddToStackList(menu->frame);

    XSync(dpy, 0);
    
    return menu;
}



static void
insertEntry(WMenu *menu, WMenuEntry *entry, int index)
{
    int i;
    
    for (i=menu->entry_no; i>index; i--) {
	menu->entries[i]->order++;
	menu->entries[i+1]=menu->entries[i];
    }
    menu->entries[index] = entry;
}


#ifdef notdef
void
wMenuInsertItem(WMenu *menu, int index, char *text, int code)
{
    WMenuEntry *entry;
  
#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddItem() \n");
	return;
    }    
#endif
    menu->flags.realized = 0;
    /* wreallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif
	tmp = wrealloc(menu->entries,
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    entry = wmalloc(sizeof(WMenuEntry));
    entry->flags.selected = 0;
    entry->flags.enabled = 1;
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = NULL;
    entry->code = code;
    entry->callback = NULL;
    if (index<0 || index>=menu->entry_no) {
	entry->order = menu->entry_no;
	menu->entries[menu->entry_no] = entry;
    } else {
	entry->order = index;
	insertEntry(menu, entry, index);
    }
    menu->entry_no++;
}
#endif /* notdef */

WMenuEntry * 
wMenuInsertCallback(WMenu *menu, int index, char *text,
		    void (*callback)(WMenu *menu, WMenuEntry *entry), 
		    void *clientdata)
{
    WMenuEntry *entry;

#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddCallback() \n");
	return NULL;
    }    
#endif

    menu->flags.realized = 0;
    /* reallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif	
	tmp = wrealloc(menu->entries,
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return NULL;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    entry = wmalloc(sizeof(WMenuEntry));
    entry->code = 0;
    entry->flags.selected = 0;
    entry->flags.enabled = 1;    
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = NULL;
    entry->clientdata = clientdata;
    entry->free_cdata = NULL;
    entry->callback = callback;
    if (index<0 || index>=menu->entry_no) {
	entry->order = menu->entry_no;
	menu->entries[menu->entry_no] = entry;
    } else {
	entry->order = index;
	insertEntry(menu, entry, index);
    }
    menu->entry_no++;
    return entry;
}


void
wMenuEntrySetCascade(WMenu *menu, WMenuEntry *entry, WMenu *cascade)
{
    menu->flags.realized = 0;
    entry->cascade = cascade;
    cascade->parent = menu;
}


#ifdef notdef
WMenuEntry*
wMenuInsertCascade(WMenu *menu, int index, char *text, WMenu *cascade)
{
    WMenuEntry *entry;

#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddCascade() \n");
	return NULL;
    }
#endif
    
    cascade->parent = menu;
    menu->flags.realized = 0;
    /* wreallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif	
	tmp = wrealloc(menu->entries, 
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return NULL;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    entry = wmalloc(sizeof(WMenuEntry));
    entry->flags.selected = 0;
    entry->flags.enabled = 1;
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = cascade;
    entry->clientdata = NULL;
    entry->callback = NULL;
    if (index<0 || index>=menu->entry_no) {
	entry->order = menu->entry_no;
	menu->entries[menu->entry_no] = entry;
    } else {
	entry->order = menu->entry_no;
	insertEntry(menu, entry, index);
    }
    menu->entry_no++;
    return entry;
}
#endif /* notdef */

void
wMenuRemoveItem(WMenu *menu, int index)
{
    int i;
    
    if (index>=menu->entry_no) return;
    wfree(menu->entries[index]);
    for (i=index; i<menu->entry_no-1; i++) {
	menu->entries[i+1]->order--;
	menu->entries[i]=menu->entries[i+1];
    }
    menu->entry_no--;
}


void 
wMenuRealize(WMenu *menu)
{
    int i;
    int width, mwidth, extra_width, bsize;
    int theight, eheight;
    WScreen *scr = menu->frame->screen_ptr;

#ifdef DEBUG
    if (!menu) {
	printf("Trying to realize NULL menu \n");
	return;
    }    
#endif
    menu->flags.realized = 1;
#ifdef I18N
    menu->title_width = XmbTextEscapement(scr->menu_title_font.font, 
					  menu->title, strlen(menu->title));
#else
    menu->title_width = XTextWidth(scr->menu_title_font.font, menu->title,
				   strlen(menu->title));
#endif
    mwidth = menu->title_width;
    theight = MIN_TITLEFONT_HEIGHT(scr->menu_title_font.height) + 9;
    eheight = scr->menu_entry_font.height + 6;
    menu->entry_height = eheight;
#ifdef I18N
    extra_width = XmbTextEscapement(scr->menu_entry_font.font, "F8", 2) + 16;
#else
    extra_width = XTextWidth(scr->menu_entry_font.font, "F8", 2) + 16;
#endif
    for (i=0; i<menu->entry_no; i++) {
	char *text, *rtext;

	text = menu->entries[i]->text;
	rtext = menu->entries[i]->rtext;
#ifdef I18N
	width = XmbTextEscapement(scr->menu_entry_font.font, text, 
				  strlen(text));
	if (rtext)
	  width += XmbTextEscapement(scr->menu_entry_font.font, rtext, 
				     strlen(rtext));
#else
	width = XTextWidth(scr->menu_entry_font.font, text, strlen(text));
	if (rtext)
	  width += XTextWidth(scr->menu_entry_font.font, rtext, strlen(rtext));
#endif
	menu->entries[i]->text_width = width;
	if (width > mwidth)
	  mwidth = width;
    }
    wCoreConfigure(menu->frame, menu->frame_x, menu->frame_y, 
		   mwidth + extra_width, theight + menu->entry_no * eheight);
    wCoreConfigure(menu->titlebar, 0, 0, mwidth + extra_width, theight);
    wCoreConfigure(menu->menu, 0, theight, mwidth + extra_width, 
		   menu->entry_no * eheight);
    
    bsize = menu->close_button->core->width;
    XMoveWindow(dpy,menu->close_button->core->window, mwidth 
		+ extra_width - 4 - bsize, (theight + 1 -bsize)/2);

    
    wTextureFreeData(menu->screen_ptr->menu_title_texture,
		     &menu->titlebar_texture_data);

    wTextureFreeData(menu->screen_ptr->menu_item_texture,
		     &menu->menu_texture_data);

    /* setup background texture */
    switch (scr->menu_item_texture->any.type) {
     case WTEX_GRADIENT:
	wTextureRender(scr, scr->menu_item_texture, 
		       &menu->menu_texture_data, menu->menu->width,
		       menu->entry_height);
	XSetWindowBackgroundPixmap(dpy, menu->menu->window, 
				   menu->menu_texture_data);
	break;
     default:
	XSetWindowBackground(dpy, menu->menu->window,
			     scr->menu_item_texture->any.color.pixel);
    }
    if (menu->flags.mapped)
      wMenuPaint(menu);
}


void
wMenuDestroy(WMenu *menu, int recurse)
{
    int i;
    WCoreWindow *frame;
    
#ifdef DEBUG
    if (!menu) {
	printf("Trying to destroy NULL menu \n");
	return;
    }
#endif
    frame = menu->frame;
    
    if (!wPreferences.obscurable_menus) {
#ifdef BUGGYSTACKING
	/* remove from stacking list */
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
	else /* this was the first window on the list */
	  frame->screen_ptr->stacking_list = frame->stacking->under;
	menu->screen_ptr->window_count--;
#else /* BUGGYSTACKING */
	RemoveFromStackList(menu->frame);
#endif /* !BUGGYSTACKING */
    }
    wTextureFreeData(menu->screen_ptr->menu_title_texture,
		     &menu->titlebar_texture_data);

    wTextureFreeData(menu->screen_ptr->menu_item_texture,
		     &menu->menu_texture_data);

    /* destroy items */
    for (i=0; i<menu->entry_no; i++) {
	wfree(menu->entries[i]->text);
	if (menu->entries[i]->rtext)
	    wfree(menu->entries[i]->rtext);
	if (recurse && menu->entries[i]->cascade!=NULL)
	  wMenuDestroy(menu->entries[i]->cascade, recurse);

	
	if (menu->entries[i]->free_cdata) {
	    (*menu->entries[i]->free_cdata)(menu->entries[i]->clientdata);
	}
	wfree(menu->entries[i]);
    }
    if (menu->entries)
	wfree(menu->entries);
    if (menu->title)
	wfree(menu->title);
    wButtonDestroy(menu->close_button);
    wCoreDestroy(menu->titlebar);
    wCoreDestroy(menu->menu);
    wCoreDestroy(menu->frame);
    wfree(menu);
}


static void
drawFrame(WScreen *scr, Window win, int y, int w, int h)
{
    XSegment segs[4];

    segs[0].x1 = 1;
    segs[0].y1 = y;
    segs[0].x2 = w-2;
    segs[0].y2 = y;
    segs[1].x1 = 1;
    segs[1].y1 = y;
    segs[1].x2 = 1;
    segs[1].y2 = y + h - 2;
    XDrawSegments(dpy, win, scr->menu_item_auxtexture->light_gc, segs, 2);
    
    segs[0].x1 = 1;
    segs[0].x2 = w-1;
    segs[0].y1 = segs[0].y2 = y + h-1;
    segs[1].x1 = segs[1].x2 = w-1;
    segs[1].y1 = y;
    segs[1].y2 = y + h-1;
    XDrawSegments(dpy, win, scr->menu_item_auxtexture->dark_gc, segs, 2);
    
    segs[0].x1 = 2;
    segs[0].y1 = segs[0].y2 = y + h-2;
    segs[0].x2 = w-2;
    segs[1].x1 = segs[1].x2 = w-2;
    segs[1].y1 = y + 1;
    segs[1].y2 = y + h-2;
    segs[2].x1 = segs[2].x2 = 0;
    segs[2].y1 = y;
    segs[2].y2 = y + h-1;
    XDrawSegments(dpy, win, scr->menu_item_auxtexture->dim_gc, segs, 3);
}


static void
paintEntry(WMenu *menu, int index, int selected)
{
    int y, w, h, tw;
    GC light, dim, dark, textGC;
    WScreen *scr=menu->frame->screen_ptr;
    Window win = menu->menu->window;
    WMenuEntry *entry=menu->entries[index];

    if (!menu->flags.realized) return;
    h = menu->entry_height;
    w = menu->menu->width;
    y = index * h;

    light = scr->menu_item_auxtexture->light_gc;
    dim = scr->menu_item_auxtexture->dim_gc;
    dark = scr->menu_item_auxtexture->dark_gc;

    /* paint background */
    if (selected) {
	XSetForeground(dpy, scr->select_menu_gc, scr->white_pixel);
	XFillRectangle(dpy, win, scr->select_menu_gc, 0, y, w, h);
	drawFrame(scr, win, y, w, h);
    } else {
	XClearArea(dpy, win, 0, y, w, h, False);

	switch (scr->menu_item_texture->any.type) {
	 case WTEX_SOLID:
	    /* draw the frame */
	    drawFrame(scr, win, y, w, h);
	    break;
	}
    }

    if (entry->cascade) {
	/* draw the cascade indicator */
	XDrawLine(dpy,win,dim, w-11, y+6, w-6, y+h/2-1);
	XDrawLine(dpy,win,light, w-11, y+h-8, w-6, y+h/2-1);
	XDrawLine(dpy,win,dark, w-12, y+6, w-12, y+h-8);
    }

    /* draw text */
    if (selected) {
	textGC = scr->select_menu_gc;
	XSetForeground(dpy, textGC, scr->black_pixel);
    } else if (entry->flags.enabled) {
	textGC = scr->menu_entry_gc;
    } else {
	textGC = scr->disabled_menu_entry_gc;
    }
#ifdef I18N
    XmbDrawString(dpy, win, scr->menu_entry_font.font, 
		  textGC, 6, 2+y+scr->menu_entry_font.y, entry->text, 
		strlen(entry->text));
#else
    XDrawString(dpy, win, textGC, 6, 2+y+scr->menu_entry_font.y, entry->text, 
		strlen(entry->text));
#endif
    if (entry->rtext && entry->cascade==NULL) {
#ifdef I18N
	tw = XmbTextEscapement(scr->menu_title_font.font, entry->rtext,
			       strlen(entry->rtext));
	XmbDrawString(dpy, win, scr->menu_entry_font.font, 
		      textGC, w-6-tw, 2+y+scr->menu_entry_font.y, entry->rtext,
		      strlen(entry->rtext));
#else
	tw = XTextWidth(scr->menu_title_font.font, entry->rtext, 
			strlen(entry->rtext));
	XDrawString(dpy, win, textGC, w-6-tw,2+y+scr->menu_entry_font.y, 
		    entry->rtext, strlen(entry->rtext));
#endif
    }
}

void
wMenuMapAt(WMenu *menu, int x, int y)
{
    int scr_width = menu->screen_ptr->scr_width;
    int scr_height = menu->screen_ptr->scr_height;

    if (!menu->flags.realized) {
	menu->flags.realized=1;
	wMenuRealize(menu);
    }
    if (x<0) x = 0;
    if (y<0) y = 0;
    if (x+menu->frame->width > scr_width)
      x = scr_width - menu->frame->width;
    if (y+menu->frame->height > scr_height)
      y = scr_height - menu->frame->height;
	
    XMoveWindow(dpy, menu->frame->window, x, y);
    menu->frame_x = x;
    menu->frame_y = y;
    XMapWindow(dpy, menu->frame->window);
    if (!wPreferences.obscurable_menus)
      wRaiseFrame(menu->frame);
    else
      XRaiseWindow(dpy, menu->frame->window);
    menu->flags.mapped = 1;
}


void
wMenuMap(WMenu *menu)
{
    if (!menu->flags.realized) {
	menu->flags.realized=1;
	wMenuRealize(menu);
    }
    if (menu->flags.app_menu) {
	menu->frame_x = menu->frame->screen_ptr->app_menu_x;
	menu->frame_y = menu->frame->screen_ptr->app_menu_y;
	XMoveWindow(dpy, menu->frame->window, menu->frame_x, menu->frame_y);
    }
    XMapWindow(dpy, menu->frame->window);
    if (!wPreferences.obscurable_menus)
      wRaiseFrame(menu->frame);
    else
      XRaiseWindow(dpy, menu->frame->window);
    menu->flags.mapped = 1;
}


void
wMenuUnmap(WMenu *menu)
{
    int i;
    
    XUnmapWindow(dpy, menu->frame->window);
    XUnmapWindow(dpy, menu->close_button->core->window);
    menu->flags.buttoned = 0;
    menu->flags.mapped = 0;
    menu->flags.open_to_left = 0;

    for (i=0; i<menu->entry_no; i++) {
	if (menu->entries[i]->cascade!=NULL 
	    && menu->entries[i]->cascade->flags.mapped 
	    && !menu->entries[i]->cascade->flags.buttoned)
	  wMenuUnmap(menu->entries[i]->cascade);
	menu->entries[i]->flags.selected=0;
    }
    menu->selected_entry = -1;
} 


void 
wMenuPaint(WMenu *menu)
{
    int i;
    WScreen *scr = menu->frame->screen_ptr;
    
    /* paint titlebar */
    wTexturePaint(scr->menu_title_texture, &menu->titlebar_texture_data,
		  menu->titlebar, False);
#ifdef I18N
    XmbDrawString(dpy, menu->titlebar->window, scr->menu_title_font.font, 
		  scr->menu_title_gc, 6, 
		  (menu->titlebar->height-scr->menu_title_font.height)/2
		  + scr->menu_title_font.y, menu->title, strlen(menu->title));
#else
    XDrawString(dpy, menu->titlebar->window, scr->menu_title_gc, 6,
		(menu->titlebar->height-scr->menu_title_font.height)/2
		+ scr->menu_title_font.y, menu->title, strlen(menu->title));
#endif
    /* paint entries */
    for(i=0; i<menu->entry_no; i++) {
	paintEntry(menu, i, i==menu->selected_entry);
    }
    if (menu->flags.buttoned)
      wButtonPaint(menu->close_button, False);
}

void
wMenuSetEnabled(WMenu *menu, int index, int enable)
{
    if (index>=menu->entry_no) return;
    menu->entries[index]->flags.enabled=enable;
    paintEntry(menu, index, index==menu->selected_entry);
}


/* ====================================================================== */


static void
selectEntry(WMenu *menu, int entry_no)
{
    WMenuEntry *entry;
    WMenu *submenu;

    if (menu->selected_entry!=entry_no) {
	/* unselect previous entry */
	if (menu->selected_entry>=0) {
	    entry = menu->entries[menu->selected_entry];
	    paintEntry(menu, menu->selected_entry, False);
	    entry = menu->entries[menu->selected_entry];
	    entry->flags.selected=0;
	    /* unmap cascade */
	    if (entry->cascade && !entry->cascade->flags.buttoned) {
		wMenuUnmap(entry->cascade);
	    }
	}
	if (entry_no<0) {
	    menu->selected_entry = -1;
	    return;
	}
	entry = menu->entries[entry_no];
	if (entry->flags.enabled) {
	    submenu = entry->cascade;
	    if (submenu!=NULL) {
		/* map cascade */
		if (!submenu->flags.mapped) {
		    int x;
		    
		    if (!submenu->flags.realized)
		      wMenuRealize(submenu);
		    if (menu->flags.open_to_left)
		      submenu->flags.open_to_left = 1;
		    if (submenu->flags.open_to_left) {
			x = menu->frame_x - submenu->frame->width;
			if (x<0) {
			    x = 0;
			    submenu->flags.open_to_left = 0;
			}
		    } else {
			x = menu->frame_x + menu->frame->width;
			if (x + submenu->frame->width
			    > menu->screen_ptr->scr_width) {
			    x = menu->frame_x - submenu->frame->width;
			    submenu->flags.open_to_left = 1;
			}
		    }
		    if (wPreferences.align_menus) {
			wMenuMapAt(entry->cascade, x, menu->frame_y);
		    } else {
			int y;
			
			y = menu->frame_y + menu->entry_height*entry_no;
			wMenuMapAt(entry->cascade, x, y);	 
		    }
		} else return;
	    }
	    menu->selected_entry = entry_no;
	    entry->flags.selected=1;
	    paintEntry(menu, entry_no, True);
	} else {
	    menu->selected_entry = -1;
	}
    }    
}

static WMenu*
findMenu(WScreen *scr, int *x_ret, int *y_ret)
{
    WMenu *menu;
    WObjDescriptor *desc;
    Window root_ret, win, junk_win;
    int x, y, wx, wy;
    unsigned int mask;
    
    XQueryPointer(dpy, scr->root_win, &root_ret, &win, &x, &y, &wx, &wy,
		  &mask);
    if (win==None) return NULL;
    if (XFindContext(dpy, win, wWinContext, (XPointer *)&desc)==XCNOENT)
      return NULL;
    if (desc->parent_type==WCLASS_MENU) {
	menu = desc->parent;
	XTranslateCoordinates(dpy, root_ret, menu->menu->window, wx, wy, 
			      x_ret, y_ret, &junk_win);
	return menu;
    }
    
    return NULL;
}


static void
closeCascade(WMenu *menu)
{
    WMenu *parent=menu->parent;

    if (!menu->flags.buttoned && !menu->flags.app_menu) {
	selectEntry(menu, -1);
	XSync(dpy, 0);
#if (MENU_BLINK_DELAY > 2)
	usleep(MENU_BLINK_DELAY/2);
#endif
	wMenuUnmap(menu);
	while (parent!=NULL && !parent->flags.buttoned) {
	    selectEntry(parent, -1);
	    wMenuUnmap(parent);
	    parent = parent->parent;
	}
	if (parent)
	  selectEntry(parent, -1);
    }
}

#define getEntryAt(menu, x, y)   ((y)<0 ? -1 : (y)/(menu->entry_height))

static void  
menuMouseDown(WObjDescriptor *desc, XButtonEvent *bev)
{
    WMenu *menu=desc->parent;
    WMenu *smenu;
    WScreen *scr=menu->frame->screen_ptr;    
    WMenuEntry *entry=NULL;
    XEvent event;
    int close_on_exit;
    int done=0;
    int entry_no;
    int x, y;

    if (!wPreferences.obscurable_menus)
      wRaiseFrame(menu->frame);
    else
      XRaiseWindow(dpy, menu->frame->window);

    /* if this event was redispatched in handleEvent() */
    close_on_exit = bev->send_event;
    /*
    if (bev->button==Button3) {
	if (!menu->flags.buttoned) {
	    if (menu->parent)
	      selectEntry(menu->parent, -1);
	    wMenuUnmap(menu);	    
	}
	return;
    }
     */
    menu = findMenu(scr, &x, &y);
    if (!menu) {
	return;
    }
    entry_no = getEntryAt(menu, x, y);
    if (entry_no>=0) {
	entry = menu->entries[entry_no];
	if (entry->flags.enabled && entry->cascade!=NULL) {
	    /* map cascade */
	    if (!entry->cascade->flags.mapped) {
		selectEntry(menu, entry_no);
	    } else if (!entry->cascade->flags.buttoned) {
		selectEntry(menu, -1);
	    }
	} else {
	    selectEntry(menu, entry_no);
	}
    }
    while (!done) {
	int x, y;
	XMaskEvent(dpy, ExposureMask|ButtonMotionMask|ButtonReleaseMask,
		   &event);
	switch (event.type) {
	 case MotionNotify:
	    smenu = findMenu(scr, &x, &y);
	    if (smenu == NULL) {
		if (menu==NULL
		    || (menu->selected_entry>=0
			&& menu->entries[menu->selected_entry]->cascade))
		    break;
		selectEntry(menu, -1);
		menu = smenu;
		break;
	    } else if (menu && menu!=smenu &&
		       !(menu->selected_entry>=0
			 && menu->entries[menu->selected_entry]->cascade)) {
		selectEntry(menu, -1);
	    }
	    menu = smenu;
	    entry_no = getEntryAt(menu, x, y);
	    selectEntry(menu, entry_no);
	    break;

	 case ButtonRelease:
	    done=1;
	    break;

	 case Expose: 
	    DispatchEvent(&event);
	    break;
 	}
    }
    if (menu && menu->selected_entry>=0) {
	entry = menu->entries[menu->selected_entry];
	/* erase menu selection */
	if (entry->callback!=NULL) {
	    /* unmap the menu, it's parents and call the callback */
	    if (!menu->flags.buttoned && !menu->flags.app_menu) {
#if (MENU_BLINK_DELAY > 0)
		int sel = menu->selected_entry;

		paintEntry(menu, sel, False);
		XSync(dpy, 0);
		usleep(MENU_BLINK_DELAY);
		paintEntry(menu, sel, True);
		XSync(dpy, 0);
		usleep(MENU_BLINK_DELAY);
		paintEntry(menu, sel, False);
		XSync(dpy, 0);
		usleep(MENU_BLINK_DELAY);
		paintEntry(menu, sel, True);
		XSync(dpy, 0);
		usleep(MENU_BLINK_DELAY);
#endif
		closeCascade(menu);
	    } else {
#if (MENU_BLINK_DELAY > 2)
		usleep(MENU_BLINK_DELAY/2);
#endif
		selectEntry(menu, -1);
	    }
	    (*entry->callback)(menu, entry);

	    /* If the user double clicks an entry, the entry will
	     * be executed twice, which is not good for things like
	     * the root menu. So, ignore any clicks that were generated 
	     * while the entry was being executed */
	    while (XCheckTypedWindowEvent(dpy, menu->menu->window, 
					  ButtonPress, &event));
	} else {
	    if (close_on_exit) {
		closeCascade(desc->parent);
	    } 

	}
    } else {
	if (close_on_exit && (!menu || entry_no>=0)) {
	    closeCascade(desc->parent);
	}
    }
}


static void
moveMenu(WMenu *menu, int x, int y)
{
    WMenu *submenu;
    
    if (!menu) return;
    menu->frame_x = x;
    menu->frame_y = y;
    XMoveWindow(dpy, menu->frame->window, x, y);
    if (menu->selected_entry>=0) {
	submenu = menu->entries[menu->selected_entry]->cascade;
	if (submenu && submenu->flags.mapped && !submenu->flags.buttoned)
	  if (wPreferences.align_menus) {
	      moveMenu(submenu, x + menu->frame->width, y);
	  } else {
	      moveMenu(submenu, x+ menu->frame->width,
		       y + submenu->entry_height*menu->selected_entry);
	  }
    }
}

/*
 *---------------------------------------------------------------------- 
 * menuTitleMouseDown--
 * 	Handle ButtonPress events on the menu titlebar.
 * 
 * Side effects:
 * 	The pointer is grabbed until the button is released, but 
 * events are dispatched properly.
 * 
 * Notes:
 * 	Moves the menu according to user action and enables the close
 * button (stick) if the menu is moved.
 *---------------------------------------------------------------------- 
 */
static void 
menuTitleMouseDown(WObjDescriptor *desc, XButtonEvent *bev)
{
    WMenu *menu=desc->parent;    
    XEvent event;
    int x=menu->frame_x, y=menu->frame_y;
    int dx=bev->x_root, dy=bev->y_root;
    int i;
    
#ifdef DEBUG
    puts("Moving menu");
#endif
    if (bev->button != Button1 && bev->button != Button2)
      return;
    if (XGrabPointer(dpy, menu->titlebar->window, False, 
		     ButtonMotionMask|ButtonReleaseMask,
		     GrabModeAsync, GrabModeAsync, None, 
		     wCursor[WCUR_MOVE], CurrentTime)!=GrabSuccess) {
	wWarning("pointer grab failed for menu move");
	return;
    }
    if (bev->state & Mod1Mask) {
	if (!wPreferences.obscurable_menus)
	  wLowerFrame(menu->frame);
	else
	  XLowerWindow(dpy, menu->frame->window);
    } else {
	if (!wPreferences.obscurable_menus)
	  wRaiseFrame(menu->frame);
	else
	  XRaiseWindow(dpy, menu->frame->window);
    }
    /* tear off the menu if it's a root menu or a cascade 
     application menu */
    if (!menu->flags.buttoned && (!menu->flags.app_menu||menu->parent!=NULL)) {
	XMapWindow(dpy, menu->close_button->core->window);
	menu->flags.buttoned=1;
	if (menu->parent) {
	    /* turn off selected menu entry in parent menu */
	    selectEntry(menu->parent, -1);
	    /* disable submenu entry in parent */
	    for (i=0; i<menu->parent->entry_no; i++) {
		if (menu->parent->entries[i]->cascade==menu) {
		    wMenuSetEnabled(menu->parent, i, False);
		    break;
		}
	    }
	}
    }    
    while(1) {
	XMaskEvent(dpy, ButtonMotionMask|ButtonReleaseMask|ExposureMask,
		   &event);
	switch (event.type) {
	 case MotionNotify:
	    x += event.xmotion.x_root - dx;
	    y += event.xmotion.y_root - dy;
	    dx = event.xmotion.x_root;
	    dy = event.xmotion.y_root;
	    moveMenu(menu, x, y);
	    break;

	 case ButtonRelease:
	    if (menu->flags.app_menu) {
		menu->frame->screen_ptr->app_menu_x = x;
		menu->frame->screen_ptr->app_menu_y = y;
	    }
#ifdef DEBUG
	    puts("End menu move");
#endif
	    XUngrabPointer(dpy, CurrentTime);
	    return;
	    
	 default:
	    DispatchEvent(&event);
	}
    }
}

/*
 *----------------------------------------------------------------------
 * menuCloseClick--
 * 	Handles mouse click on the close butoon of menus. The menu is
 * closed when the button is clicked. 
 * 
 * Side effects:
 * 	None
 *---------------------------------------------------------------------- 
 */
static void 
menuCloseClick(WObjDescriptor *desc, XEvent *unused)
{
    WMenu *menu=desc->parent;
    WMenu *parent=menu->parent;
    int i;
    
    if (parent) {
	for (i=0; i<parent->entry_no; i++) {
	    if (parent->entries[i]->cascade==menu) {
		wMenuSetEnabled(parent, i, True);
		break;
	    }
	}
    }
    wMenuUnmap(menu);
}



