/* 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "version.h"
#include "screen.h"
#include "wcore.h"
#include "window.h"

extern *DisplayName;

#ifdef USECPP
static void
putdef(char *line, char *name, char *value)
{
    if (!value) {
	wWarning("could not define value for %s for cpp", name);
	return;
    }
    strcat(line, name);
    strcat(line, value);
}



static void
putidef(char *line, char *name, int value)
{
    char tmp[64];
    sprintf(tmp, "%i", value);
    strcat(line, name);
    strcat(line, tmp);
}


static char*
username()
{
    char *tmp;
    
    tmp = getlogin();
    if (!tmp) {
	struct passwd *user;

	user = getpwuid(getuid());
	if (!user) {
	    wSysError("could not get password entry for UID %i", getuid());
	    return NULL;
	}
	if (!user->pw_name) {
	    return NULL;
	} else {
	    return user->pw_name;
	}
    }
    return tmp;
}
       
char *
MakeCPPArgs(char *path)
{
    int i;
    char buffer[MAXLINE], *buf, *line;
    char version[64];
    Visual *visual;
    
    line = wmalloc(MAXLINE);
    *line = 0;
    i=1;
    if (gethostname(buffer, MAXLINE)==0) {
	putdef(line, " -DHOST=", buffer);
    }
    buf = username();
    if (buf)
      putdef(line, " -DUSER=", buf);
    putidef(line, " -DUID=", getuid());
    buf = XDisplayName(DisplayString(dpy));
    putdef(line, " -DDISPLAY=", buf);
    putidef(line, " -DMAJ_VERSION=", MAJ_VERSION);
    putidef(line, " -DMIN_VERSION=", MIN_VERSION);
    sprintf(version, "%d.%d.%d", MAJ_VERSION, MIN_VERSION, SUBMIN_VERSION);
    putdef(line, " -DWM_VERSION=", version);
    
    visual = DefaultVisual(dpy, DefaultScreen(dpy));
    putidef(line, " -DVISUAL=", visual->class);
    
    putidef(line, " -DDEPTH=", DefaultDepth(dpy, DefaultScreen(dpy)));

    putidef(line, " -DSCR_WIDTH=", WidthOfScreen(DefaultScreenOfDisplay(dpy)));
    putidef(line, " -DSCR_HEIGHT=", 
	    HeightOfScreen(DefaultScreenOfDisplay(dpy)));

    strcpy(buffer, path);    
    buf = strrchr(buffer, '/');
    if (buf) *buf = 0; /* trunc filename */
    putdef(line, " -I", buffer);
#ifdef XPM
    strcat(line, " -DXPM");
#endif
#ifdef DITHERING
    strcat(line, " -DDITHERING");
#endif
    
#undef arg
#ifdef DEBUG
    puts("CPP ARGS");
    puts(line);
#endif
    return line;
}
#endif /* USECPP */


WWindow*
NextFocusWindow(WScreen *scr)
{
    WWindow *tmp, *wwin, *closest, *min;
    Window d;
    
    wwin = scr->focused_window;
    tmp = wwin->prev;
    closest = NULL;
    min = wwin;
    d = 0xffffffff;
    while (tmp) {
	if ((tmp->flags.mapped || tmp->flags.shaded) 
	    && tmp->window_flags.focusable) {
	    if (min->client_win > tmp->client_win)
	      min = tmp;
	    if (tmp->client_win > wwin->client_win
		&& (!closest
		    || (tmp->client_win - wwin->client_win) < d)) {
		closest = tmp;
		d = tmp->client_win - wwin->client_win;
	    }
	}
	tmp = tmp->prev;
    }
    if (!closest||closest==wwin)
      return min;
    return closest;
}


WWindow*
PrevFocusWindow(WScreen *scr)
{
    WWindow *tmp, *wwin, *closest, *max;
    Window d;
    
    wwin = scr->focused_window;
    tmp = wwin->prev;
    closest = NULL;
    max = wwin;
    d = 0xffffffff;
    while (tmp) {
	if ((tmp->flags.mapped || tmp->flags.shaded)
	    && tmp->window_flags.focusable) {
	    if (max->client_win < tmp->client_win)
	      max = tmp;
	    if (tmp->client_win < wwin->client_win
		&& (!closest
		    || (wwin->client_win - tmp->client_win) < d)) {
		closest = tmp;
		d = wwin->client_win - tmp->client_win;
	    }
	}
	tmp = tmp->prev;
    }
    if (!closest||closest==wwin)
      return max;
    return closest;
}


void
SlideWindow(Window win, int from_x, int from_y, int to_x, int to_y,
	    int steps, int delay)
{
    float dx, dy, x=from_x, y=from_y;

    dx = (float)(to_x-from_x) / (float)steps;
    dy = (float)(to_y-from_y) / (float)steps;
    if (dx == 0.0 && to_x != from_x) x = to_x;
    if (dy == 0.0 && to_y != from_y) y = to_y;
    while ((int)x != to_x || (int)y != to_y) {
	x += dx;
	y += dy;
	if ((dx<0 && (int)x < to_x)
	    || (dx>0 && (int)x > to_x))
	  x = (float)to_x;
	if ((dy<0 && (int)y < to_y)
	    || (dy>0 && (int)y > to_y))
	  y = (float)to_y;
	XMoveWindow(dpy, win, (int)x, (int)y);
	XSync(dpy, 0);
	if (delay>0)
	  usleep(delay);
    }
    XMoveWindow(dpy, win, to_x, to_y);
}
