/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the wfree Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the wfree Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef XPM
#include <X11/xpm.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "raster.h"
#include "screen.h"
#include "pixmap.h"


extern WPreferences wPreferences;


#ifdef DITHERING 
static int
allocatePseudoColor(WScreen *scr, WScreenDescriptor *ws, 
		    int *rw, int *gw, int *bw)
{
    XColor *colors;
    XColor avcolors[256];
    int avncolors;
    int i, ncolors, r, g, b;
    int retries;

    ws->colors = NULL;
    if (*rw<=1 || *gw<=1 || *bw<=1)
      return 0;
    ncolors = *rw * *gw * *bw;
    colors = wmalloc(sizeof(XColor)*ncolors);
    
    i=0;
    for (r=0; r<*rw; r++) {
	for (g=0; g<*gw; g++) {
	    for (b=0; b<*bw; b++) {
		colors[i].red=(r*0xffff) / (*rw-1);
		colors[i].green=(g*0xffff) / (*gw-1);
		colors[i].blue=(b*0xffff) / (*bw-1);
		colors[i].flags = DoRed|DoGreen|DoBlue;
		i++;
	    }
	}
    }
    /* try to allocate the colors */
    for (i=0; i<ncolors; i++) {
	if (!XAllocColor(dpy, scr->w_colormap, &(colors[i]))) {
	    colors[i].flags = 0; /* failed */
	} else {
	    colors[i].flags = DoRed|DoGreen|DoBlue;	    
	}
    }
    /* try to allocate close values for the colors that couldn't 
     * be allocated before */
    avncolors = (1<<scr->w_depth>256 ? 256 : 1<<scr->w_depth);
    for (i=0; i<avncolors; i++) avcolors[i].pixel = i;

    XQueryColors(dpy, scr->w_colormap, avcolors, avncolors);

    for (i=0; i<ncolors; i++) {
	if (colors[i].flags==0) {
	    int j;
	    unsigned long cdiff=0xffffffff, diff;
	    unsigned long closest=0;
	    
	    retries = 2;
	    
	    while (retries--) {
		/* find closest color */
		for (j=0; j<avncolors; j++) {
		    r = (colors[i].red - avcolors[i].red)>>8;
		    g = (colors[i].green - avcolors[i].green)>>8;
		    b = (colors[i].blue - avcolors[i].blue)>>8;
		    diff = r*r + g*g + b*b;
		    if (diff<cdiff) {
			cdiff = diff;
			closest = j;
		    }
		}
		/* allocate closest color found */
		colors[i].red = avcolors[closest].red;
		colors[i].green = avcolors[closest].green;
		colors[i].blue = avcolors[closest].blue;
		if (XAllocColor(dpy, scr->w_colormap, &colors[i])) {
		    colors[i].flags = DoRed|DoGreen|DoBlue;
		    break; /* succeeded, don't need to retry */
		}
#ifdef DEBUG
		printf("close color allocation failed. Retrying...\n");
#endif
	    }
	}
    }
    ws->colors = colors;
    return 1;
}


static inline int 
get_shifts(unsigned long mask)
{
    int i=0;
    
    while (mask) {
	mask>>=1;
	i++;
    }
    return i-1;
}

/*
 *---------------------------------------------------------------------- 
 * wInitRaster--
 * 	Initializes a screen descriptor and colormaps (if needed),
 * allocating colors and other resources.
 * 	r_weight, g_weight, b_weight specifies how many colors
 * to use for each component on PseudoColor displays.
 * 
 * Returns:
 * 	NULL on error
 * 	The screen descriptor
 * 
 * Notes:
 * 	If color allocation fails for PseudoColor displays, it will
 * retry with less colors than requested.
 * 
 *---------------------------------------------------------------------- 
 */
WScreenDescriptor *
wInitRaster(WScreen *scr, int r_weight, int g_weight, int b_weight)
{
    WScreenDescriptor *ws;

    ws = wmalloc(sizeof(WScreenDescriptor));

    ws->flags.dither = wPreferences.use_dithering;
    ws->flags.mono = (scr->w_depth==1);
    
    if (scr->w_visual->class==TrueColor) {
	/* calculate component shift values for color composition */
	ws->r_shift = 15 - get_shifts(scr->w_visual->red_mask);
	ws->g_shift = 15 - get_shifts(scr->w_visual->green_mask);
	ws->b_shift = 15 - get_shifts(scr->w_visual->blue_mask);
	ws->r_colors = scr->w_visual->red_mask;
	while ((ws->r_colors&1)==0) ws->r_colors>>=1;
	ws->g_colors = scr->w_visual->green_mask;
	while ((ws->g_colors&1)==0) ws->g_colors>>=1;
	ws->b_colors = scr->w_visual->blue_mask;
	while ((ws->b_colors&1)==0) ws->b_colors>>=1;

	ws->flags.true_color = 1;
    } else if (ws->flags.dither) {
	ws->flags.true_color = 0;
	
	if (scr->w_visual->class==PseudoColor) {
	    if (r_weight*g_weight*b_weight>scr->w_visual->map_entries) {
		wWarning("invalid dither colormap size specification. disabling");
		ws->flags.dither = 0;
	    }
	    if (allocatePseudoColor(scr, ws, &r_weight, &g_weight, &b_weight)) {
		ws->r_colors = r_weight;
		ws->g_colors = g_weight;
		ws->b_colors = b_weight;
		ws->ncolors = r_weight*g_weight*b_weight;
	    } else {
		fprintf(stderr, "Could not build PseudoColor color table\n");
		ws->flags.mono = 1;
	    }
	} /* TODO: other visual types */
    }
    return ws;
}


static unsigned long
getTrueColor(WScreen *scr, XColor *color)
{
    int r, g, b, rs, gs, bs;

    rs = scr->rscreen->r_shift;
    gs = scr->rscreen->g_shift;
    bs = scr->rscreen->b_shift;
    r = (rs < 0 ? color->red << (-rs) : color->red >> rs);

    g = (gs < 0 ? color->green << (-gs) : color->green >> gs);

    b = (bs < 0 ? color->blue << (-bs) : color->blue >> bs); 

    color->pixel = ((r & scr->w_visual->red_mask) 
		    | (g & scr->w_visual->green_mask) 
		    | (b & scr->w_visual->blue_mask));

    return color->pixel;
}



int
GetColor(WScreen *scr, XColor *color)
{
    WScreenDescriptor *ws = scr->rscreen;
    int r, g, b;
    int rw, gw, bw;
    int index;
    
    rw = ws->r_colors;
    gw = ws->g_colors;
    bw = ws->b_colors;
    switch (scr->w_visual->class) {
     case PseudoColor:
	if (ws->flags.true_color) {
	    r = (color->red * (rw-1))/0xffff;
	    g = (color->green * (gw-1))/0xffff;
	    b = (color->blue * (bw-1))/0xffff;
	    index = r*gw*bw + g*bw + b;
	    color->pixel = ws->colors[index].pixel;
	} else {
	    r = color->red >> 8;
	    g = color->green >> 8;
	    b = color->blue >> 8;
	    if ((r*r + g*g + b*b) > 49152) {
		color->red = 0xffff;
		color->green = 0xffff;
		color->blue = 0xffff;
		color->pixel = scr->white_pixel;
	    } else {
		color->red = 0;
		color->green = 0;
		color->blue = 0;
		color->pixel = scr->black_pixel;
	    }
	}
	break;
	
     case TrueColor:
	getTrueColor(scr, color);
	break;
	
    }
    return 1;
}


inline void 
AllocColor(WScreen *screen, XColor *color)
{
    if (!screen->rscreen->flags.dither || !wPreferences.save_colors) {
	XAllocColor(dpy, screen->w_colormap, color);
    } else
      GetColor(screen, color);
}


inline void
FreeColor(WScreen *screen, unsigned long pixel)
{
    unsigned long colors[1];
    if (!screen->rscreen->flags.dither || !wPreferences.save_colors) {
	colors[0] = pixel;
	XFreeColors(dpy, screen->colormap, colors, 1, 0);
    }
}


/* this does not dithers the image, it only converts WImage to pixmap format */
int 
DitherImage(WScreen *scr, WImage *image, Pixmap *pixmap)
{
    WScreenDescriptor *ws = scr->rscreen;
    XImage *ximg;
    char *imgdata;

    *pixmap = XCreatePixmap(dpy, scr->w_win, image->width, image->height,
			    scr->w_depth);

    ximg = XCreateImage(dpy, scr->w_visual, scr->w_depth, ZPixmap, 0, NULL,
			image->width, image->height, XBitmapPad(dpy), 0);

    ximg->data = imgdata = malloc(ximg->bytes_per_line*ximg->height);
    if (!ximg->data) {
	XDestroyImage(ximg);
	XFreePixmap(dpy, *pixmap);
	*pixmap = None;
	return 0;
    }

    if (ws->flags.mono) {

    } else {
	int x, y, i;
	int r, g, b, rw, gw, bw, rm, gm, bm;
	unsigned long pixel;

	rw = ws->r_colors;
	gw = ws->g_colors;
	bw = ws->b_colors;
#define RED(i) (image->data[(i)*3])
#define GREEN(i) (image->data[(i)*3+1])
#define BLUE(i)	(image->data[(i)*3+2])
	switch (scr->w_visual->class) {
	 case PseudoColor:
	    if (scr->w_depth==8) {
		i = 0;
		for (y=0; y<image->height; y++) {
		    for (x=0; x<image->width; x++) {
			pixel = RED(i)*gw*bw + GREEN(i)*bw + BLUE(i);
			imgdata[i] = ws->colors[pixel].pixel;
			i++;
		    }
		}
	    }
	    break;

	 case TrueColor:
	    rw = ws->r_shift;
	    gw = ws->g_shift;
	    bw = ws->b_shift;
	    rm = 0xffff/ws->r_colors;
	    gm = 0xffff/ws->g_colors;
	    bm = 0xffff/ws->b_colors;
	    i=0;
	    for (y=0; y<image->height; y++) {
		for (x=0; x<image->width; x++) {
		    r = (rw<0 ? RED(i)*rm << (-rw) : RED(i)*rm >> rw);
		    g = (gw<0 ? GREEN(i)*gm << (-gw) : GREEN(i)*gm >> gw);
		    b = (bw<0 ? BLUE(i)*bm << (-bw) : BLUE(i)*bm >> bw); 
		    pixel = ((r & scr->w_visual->red_mask)
			     | (g & scr->w_visual->green_mask) 
			     | (b & scr->w_visual->blue_mask));
		    XPutPixel(ximg, x, y, pixel);
		    i++;
		}
	    }
	    break;
	}
    }
#undef RED
#undef GREEN
#undef BLUE
    XSetClipMask(dpy, scr->copy_gc, None);
    XPutImage(dpy, *pixmap, scr->copy_gc, ximg, 0, 0, 0, 0, 
	      image->width, image->height);
    XDestroyImage(ximg);
    return 1;
}

#endif /* DITHERING */
