/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "actions.h"
#include "menu.h"
#include "funcs.h" 
#include "dialog.h"
#include "keybind.h"
#include "stacking.h"
#ifdef DOCK
#include "dock.h"
#endif

extern char *MenuConfigPath;
extern wCursor[WCUR_LAST];

extern WPreferences wPreferences;

extern WShortKey *wKeyBindings[WKBD_LAST];

static char **CommandTable=NULL;
static int LastCommand=0;
static int TableSize=0;


/*
 * Syntax:
 * # main menu
 * "Menu Name" MENU
 * 	"Title" EXEC command_to_exec -params
 * 	"Submenu" MENU
 * 		"Title" EXEC command_to_exec -params
 * 	"Submenu" END
 * 	"Workspaces" WORKSPACE_MENU
 * 	"Title" built_in_command
 * "Menu Name" END
 * 
 * Built-in commands:
 * 
 * SHUTDOWN - closes the X server
 * REFRESH - forces the desktop to be repainted
 * EXIT - exit the window manager
 * EXEC <program> - execute an external program
 * WORKSPACE_MENU - places the workspace submenu
 * ARRANGE_ICONS
 * RESTART [<window manager>] - restarts the window manager
 * SHOW_ALL - unhide all windows on workspace
 * HIDE_OTHERS - hides all windows excep the focused one
 */


/* menu commands */

/* change this to a workspace manager call */
static void
execCommand(WMenu *menu, WMenuEntry *entry)
{
    XGrabPointer(dpy, menu->frame->screen_ptr->root_win, True, 0,
		 GrabModeAsync, GrabModeAsync, None, wCursor[WCUR_WAIT],
		 CurrentTime);
    XSync(dpy, 0);
    if (fork()==0) {
	execl("/bin/sh", "/bin/sh", "-c", CommandTable[(int)entry->clientdata],
	      NULL);
	exit(0);
    }
    XUngrabPointer(dpy, CurrentTime);
    XSync(dpy, 0);
}

static void
exitCommand(WMenu *menu, WMenuEntry *entry)
{
    /* prevent reentrant calls */
    wMenuSetEnabled(menu, entry->order, False);
    if (wMessageDialog(menu->frame->screen_ptr, "Exit", "Exit window manager?",
		       WD_EXIT_CONFIRM)==WDB_EXIT) {
#ifdef DEBUG
	printf("Exiting WindowMaker.\n");
#endif
	if (menu->frame->screen_ptr->dock)
	  wDockSaveState(menu->frame->screen_ptr->dock, 
			 find_file(DEF_USERPREF_PATH, DEF_DOCKCONFIG));
	RestoreDesktop(menu->frame->screen_ptr);
	exit(0);
    }
    wMenuSetEnabled(menu, entry->order, True);
}


static void
shutdownCommand(WMenu *menu, WMenuEntry *entry)
{
    /* prevent reentrant calls */
    wMenuSetEnabled(menu, entry->order, False);
    if (wMessageDialog(menu->frame->screen_ptr, "Close X session", 
	"Close Window System session?\n(all applications will be closed)",
		       WD_EXIT_CONFIRM)==WDB_EXIT) {
	printf("Closing...\n");
	if (menu->frame->screen_ptr->dock)
	  wDockSaveState(menu->frame->screen_ptr->dock, 
			 find_file(DEF_USERPREF_PATH, DEF_DOCKCONFIG));
 	WipeDesktop(menu->frame->screen_ptr);
	exit(0);
    }
    wMenuSetEnabled(menu, entry->order, True);
}


static void
restartCommand(WMenu *menu, WMenuEntry *entry)
{
    WWindow *wwin;
    
    wwin = menu->frame->screen_ptr->focused_window;
    while (wwin) {
	wWindowSaveState(wwin);
	wwin = wwin->prev;
    }
    if (menu->frame->screen_ptr->dock)
      wDockSaveState(menu->frame->screen_ptr->dock, 
		     find_file(DEF_USERPREF_PATH, DEF_DOCKCONFIG));
    RestoreDesktop(menu->frame->screen_ptr);
    Restart((char*)entry->clientdata);
}


static void
refreshCommand(WMenu *menu, WMenuEntry *entry)
{
    wRefreshDesktop(menu->frame->screen_ptr);
}


static void
arrangeIconsCommand(WMenu *menu, WMenuEntry *entry)
{
    wArrangeIcons(menu->screen_ptr);
}

static void
showAllCommand(WMenu *menu, WMenuEntry *entry)
{
    wShowAllWindows(menu->screen_ptr);
}

static void
hideOthersCommand(WMenu *menu, WMenuEntry *entry)
{
    wHideOtherApplications(menu->screen_ptr->focused_window);
}

static void
switchWSCommand(WMenu *menu, WMenuEntry *entry)
{
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+2, True);
    wChangeWorkspace(menu->screen_ptr, (int)entry->clientdata-1);
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+2, False);
    /* don't let user destroy current workspace */
    if (menu->screen_ptr->workspace == menu->screen_ptr->last_workspace) {
	wMenuSetEnabled(menu, 1, False);
    } else {
	wMenuSetEnabled(menu, 1, True);
    }
}


static void
newWSCommand(WMenu *menu, WMenuEntry *foo)
{
    char title[MAX_MENU_TEXT_LENGTH+6];
    int ws;
    WMenuEntry *entry;

    wMenuSetEnabled(menu, menu->screen_ptr->workspace+2, True);
    ws = wNewWorkspace(menu->screen_ptr);
    if (ws<0) {
	wMenuSetEnabled(menu, menu->screen_ptr->workspace+2, False);
	wMenuSetEnabled(menu, 0, False);
	return;
    }
    if (ws>=MAX_WORKSPACES) {
	wMenuSetEnabled(menu, 0, False);
    }
    sprintf(title, "Switch to %i", ws+1);
    entry = wMenuAddCallback(menu, title, switchWSCommand, (void*)(ws+1));
    if (ws<9) {
	int kcode;
	if (wKeyBindings[WKBD_WORKSPACE1+ws]) {
	    kcode = wKeyBindings[WKBD_WORKSPACE1+ws]->keycode;
	    entry->rtext = 
	      strdup(XKeysymToString(XKeycodeToKeysym(dpy, kcode, 0)));
	}
    }
    wMenuRealize(menu);
    /* disable current workspace entry */
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+2, False);
    wMenuPaint(menu);
    /* disable destroy last entry */
    wMenuSetEnabled(menu, 1, False);
}


static void
deleteWSCommand(WMenu *menu, WMenuEntry *entry)
{
    int i;
    
    i = menu->screen_ptr->last_workspace;
    wDeleteLastWorkspace(menu->screen_ptr);
    if (i!=menu->screen_ptr->last_workspace) {
	wMenuRemoveItem(menu, i+2);
	wMenuRealize(menu);
	wMenuPaint(menu);
    }
    /* don't let user destroy current workspace */
    if (menu->screen_ptr->workspace == menu->screen_ptr->last_workspace) {
	wMenuSetEnabled(menu, 1, False);
    } else {
	wMenuSetEnabled(menu, 1, True);
    }
}

/*******************************/

/* menu config. parser */

static void
destroyCommands(char **table, int size)
{
    int i;
    
    if (size<=0) return;
    for (i=0; i<size; i++) {
	free(table[i]);
    }
    wfree(table);
}


static int 
registerCommand(char *command)
{
    /* enlarge table */
    if (TableSize>=LastCommand) {
	char **oldtable=CommandTable;
	
	TableSize += 5;
	CommandTable = wrealloc(CommandTable, sizeof(char *)*TableSize);
	if (CommandTable==NULL) {
	    wWarning("out of memory to store menu commands. Nuking it.");
	    destroyCommands(oldtable, LastCommand);
	    return -1;
	}
    }
    
    CommandTable[LastCommand] = strdup(command);
    return LastCommand++;
}



static void
addWorkspaceMenu(WScreen *scr, WMenu *menu, char *title)
{
    char wstitle[MAX_MENU_TEXT_LENGTH+6];
    WMenu *wsmenu;
    WMenuEntry *entry;
    int i;

    wsmenu = wMenuCreate(scr, "Workspace", False);
    if (!wsmenu) {
	wWarning("could not create Workspace menu");
	return;
    }
    wMenuAddCallback(wsmenu, "New", newWSCommand, NULL);
    wMenuAddCallback(wsmenu, "Destroy Last", deleteWSCommand, NULL);
    if (scr->last_workspace>=MAX_WORKSPACES) {
	wMenuSetEnabled(wsmenu, 0, False);
    }
    for (i=0; i<=scr->last_workspace; i++) {
	sprintf(wstitle, "Switch to %i", i+1);
	entry = wMenuAddCallback(wsmenu, wstitle, switchWSCommand, 
				 (void*)(i+1));
	if (i<10) {
	    int kcode;
	    if (wKeyBindings[WKBD_WORKSPACE1+i]) {
		kcode = wKeyBindings[WKBD_WORKSPACE1+i]->keycode;
		entry->rtext = 
		  strdup(XKeysymToString(XKeycodeToKeysym(dpy, kcode, 0)));
	    }
	}
    }
    wMenuSetEnabled(wsmenu, scr->workspace+2, False);
    wMenuSetEnabled(wsmenu, 1, False);
    wMenuEntrySetCascade(menu, wMenuAddCallback(menu, title, NULL, NULL), 
			 wsmenu);
    scr->workspace_menu = wsmenu;    
}


static WMenu*
parseCascade(WScreen *scr, WMenu *menu, FILE *file)
{
    char linebuf[MAXLINE];
    char title[MAXLINE];
    char command[MAXLINE];
    char params[MAXLINE];
    char *line;
    
    while (!feof(file)) {
	fgets(linebuf, MAXLINE, file);
	line = cropline(linebuf);
	if (line[0]==0 || line[0]=='#' || (line[0]=='/' && line[1]=='/')) 
	  continue;
	
	if (separateline(line, title, command, params) < 2) {
	    wWarning("syntax error in menu config. %s", line);
	    goto error;
	}

	if (strcmp(command, "MENU")==0) {
	    WMenu *cascade;

	    cascade = wMenuCreate(scr, title, False);
	    if (parseCascade(scr, cascade, file)==NULL) {
		wMenuDestroy(cascade, True);
	    } else {
		wMenuEntrySetCascade(menu, 
				     wMenuAddCallback(menu, title, NULL, NULL),
				     cascade);
	    }
	} else {
	    /* see if the menu has ended */
	    if (strcmp(command, "END")==0) {
		return menu;
	    }
	    /* add new item */
	    if (strcmp(command, "EXEC")==0) {
		int code;

		code = registerCommand(params);
		if (code<0)
		  goto error;
		wMenuAddCallback(menu, title, execCommand, (void*)code);

	    } else if (strcmp(command, "EXIT")==0) {
		wMenuAddCallback(menu, title, exitCommand, NULL);
		
	    } else if (strcmp(command, "SHUTDOWN")==0) {
		wMenuAddCallback(menu, title, shutdownCommand, NULL);
		
	    } else if (strcmp(command, "REFRESH")==0) {
		wMenuAddCallback(menu, title, refreshCommand, NULL);
		
	    } else if (strcmp(command, "WORKSPACE_MENU")==0) {
		if (scr->workspace_menu)
		  wWarning("menu can only have one WORKSPACE_MENU");
		else
		  addWorkspaceMenu(scr, menu, title);
	    } else if (strcmp(command, "ARRANGE_ICONS")==0) {
		wMenuAddCallback(menu, title, arrangeIconsCommand, NULL);
	    } else if (strcmp(command, "HIDE_OTHERS")==0) {
		wMenuAddCallback(menu, title, hideOthersCommand, NULL);
	    } else if (strcmp(command, "SHOW_ALL")==0) {
		wMenuAddCallback(menu, title, showAllCommand, NULL);
	    } else if (strcmp(command, "RESTART")==0) {
		WMenuEntry *entry;

		entry = wMenuAddCallback(menu, title, restartCommand, 
					 strdup(params));
		entry->free_cdata = free;
	    } else {
		wWarning("unknown command \"%s\" in menu config.", command);
	    }
	}
    }
    wWarning("syntax error in menu config. END declaration missing");
    return menu;

    error:
    if (menu)
      wMenuDestroy(menu, True);
    return NULL;
}


static WMenu*
parseMenu(WScreen *scr, FILE *file)
{
    WMenu *menu=NULL;
    char linebuf[MAXLINE];
    char title[MAXLINE];
    char command[MAXLINE];
    char params[MAXLINE];
    char *line;
    
    while (!feof(file)) {
	fgets(linebuf, MAXLINE, file);
	line = cropline(linebuf);
	if (line[0]==0 || line[0]=='#' || (line[0]=='/' && line[1]=='/')) 
	  continue;

	if (separateline(line, title, command, params) < 2) {
	    wWarning("syntax error in menu config. %s", line);
	    return NULL;
	}

	if (strcasecmp(command, "MENU")==0) {
	    menu = wMenuCreate(scr, title, True);
	    return parseCascade(scr, menu, file);
	} else {
	    wWarning("no title given for the root menu");
	    return NULL;
	}
    }
    return NULL;
}

static WMenu *
makeDefaultMenu(WScreen *scr)
{
    WMenu *menu=NULL;
    
    menu = wMenuCreate(scr, "Commands", True);
    wMenuAddCallback(menu, "Exit...", exitCommand, NULL);
    return menu;
}


static void
raiseMenus(WMenu *menu)
{
    int i;
    
    if (menu->flags.mapped) {
	if (!wPreferences.obscurable_menus)
	  wRaiseFrame(menu->frame);
	else
	  XRaiseWindow(dpy, menu->frame->window);
    }
    for (i=0; i<menu->entry_no; i++) {
	if (menu->entries[i]->cascade)
	  raiseMenus(menu->entries[i]->cascade);
    }
}



void 
RefreshWorkspaceMenu(WScreen *scr)
{
    WMenu *menu = scr->workspace_menu;
    char title[MAX_MENU_TEXT_LENGTH+6];
    int i;

    if (!menu)
      return;

    for (i=0; i<=scr->last_workspace+2; i++) 
      wMenuSetEnabled(menu, i, True);
    if (menu->entry_no-2<=scr->last_workspace) {
	WMenuEntry *entry;

	for (i=menu->entry_no-2; i<=scr->last_workspace; i++) {
	    sprintf(title, "Switch to %i", i+1);
	    entry = wMenuAddCallback(menu, title, switchWSCommand, 
				     (void*)(i+1));
	    if (i<10) {
		int kcode;
		if (wKeyBindings[WKBD_WORKSPACE1+i]) {
		    kcode = wKeyBindings[WKBD_WORKSPACE1+i]->keycode;
		    entry->rtext = 
		      strdup(XKeysymToString(XKeycodeToKeysym(dpy, kcode, 0)));
		}
	    }
	}
	wMenuRealize(menu);
	/* disable current workspace entry */
	wMenuSetEnabled(menu, scr->workspace+2, False);
	wMenuPaint(menu);
	/* disable "destroy last" entry */
	wMenuSetEnabled(menu, 1, False);
    } else {
	wMenuSetEnabled(menu, scr->workspace+2, False);
    }
    wMenuSetEnabled(menu, 0, True);
    wMenuSetEnabled(menu, 1, (scr->workspace!=scr->last_workspace));
}

/*
 *---------------------------------------------------------------------- 
 * OpenRootMenu--
 * 	Opens the root menu, parsing the menu configuration file if it
 * wasn't already or the config file was changed.
 *	If the menu is already mapped and is not sticked to the
 * root window, it will be unmapped.
 * 
 * Side effects:
 * 	The command table and the menu may be remade.
 * 
 * Notes:
 * 	If the menu was never mapped or the configuration file was 
 * changed since the last mapping the menu will be reconstructed. 
 * If the menu reconstruction fails, the previous menu will be used.
 *---------------------------------------------------------------------- 
 */
void
OpenRootMenu(WScreen *scr, int x, int y)
{
    WMenu *menu=NULL;
    FILE *file=NULL;
    struct stat stats;
    static time_t last_changed=0;
#ifdef USECPP
    char command[MAXLINE], *args;
    int cpp=0;
#endif

    if (scr->root_menu && scr->root_menu->flags.mapped) {
	if (!scr->root_menu->flags.buttoned)
	  wMenuUnmap(scr->root_menu);
	else
	  raiseMenus(scr->root_menu);
	return;
    }
    if (!MenuConfigPath) {
	MenuConfigPath = find_file(DEF_CONFIG_PATHS, DEF_MENUCONFIG);
    }
    if (MenuConfigPath) {
	if (stat(MenuConfigPath, &stats)<0) {
	    wSysError("could not stat menu configuration file \"%s\"", 
		      MenuConfigPath);
	    return;
	}
	/* user changed menu config. Reread menu definition */
	if (scr->root_menu==NULL || stats.st_mtime!=last_changed) {
	    char **cmd_table;
	    int last_cmd;
	    int table_size;
	    
#ifdef DEBUG
	    puts("Reparsing menu configs");
#endif	
	    last_changed = stats.st_mtime;

#ifdef USECPP
	    args = MakeCPPArgs(MenuConfigPath);
	    if (!args) {
		wWarning("could not make arguments for menu file preprocessor");
	    } else {
		sprintf(command, "%s %s %s", CPP_PATH, args, MenuConfigPath);
		wfree(args);
		file = popen(command, "r");
		if (!file) {
		    wSysError("could not open menu configuration preprocessor");
		} else {
		    cpp = 1;
		}
	    }
#endif /* USECPP */
	    if (!file) {
		file = fopen(MenuConfigPath, "r");
		if (file == NULL) {
		    wSysError("could not open menu configuration file \"%s\"", 
			      MenuConfigPath);
		    return;
		}
	    }
	    /* save the current command table */
	    cmd_table = CommandTable;
	    last_cmd = LastCommand;
	    table_size = TableSize;
	    /* reset commands */
	    CommandTable = NULL;
	    LastCommand = 0;
	    TableSize = 0;

	    scr->workspace_menu = NULL;
	    menu = parseMenu(scr, file);
#ifdef CPP
	    if (cpp) {
		pclose(file):
	    } else {
		fclose(file);
	    }
#else
	    fclose(file);
#endif
	    if (menu!=NULL) {
		if (scr->root_menu) {
		    wMenuDestroy(scr->root_menu, True);
		    /* destroy the old command table */
		    destroyCommands(cmd_table, last_cmd);
		}
		scr->root_menu = menu;
	    } else {
		wMessageDialog(scr, "Error", "Could not read menu configuration\n"\
       			"(see console log for details)", WD_INFORMATION);
		/* restore the old command table */
		destroyCommands(CommandTable, LastCommand);
		CommandTable = cmd_table;
		LastCommand = last_cmd;
		TableSize = table_size;
	    }
	}
    }
    menu = scr->root_menu;
    if (!menu) {
	menu = makeDefaultMenu(scr);
	scr->root_menu = menu;
    }
    if (menu) {
	wMenuMapAt(menu, x-menu->frame->width/2, y-menu->titlebar->height/2);
    }
}


