/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "texture.h"
#include "def_pixmaps.h"
#include "screen.h"
#include "pixmap.h"
#include "button.h"
#include "menu.h"
#include "resource.h"
#include "raster.h"
#include "funcs.h"
#include "actions.h"
#include "properties.h"
#include "setup.h"

#define EVENT_MASK (LeaveWindowMask|EnterWindowMask|PropertyChangeMask\
		|SubstructureNotifyMask|PointerMotionMask \
		|SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask\
		|KeyPressMask|KeyReleaseMask)


#define STIPPLE_WIDTH 2
#define STIPPLE_HEIGHT 2
static char STIPPLE_DATA[] = {0x02, 0x01};

/**** Global variables ****/

extern Cursor wCursor[WCUR_LAST];
extern WResource wResourceDefault;
extern WPreferences wPreferences;
extern char *PixmapPath;


/*
 *----------------------------------------------------------------------
 * alreadyRunningError--
 * 	X error handler used to catch errors when trying to do
 * XSelectInput() on the root window. These errors probably mean that
 * there already is some other window manager running.
 * 
 * Returns:
 * 	Nothing, unless something really evil happens...
 * 
 * Side effects:
 * 	The program is aborted.
 *---------------------------------------------------------------------- 
 */
static int 
alreadyRunningError(Display *dpy, XErrorEvent *error)
{
    wFatal("it seems that there already is a window manager running");
    exit(1);
    return -1;
}


/*
 *---------------------------------------------------------------------- 
 * allocButtonPixmaps--
 * 	Allocate pixmaps used on window operation buttons (those in the
 * titlebar). The pixmaps are linked to the program. If XPM is supported
 * XPM pixmaps are used otherwise, equivalent bitmaps are used.
 * 
 * Returns:
 * 	Nothing
 * 
 * Side effects:
 * 	Allocates shared pixmaps for the screen. These pixmaps should 
 * not be freed by anybody.
 *---------------------------------------------------------------------- 
 */
static void
allocButtonPixmaps(WScreen *scr)
{
    WPixmap *pix;

    /* create predefined pixmaps */
#ifdef XPM
    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE_XPM);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_CLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE_XPM);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_BROKENCLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY_XPM);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_ICONIFY] = pix;
                
#ifdef NEWSTYLE
    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE_XPM2);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE_XPM2);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY_XPM2);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY] = pix;
#endif /* NEWSTYLE */
    /* create predefined bitmaps (xpm not available) */
#else /* !XPM */
    pix = wPixmapCreateFromXBMData(scr, (char*)PRED_CLOSE_XBM_DATA, 
				   (char*)PRED_CLOSE_XBM_DATA, 
				   PRED_CLOSE_XBM_WIDTH, 
				   PRED_CLOSE_XBM_HEIGHT, 
				   scr->black_pixel, scr->white_pixel);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_CLOSE] = pix;

    pix = wPixmapCreateFromXBMData(scr, (char*)PRED_BROKEN_CLOSE_XBM_DATA, 
				   (char*)PRED_BROKEN_CLOSE_XBM_DATA, 
				   PRED_BROKEN_CLOSE_XBM_WIDTH, 
				   PRED_BROKEN_CLOSE_XBM_HEIGHT, 
				   scr->black_pixel, scr->white_pixel);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_BROKENCLOSE] = pix;

    pix = wPixmapCreateFromXBMData(scr, (char*)PRED_ICONIFY_XBM_DATA, 
				   (char*)PRED_ICONIFY_XBM_DATA, 
				   PRED_ICONIFY_XBM_WIDTH, 
				   PRED_ICONIFY_XBM_HEIGHT, 
				   scr->black_pixel, scr->white_pixel);
    if (pix)
      pix->shared = 1;
    scr->b_pixmaps[WBUT_ICONIFY] = pix;
#ifdef NEWSTYLE
    scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE] = scr->b_pixmaps[WBUT_CLOSE];

    scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE]=scr->b_pixmaps[WBUT_BROKENCLOSE];

    scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY] = scr->b_pixmaps[WBUT_ICONIFY];
#endif /* NEWSTYLE */
 
#endif /* XPM */
}

/*
 *----------------------------------------------------------------------
 * makeIconBackImage--
 * 	Creates a plain 24bit RGB WImage of a gray square to be
 * used as a default icon background.
 * 
 * Returns:
 * 	The image to be used as icon background.
 * 
 * TODO: Draw borders.
 * 	Make a general routine to make a background with user defined
 * textures.
 *---------------------------------------------------------------------- 
 */
#ifdef DITHERING
static WImage *
makeIconBackImage(WScreen *scr)
{
    WImage *img;
    
    img = wmalloc(sizeof(WImage));
    img->data = wmalloc(ICON_WIDTH*ICON_HEIGHT*3);
    memset(img->data, 0xbe, ICON_WIDTH*ICON_HEIGHT*3);
    img->alpha = NULL;
    img->width = ICON_WIDTH;
    img->height = ICON_HEIGHT;
    return img;
}
#endif

/*
 *----------------------------------------------------------------------
 * makeIconBackPixmap--
 * 	Creates a pixmap used as the icon background.
 *
 * Returns:
 * 	A pixmap of the default screen depth (not DefaultDepth())
 * 
 * TODO: See makeIconBackImage()
 *---------------------------------------------------------------------- 
 */
static Pixmap
makeIconBackPixmap(WScreen *scr)
{
    Pixmap pix;
    XColor color;
    GC gc;
    XGCValues gcv;
    WTexture *tex;
    
    pix = XCreatePixmap(dpy, scr->w_win, ICON_WIDTH, ICON_HEIGHT, 
			scr->w_depth);
    if (!XParseColor(dpy, scr->colormap, DEF_ICON_TEXTURE_COLOR, &color)){
	wWarning("could not parse color %s", DEF_ICON_TEXTURE_COLOR);
	color.red=0xbebe;
	color.green=0xbebe;
	color.blue=0xbebe;
    }
    tex = (WTexture*)wTextureMakeSolid(scr, &color, WREL_ICON);
    
    gcv.foreground = tex->any.color.pixel;
    gc = XCreateGC(dpy, scr->w_win, GCForeground, &gcv);
    XFillRectangle(dpy, pix, gc, 0, 0, ICON_WIDTH, ICON_HEIGHT);
    XFreeGC(dpy, gc);

    
    XDrawRectangle(dpy, pix, tex->solid.light_gc, 0, 0, ICON_WIDTH, 1);
    XDrawRectangle(dpy, pix, tex->solid.light_gc, 0, 0, 1, ICON_HEIGHT);
    
    XDrawLine(dpy, pix, tex->solid.dark_gc, 0, ICON_HEIGHT-1, 
	      ICON_WIDTH, ICON_HEIGHT-1);
    XDrawLine(dpy, pix, tex->solid.dim_gc, 1, ICON_HEIGHT-2, 
	      ICON_WIDTH-2, ICON_HEIGHT-2);
    
    XDrawLine(dpy, pix, tex->solid.dark_gc, ICON_WIDTH-1, 0, 
	      ICON_WIDTH-1, ICON_HEIGHT-1);
    XDrawLine(dpy, pix, tex->solid.dim_gc, ICON_WIDTH-2, 1, 
	      ICON_WIDTH-2, ICON_HEIGHT-2);
    wTextureDestroy(scr, tex);
    return pix;
}


/*
 *---------------------------------------------------------------------- 
 * loadFont--
 * 	Loads a single font into the WFont structure, initializing
 * data in it. If the font could not be loaded, fixed will be used.
 * If fixed can't be loaded either, the program is aborted.
 * 
 * Returns:
 * 	None
 * 
 * TODO: I18N support (current does not work)
 *---------------------------------------------------------------------- 
 */
static void
loadFont(char *font_name, WFont *font)
{
#ifdef I18N
    char **missing;
    int num_missing = 0;
    char *default_string;
    int skyline;   
 
    font->font = XCreateFontSet(dpy, font_name, &missing, &num_missing, 
				&default_string);
    if (num_missing > 0) {
        wWarning("The following character sets are missing:");
        for (skyline = 0; skyline < num_missing; skyline++)
	  wWarning(missing[skyline]);
        XFreeStringList(missing);
        wWarning("The string \"%s\" will be used in place",
		 default_string);
        wWarning("of any characters from those sets.");
    }
    if (!font->font) {
	wWarning("could not create font set %s. Using fixed", font_name);
	font->font = XCreateFontSet(dpy, "fixed", &missing, &num_missing, 
				    &default_string);
	if (num_missing > 0) {
	    XFreeStringList(missing);
	}
	if (!font->font) {
	    wFatal("could not load any usable font set");
	    wAbort();
	    exit(1);
	}
    }
    font->height = XExtentsOfFontSet(font->font)->max_logical_extent.height;
    font->y = XExtentsOfFontSet(font->font)->max_logical_extent.y + 23;
#else
    font->font = XLoadQueryFont(dpy, font_name);
    if (!font->font) {
	wWarning("could not load font %s. Using fixed", font_name);
	font->font = XLoadQueryFont(dpy, "fixed");
	if (!font->font) {
	    wFatal("could not load any usable font");
	    wAbort();
	    exit(1);
	}
    }
    font->height = font->font->ascent+font->font->descent;
    font->y = font->font->ascent;
#endif /* !I18N */
}


/*
 *----------------------------------------------------------------------
 * parseColor--
 * 	Parses a color name. If the color could not be parsed, the color
 * will be set to black if def_black is true or white otherwise.
 * 
 * Returns:
 * 	0 if the color could not be parsed or
 * 	1 on success
 *---------------------------------------------------------------------- 
 */
static int
parseColor(WScreen *scr, char *color_name, XColor *color, int def_black)
{
    if (!XParseColor(dpy, scr->colormap, color_name, color)) {
	wWarning("could not parse color %s", color_name);
	color->red = (def_black ? 0 : 0xffff);
	color->green = (def_black ? 0 : 0xffff);
	color->blue = (def_black ? 0 : 0xffff);
	return 0;
    }
    return 1;
}


/*
 *---------------------------------------------------------------------- 
 * allocFonts--
 * 	Allocates all fonts needed by the window manager. The font names
 * are taken from what was read in preferences. Default names from
 * wconfig.h are used if some setting was not specified. If some font 
 * cannot be loaded, fixed is used. If it can't be loaded too, the 
 * program is aborted.
 *
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	Fonts are loaded and the font names in preferences are freed.
 *
 *---------------------------------------------------------------------- 
 */
static void
allocFonts(WScreen *scr)
{
    if (wPreferences.title_font) {
	loadFont(wPreferences.title_font, &scr->title_font);
	free(wPreferences.title_font);
    } else
      loadFont(DEF_TITLE_FONT, &scr->title_font);
    
    if (wPreferences.menu_title_font) {
	loadFont(wPreferences.menu_title_font, &scr->menu_title_font);
	free(wPreferences.menu_title_font);
    } else
      loadFont(DEF_MENU_TITLE_FONT, &scr->menu_title_font);
    
    if (wPreferences.menu_entry_font) {
	loadFont(wPreferences.menu_entry_font, &scr->menu_entry_font);
	free(wPreferences.menu_entry_font);
    } else
      loadFont(DEF_MENU_ENTRY_FONT, &scr->menu_entry_font);

    if (wPreferences.icon_title_font) {
	loadFont(wPreferences.icon_title_font, &scr->icon_title_font);
	free(wPreferences.icon_title_font);
    } else
      loadFont(DEF_ICON_TITLE_FONT, &scr->icon_title_font);

    if (wPreferences.panel_title_font) {
	loadFont(wPreferences.panel_title_font, &scr->panel_title_font);
	free(wPreferences.panel_title_font);
    } else
      loadFont(DEF_PANEL_TITLE_FONT, &scr->panel_title_font);
    
    if (wPreferences.panel_text_font) {
	loadFont(wPreferences.panel_text_font, &scr->panel_text_font);
	free(wPreferences.panel_text_font);
    } else
      loadFont(DEF_PANEL_TEXT_FONT, &scr->panel_text_font);
    
    if (wPreferences.info_font) {
	loadFont(wPreferences.info_font, &scr->info_text_font);
	free(wPreferences.info_font);
    } else
      loadFont(DEF_INFO_TEXT_FONT, &scr->info_text_font);
}


static WTexture*
makeTexture(scr, border, type, scolor, gcolor, def_scolor, def_black)
WScreen *scr;			       /* screen descriptor */
int border;			       /* border type */
int type;			       /* type of texture */
char *scolor;			       /* color for solid */
W2Color *gcolor;		       /* colors for gradient */
char *def_scolor;		       /* default solid color */
int def_black;			       /* assume black on failure? */
{
    XColor color, color2;
    WTexture *tex;
    int ok=1;

    switch (type) {
     case WTEX_VGRADIENT:
     case WTEX_HGRADIENT:
     case WTEX_DGRADIENT:
	if (scr->w_visual->class==PseudoColor) {
	    ok=0;	    
	    if (scr->w_depth!=8)
	      wWarning("Gradient texture not supported for your display "\
			"(PseudoColor %ibpp)", scr->w_depth);
	    else if (!wPreferences.use_dithering)
	      wWarning("You must enable dithering to be able to use "\
			"gradients with this display");
	    else ok=1;
	}
	if ((scr->w_visual->class==PseudoColor && ok)
	    || scr->w_visual->class==TrueColor) {
	    if (gcolor->color1 && gcolor->color2
		&& parseColor(scr, gcolor->color1, &color, 0)
		&& parseColor(scr, gcolor->color2, &color2, 0)) {
		tex = (WTexture*)wTextureMakeGradient(scr, type, &color,
						      &color2, border);
		if (tex) break;
	    }
	} else
	  wWarning("Gradient texture available only for PseudoColor and "\
			"TrueColor displays");
	  
	/* if others failed, make solid */
     case WTEX_SOLID:
     default:
	if (!scolor) {
	    parseColor(scr, def_scolor, &color, def_black);
	} else {
	    parseColor(scr, scolor, &color, def_black);
	}
	tex = (WTexture*)wTextureMakeSolid(scr, &color, border);
	break;
    }
    if (gcolor->color1)
      wfree(gcolor->color1);
    if (gcolor->color2)
      wfree(gcolor->color2);
    if (scolor)
      free(scolor);

    if (!tex) {
	wFatal("could not allocate texture");
	wAbort();
	exit(1);
    }
    return tex;
}


/*
 *----------------------------------------------------------------------
 * initTextures--
 * 	Creates textures to be used for decorations, depending on
 * user preferences or default values in wconfig.h
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	Textures are allocated for the screen and strings allocated
 * during preference file read are freed.
 *----------------------------------------------------------------------
 */
static void
initTextures(WScreen *scr)
{    
    /* focused titlebar */
    scr->focused_texture = makeTexture(scr, WREL_TITLEBAR,
				       wPreferences.ftitle_texture,
				       wPreferences.ftitle_bcolor,
				       &wPreferences.ftitle_gradient,
				       DEF_FOCUSED_TEXTURE_COLOR, 1);
    scr->icon_title_texture = 
      wTextureMakeSolid(scr, &scr->focused_texture->any.color, WREL_TITLEBAR);

    /* semi-focused titlebar */
    scr->pfocused_texture = makeTexture(scr, WREL_TITLEBAR,
					wPreferences.ptitle_texture,
					wPreferences.ptitle_bcolor,
					&wPreferences.ptitle_gradient,
					DEF_PREV_FOCUSED_TEXTURE_COLOR, 0);

    /* unfocused titlebar */
    scr->unfocused_texture = makeTexture(scr, WREL_TITLEBAR,
					 wPreferences.utitle_texture,
					 wPreferences.utitle_bcolor,
					 &wPreferences.utitle_gradient,
					 DEF_UNFOCUSED_TEXTURE_COLOR, 0);

    /* menu title */
    scr->menu_title_texture = makeTexture(scr, WREL_MENUTITLE,
					  wPreferences.menu_title_texture,
					  wPreferences.menu_title_bcolor,
					  &wPreferences.menu_title_gradient,
					  DEF_FOCUSED_TEXTURE_COLOR, 1);

    /* menu text */
    scr->menu_item_texture = makeTexture(scr, WREL_MENUENTRY,
					 wPreferences.menu_text_texture,
					 wPreferences.menu_text_bcolor,
					 &wPreferences.menu_text_gradient,
					 DEF_UNFOCUSED_TEXTURE_COLOR, 1);

    if (scr->menu_item_texture->any.type==WTEX_SOLID) {
	scr->menu_item_auxtexture = (WTexSolid*)scr->menu_item_texture;
    } else {
	scr->menu_item_auxtexture = 
	  wTextureMakeSolid(scr,&scr->menu_item_texture->any.color, WREL_FLAT);
    }
    
    /* pushed buttons */
    {
	XColor color;
#ifdef NEWSTYLE    
	color.red = 0x7777;
	color.green = 0x7777;
	color.blue = 0x7777;
#else    
	color.red = 0xffff;
	color.green = 0xffff;
	color.blue = 0xffff;
#endif
	scr->pushed_button_texture = (WTexture*)wTextureMakeSolid(scr, &color,
								  WREL_SUNKEN);
	color.red = 0xbebe;
	color.green = 0xbebe;
	color.blue = 0xbebe;
	scr->normal_button_texture =
	  (WTexture*)wTextureMakeSolid(scr, &color, WREL_RAISED);
    }
#undef GETGRADCOLORS
    /* resizebar */
    scr->resizebar_texture = 
      (WTexture*)wTextureMakeSolid(scr, &scr->unfocused_texture->any.color, 
				   WREL_FLAT);
}


/*
 *----------------------------------------------------------------------
 * allocResources--
 * 	Allocates resources (fonts, GCs, textures and pixmaps) for the
 * screen.
 * 
 * Returns:
 * 	Nothing.
 * 
 * Side effects:
 * 	GCs, colors, textures etc. are allocated for the screen.
 *---------------------------------------------------------------------- 
 */
static void
allocResources(WScreen *scr)
{
    XGCValues gcv;
    XColor color;
    WPixmap *pix;
    unsigned long mtextcolor;
    int gcm;

    scr->stipple_bitmap = 
      XCreateBitmapFromData(dpy, scr->w_win, STIPPLE_DATA, STIPPLE_WIDTH,
			    STIPPLE_HEIGHT);
    /* fonts */    
    allocFonts(scr);

    gcv.stipple = scr->stipple_bitmap;
    gcv.foreground = scr->white_pixel;
    gcv.fill_style = FillStippled;
    gcm = GCForeground|GCStipple|GCFillStyle;
    scr->stipple_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
    
    gcm = GCForeground;
    /* menu GCs */
    if (!wPreferences.menu_title_fcolor) {
	parseColor(scr, DEF_FOCUSED_TEXT_COLOR, &color, 1);
    } else {
	if (!parseColor(scr, wPreferences.menu_title_fcolor, &color, 1)) {
	    free(wPreferences.menu_title_fcolor);
	    parseColor(scr, DEF_FOCUSED_TEXT_COLOR, &color, 1);
	}
    }
    AllocColor(scr, &color);
    scr->mtitle_pixel = color.pixel;
    gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->menu_title_font.font->fid;
    gcm |= GCFont;
#endif
    scr->menu_title_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    if (!wPreferences.menu_text_fcolor) {
	parseColor(scr, DEF_UNFOCUSED_TEXT_COLOR, &color, 0);
    } else {
	if (!parseColor(scr, wPreferences.menu_text_fcolor, &color, 0)) {
	    free(wPreferences.menu_text_fcolor);
	    parseColor(scr, DEF_UNFOCUSED_TEXT_COLOR, &color, 0);
	}
    }
    AllocColor(scr, &color);
    scr->mtext_pixel = color.pixel;
    mtextcolor = gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->menu_entry_font.font->fid;
#endif
    scr->menu_entry_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    /* selected menu entry GC */
    gcm = GCForeground|GCBackground;
#ifndef I18N
    gcv.font = scr->menu_entry_font.font->fid;
    gcm |= GCFont;
#endif
    gcv.foreground = scr->white_pixel;
    gcv.background = scr->white_pixel;
    scr->select_menu_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    if (!wPreferences.menu_text_dcolor) {
	color.red = 0x7777;
	color.green = 0x7777;
	color.blue = 0x7777;
    } else {
	if (!parseColor(scr, wPreferences.menu_text_dcolor, &color, 0)) {
	    color.red = 0x7777;
	    color.green = 0x7777;
	    color.blue = 0x7777;
	    free(wPreferences.menu_text_dcolor);
	}
    }
    AllocColor(scr, &color);
    scr->dtext_pixel = color.pixel;
    gcm = GCForeground|GCBackground;
#ifndef I18N
    gcv.font = scr->menu_entry_font.font->fid;
    gcm |= GCFont;
#endif    
    if (mtextcolor == color.pixel) {
	gcv.foreground = scr->white_pixel;
	gcv.background = scr->black_pixel;
	gcv.stipple = scr->stipple_bitmap;
	gcv.fill_style = FillStippled;
	gcm |= GCStipple|GCFillStyle;
	scr->disabled_menu_entry_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
    } else {
	gcv.foreground = color.pixel;
	scr->disabled_menu_entry_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
    }

    /* frame GC */
    parseColor(scr, DEF_FRAME_COLOR, &color, 0);

    gcv.foreground = ~color.pixel;
    gcv.function = GXxor;
    gcv.line_width = DEF_FRAME_THICKNESS;
    gcv.subwindow_mode = IncludeInferiors;
    scr->frame_gc = XCreateGC(dpy, scr->root_win, GCForeground
			      |GCFunction|GCSubwindowMode|GCLineWidth, &gcv);
    
    /* line GC */
    gcv.foreground = ~color.pixel;
    gcv.function = GXxor;
    gcv.subwindow_mode = IncludeInferiors;
#ifndef I18N
    gcv.font = scr->info_text_font.font->fid;
    gcm = GCForeground|GCFont|GCFunction|GCSubwindowMode;
#else
    gcm = GCForeground|GCFunction|GCSubwindowMode;
#endif
    scr->line_gc = XCreateGC(dpy, scr->root_win, gcm, &gcv);

    /* copy GC */
    gcv.foreground = scr->white_pixel;
    gcv.background = scr->black_pixel;
    scr->copy_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCBackground, &gcv);
    

    /* window title text GCs */
#ifndef I18N
    gcv.font = scr->title_font.font->fid;
    gcm = GCForeground|GCFont;
#else
    gcm = GCForeground;
#endif

    /* focused window title */
    if (!wPreferences.ftitle_fcolor) {
	parseColor(scr, DEF_FOCUSED_TEXT_COLOR, &color, 1);
    } else {
	if (!parseColor(scr, wPreferences.ftitle_fcolor, &color, 1)) {
	    free(wPreferences.ftitle_fcolor);
	    parseColor(scr, DEF_FOCUSED_TEXT_COLOR, &color, 1);
	}
    }
    AllocColor(scr, &color);
    scr->ftitle_pixel = color.pixel;
    gcv.foreground = color.pixel;
    scr->focused_text_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    /* icon title GC */
    gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->icon_title_font.font->fid;
#endif
    scr->icon_title_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    /* semi-focused window title */
    if (!wPreferences.ptitle_fcolor) {
	parseColor(scr, DEF_PREV_FOCUSED_TEXTURE_COLOR, &color, 1);
    } else {
	if (!parseColor(scr, wPreferences.ptitle_fcolor, &color, 1)) {
	    free(wPreferences.ptitle_fcolor);
	    parseColor(scr, DEF_PREV_FOCUSED_TEXTURE_COLOR, &color, 1);
	}
    }
    AllocColor(scr, &color);
    scr->ptitle_pixel = color.pixel;
    gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->title_font.font->fid;
#endif
    scr->pfocused_text_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);


    /* unfocused window title */
    if (!wPreferences.utitle_fcolor) {
	parseColor(scr, DEF_UNFOCUSED_TEXT_COLOR, &color, 1);
    } else {
	if (!parseColor(scr, wPreferences.utitle_fcolor, &color, 1)) {
	    free(wPreferences.utitle_fcolor);
	    parseColor(scr, DEF_UNFOCUSED_TEXT_COLOR, &color, 1);
	}
    }
    AllocColor(scr, &color);
    scr->utitle_pixel = color.pixel;
    gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->title_font.font->fid;
#endif
    scr->unfocused_text_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

    /* move/size display GC */
    gcv.foreground = color.pixel;
#ifndef I18N
    gcv.font = scr->info_text_font.font->fid;
    gcm = GCForeground|GCFont;
#else
    gcm = GCForeground;
#endif
    scr->info_text_gc = XCreateGC(dpy, scr->root_win, gcm, &gcv);

    /* textures */
    initTextures(scr);
    
#ifdef XPM
    /* load pixmaps */
    pix = wPixmapCreateFromXPMData(scr, BUTTON_ARROW_XPM);
    if (pix!=NULL)
      pix->shared = 1;
    scr->button_arrow = pix;
#else /* !XPM */
    pix = wPixmapCreateFromXBMData(scr, (char*)BUTTON_ARROW_XBM_DATA, 
				   (char*)BUTTON_ARROW_MASK_XBM_DATA,
				   BUTTON_ARROW_XBM_WIDTH, 
				   BUTTON_ARROW_XBM_HEIGHT,
				   scr->black_pixel, scr->white_pixel);
    if (pix!=NULL)
      pix->shared = 1;
    scr->button_arrow = pix;
#endif /* XPM */

    if (wPreferences.icon_back_file) {
#ifdef DITHERING
	scr->icon_back_image = LoadImageFile(scr,wPreferences.icon_back_file);
#endif
	pix = wPixmapLoadFile(scr, wPreferences.icon_back_file, NULL);
	if (pix) {
	    scr->icon_back_pixmap = pix->image;
	    if (pix->mask)
	      XFreePixmap(dpy, pix->mask);
	    wfree(pix);
	} else {
	    scr->icon_back_pixmap = None;
	}
    }
#ifdef DITHERING    
    if (scr->icon_back_image==NULL)
      scr->icon_back_image = makeIconBackImage(scr);
#endif    
    if (scr->icon_back_pixmap==None)
      scr->icon_back_pixmap = makeIconBackPixmap(scr);
    
    if (wResourceDefault.icon_file) {
	scr->default_icon = LoadIcon(scr, wResourceDefault.icon_file,
				     wResourceDefault.icon_mask_file, 0);
	   
	scr->default_titled_icon = LoadIcon(scr, wResourceDefault.icon_file, 
					    wResourceDefault.icon_mask_file, 
					    scr->icon_title_font.height);
    }
    /* make default flat icons if they could not be loaded */
    if (scr->default_icon==None)
      scr->default_icon = scr->icon_back_pixmap;
    
    if (scr->default_titled_icon == None) {
	int title_height = scr->icon_title_font.height;
	  
	scr->default_titled_icon = XCreatePixmap(dpy, scr->w_win, ICON_WIDTH, 
						 ICON_HEIGHT, scr->w_depth);
	XCopyArea(dpy, scr->icon_back_pixmap, scr->default_titled_icon,
		  scr->copy_gc, 0, 0, ICON_WIDTH, ICON_HEIGHT, 0, 0);
	XFillRectangle(dpy, scr->default_titled_icon,
		       scr->icon_title_texture->normal_gc, 0, 0, ICON_WIDTH, 
		       title_height+1);
	XDrawLine(dpy, scr->default_titled_icon,
		  scr->icon_title_texture->light_gc, 0, 0, ICON_WIDTH, 0);
	XDrawLine(dpy, scr->default_titled_icon, 
		  scr->icon_title_texture->light_gc, 0, 0, 0, 
		  title_height+1);
	XDrawLine(dpy, scr->default_titled_icon, 
		  scr->icon_title_texture->dim_gc, ICON_WIDTH-1, 0, 
		  ICON_WIDTH-1, title_height+1);
    }

#ifdef DOCK
    if (!wPreferences.flags.nodock) {
	scr->dock_dots = 
	  wPixmapLoadFile(scr, 
			  find_file(PixmapPath, wPreferences.icon_dots_file),
			  NULL);
    }
#endif
    /* titlebar button pixmaps */
    allocButtonPixmaps(scr);
}


/*
 *----------------------------------------------------------------------
 * createInternalWindows--
 * 	Creates some windows used internally by the program. One to
 * receive input focus when no other window can get it and another
 * to display window geometry information during window resize/move.
 * 
 * Returns:
 * 	Nothing
 * 
 * Side effects:
 * 	Windows are created and some colors are allocated for the
 * window background.
 *---------------------------------------------------------------------- 
 */
static void
createInternalWindows(WScreen *scr)
{
    int vmask;
    XSetWindowAttributes attribs;

    /* window for displaying geometry information during resizes and moves */
    vmask = CWBackPixmap|CWBackPixel|CWCursor|CWSaveUnder|CWOverrideRedirect;
    attribs.save_under = True;
    attribs.override_redirect = True;    
    attribs.cursor = wCursor[WCUR_DEFAULT];
    attribs.background_pixmap = None;
    attribs.background_pixel = scr->unfocused_texture->any.color.pixel;
    scr->geometry_display =
      XCreateWindow(dpy, scr->root_win, (scr->scr_width-80)/2, 
		    (scr->scr_height-20)/2, 80, 20, 1, scr->depth, 
		    CopyFromParent, scr->visual, vmask, &attribs);
    
    /* InputOnly window to get the focus when no other window can get it */
    vmask = CWEventMask|CWOverrideRedirect;
    attribs.event_mask = KeyPressMask|FocusChangeMask;
    attribs.override_redirect = True;
    scr->no_focus_win=XCreateWindow(dpy,scr->root_win,-10, -10, 4, 4, 0, 0,
				    InputOnly,CopyFromParent, vmask, &attribs);
    XSelectInput(dpy, scr->no_focus_win, KeyPressMask|KeyReleaseMask);
    XMapWindow(dpy, scr->no_focus_win);

    XSetInputFocus (dpy, scr->no_focus_win, RevertToParent, CurrentTime);
 
#ifdef DOCK
    /* shadow window for dock buttons */
    vmask = CWBackPixmap|CWBackPixel|CWCursor|CWSaveUnder|CWOverrideRedirect;
    attribs.save_under = True;
    attribs.override_redirect = True;
    attribs.background_pixmap = None;
    attribs.background_pixel = scr->white_pixel;
    scr->dock_shadow =
      XCreateWindow(dpy, scr->root_win, 0, 0, ICON_WIDTH, ICON_HEIGHT, 0,
		    scr->depth, CopyFromParent, scr->visual, vmask, &attribs);
#endif
}


/*
 *----------------------------------------------------------------------
 * wScreenInit--
 * 	Initializes the window manager for the given screen and 
 * allocates a WScreen descriptor for it. Many resources are allocated
 * for the screen and the root window is setup appropriately.
 *
 * Returns:
 * 	The WScreen descriptor for the screen.
 * 
 * Side effects:
 * 	Many resources are allocated and the IconSize property is
 * set on the root window.
 *	The program can be aborted if some fatal error occurs.
 * 
 * TODO: User specifiable visual.
 *---------------------------------------------------------------------- 
 */
WScreen *
wScreenInit(int screen_number)
{
    WScreen *scr;
    XIconSize icon_size[1];
    /*
    char buffer[16];
     */
    
    scr = wmalloc(sizeof(WScreen));
    memset(scr, 0, sizeof(WScreen));

    /* initialize globals */
    scr->screen = screen_number;
    scr->root_win = RootWindow(dpy, screen_number);
    scr->visual = DefaultVisual(dpy, screen_number);    
    scr->depth = DefaultDepth(dpy, screen_number);
    scr->colormap = DefaultColormap(dpy, screen_number);

    /* change the following to user settable */
    scr->w_win = scr->root_win;
    scr->w_visual = scr->visual;
    scr->w_depth = scr->depth;

    scr->black_pixel = BlackPixel(dpy, screen_number);
    scr->white_pixel = WhitePixel(dpy, screen_number);
    
    scr->w_colormap = scr->colormap;

    scr->scr_width = WidthOfScreen(ScreenOfDisplay(dpy, screen_number));
    scr->scr_height = HeightOfScreen(ScreenOfDisplay(dpy, screen_number));

    XSetErrorHandler((XErrorHandler)alreadyRunningError);
    XSelectInput(dpy, scr->root_win, EVENT_MASK);
    XSync(dpy, 0);
    XSetErrorHandler((XErrorHandler)NULL);

    XDefineCursor(dpy, scr->root_win, wCursor[WCUR_DEFAULT]);
#ifdef DITHERING
    /* screen descriptor for raster graphic library */
    if (wPreferences.cmap.red*wPreferences.cmap.blue
	*wPreferences.cmap.green>(1<<scr->w_depth)) {
	wPreferences.cmap.red = 1<<(scr->w_depth/3);
	wPreferences.cmap.green = 1<<(scr->w_depth/3);
	wPreferences.cmap.blue = 1<<(scr->w_depth/3);
	wWarning("colormap size value(s) too large");
    }
    if (wPreferences.cmap.red<2)
      wPreferences.cmap.red=2;
    if (wPreferences.cmap.green<2)
      wPreferences.cmap.green=2;
    if (wPreferences.cmap.blue<2)
      wPreferences.cmap.blue=2;
    scr->rscreen = wInitRaster(scr, wPreferences.cmap.red,
			       wPreferences.cmap.green, 
			       wPreferences.cmap.blue);
#else
    scr->rscreen = NULL;
#endif /* DITHERING */  
    /* allocate shared resources */
    allocResources(scr);
    createInternalWindows(scr);
    /* set icon sizes we can accept */
    icon_size[0].min_width = 1;
    icon_size[0].min_height = 1;
    icon_size[0].max_width = ICON_WIDTH;
    icon_size[0].max_height = ICON_HEIGHT;
    icon_size[0].width_inc = 1;
    icon_size[0].height_inc = 1;
    XSetIconSizes(dpy, scr->root_win, icon_size, 1);

    /* setup window manager selection stuff */
    /*
    sprintf(buffer, "SM_S%i", screen_number);
    scr->selection = XInternAtom(dpy, buffer, False);
    if (XGetSelectionOwner(dpy, scr->selection)!=None) {
	wWarning("taking over selection ownership for %s", buffer);
    }
    XSetSelectionOwner(dpy, scr->selection, scr->selection_owner, CurrentTime);
    */
    
    /* setup WindowMaker protocols property in the root window*/
    PropSetWMakerProtocols(scr->root_win);
    
    return scr;
}

