/* setup.c - dynamic setup module interface
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "wconfig.h"
#include "generic/wwmlib.h"

#ifdef DYNSETUP

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>

#include "screen.h"
#include "raster.h"
#include "window.h"
#include "setup.h"
#include "version.h"


/**** Global Variables ****/
extern WPreferences wPreferences;


static int
loadFont(WFont *font, char *font_name)
{
    XFontStruct *new;
    
    new = XLoadQueryFont(dpy, font_name);
    if (!new) {
	return 0;
    }
    XFreeFont(dpy, font->font);
    font->font = new;
    font->height = new->ascent+new->descent;
    font->y = new->ascent;
    return 1;
}


static void
realizeMenus(WMenu *menu)
{
    int i;

    wMenuRealize(menu);
    for (i=0; i<menu->entry_no; i++) {
	if (menu->entries[i]->cascade)
	  realizeMenus(menu->entries[i]->cascade);
    }
}


void
wSetupOpen(WScreen *scr, char *prog)
{
    pid_t pid;
    int inp[2], outp[2];

    if (scr->setup_pipe_out) 
      return;

    if (pipe(inp)<0 || pipe(outp)<0) {
	wSysError("could not open pipe for setup panel");
	return;
    }
    scr->setup_pipe_in = inp[0];
    scr->setup_pipe_out = outp[1];

    pid = fork();
    if (pid<0) {
	wSysError("could not fork setup panel process");
	return;
    } else if (pid==0) {
	char *argv[2];
	
	argv[0] = prog;
	argv[1] = NULL;
	close(inp[0]);
	close(outp[1]);
	close(0);
	dup(outp[0]);
	close(1);
	dup(inp[1]);
	if (execvp(prog, argv)<0) {
	    wSysError("could not exec setup panel:%s", prog);
	    exit(1);
	}
    } else {
	close(inp[1]);
	close(outp[0]);
    }
    scr->setup_panel_pid = pid;
}


void
wSetupClose(WScreen *scr)
{
    if (scr->setup_pipe_out) {
	close(scr->setup_pipe_in);
	close(scr->setup_pipe_out);
	scr->setup_pipe_out = 0;
	kill(scr->setup_panel_pid, SIGTERM);
    }
}
	    
static int
resetFont(WScreen *scr, int font, char *font_name)
{
    int success = 0;

    switch (font) {
     case WDS_WINDOW_TITLE_FONT:
	if (loadFont(&scr->title_font, font_name)) {
	    XSetFont(dpy, scr->unfocused_text_gc, scr->title_font.font->fid);
	    XSetFont(dpy, scr->focused_text_gc, scr->title_font.font->fid);
	    if (wPreferences.title_font)
	      free(wPreferences.title_font);
	    wPreferences.title_font = strdup(font_name);
	    success = 1;
	}
	break;
	
     case WDS_MENU_TITLE_FONT:
	if (loadFont(&scr->menu_title_font, font_name)) {
	    XSetFont(dpy, scr->menu_title_gc, scr->menu_title_font.font->fid);
	    if (wPreferences.menu_title_font)
	      free(wPreferences.menu_title_font);
	    wPreferences.menu_title_font = strdup(font_name);
	    success = 1;
	}
	break;
	
     case WDS_MENU_ENTRY_FONT:
	if (loadFont(&scr->menu_entry_font, font_name)) {
	    XSetFont(dpy, scr->menu_entry_gc, scr->menu_entry_font.font->fid);
	    XSetFont(dpy, scr->select_menu_gc, scr->menu_entry_font.font->fid);
	    XSetFont(dpy, scr->disabled_menu_entry_gc, 
		     scr->menu_entry_font.font->fid);
	    if (wPreferences.menu_entry_font)
	      free(wPreferences.menu_entry_font);
	    wPreferences.menu_entry_font = strdup(font_name);
	    success = 1;
	}
	break;
    }
    return success;
}


static int
resetColor(WScreen *scr, int color, char *color_name)
{
    int success=0;
    XColor col;

    if (XParseColor(dpy, scr->colormap, color_name, &col)) {
	success = 1;
	AllocColor(scr, &col);
	switch (color) {
	 case WDS_FTITLE_FORE_COLOR:
	    if (wPreferences.ftitle_fcolor)
	      free(wPreferences.ftitle_fcolor);
	    wPreferences.ftitle_fcolor = strdup(color_name);
	    FreeColor(scr, scr->ftitle_pixel);
	    scr->ftitle_pixel = col.pixel;
	    XSetForeground(dpy, scr->focused_text_gc, col.pixel);
	    break;
	    
	 case WDS_UTITLE_FORE_COLOR:
	    if (wPreferences.utitle_fcolor)
	      free(wPreferences.utitle_fcolor);
	    wPreferences.utitle_fcolor = strdup(color_name);
	    FreeColor(scr, scr->utitle_pixel);
	    scr->utitle_pixel = col.pixel;
	    XSetForeground(dpy, scr->unfocused_text_gc, col.pixel);
	    break;
	    
	 case WDS_MENU_TITLE_FORE_COLOR:
	    if (wPreferences.menu_title_fcolor)
	      free(wPreferences.menu_title_fcolor);
	    wPreferences.menu_title_fcolor = strdup(color_name);
	    FreeColor(scr, scr->mtitle_pixel);
	    scr->mtitle_pixel = col.pixel;
	    XSetForeground(dpy, scr->menu_title_gc, col.pixel);
	    break;
	    
	 case WDS_MENU_TEXT_DISABLED_COLOR:
	    if (wPreferences.menu_text_dcolor)
	      free(wPreferences.menu_text_dcolor);
	    wPreferences.menu_text_dcolor = strdup(color_name);
	    FreeColor(scr, scr->dtext_pixel);
	    scr->dtext_pixel = col.pixel;
	    XSetForeground(dpy, scr->disabled_menu_entry_gc, col.pixel);
	    break;
	    
	 case WDS_MENU_TEXT_FORE_COLOR:
	    if (wPreferences.menu_text_fcolor)
	      free(wPreferences.menu_text_fcolor);
	    wPreferences.menu_text_fcolor = strdup(color_name);
	    FreeColor(scr, scr->mtext_pixel);
	    scr->mtext_pixel = col.pixel;
	    XSetForeground(dpy, scr->menu_entry_gc, col.pixel);
	    break;
	}
    }
    return success;
}



void
wSetupCheck(WScreen *scr)
{
    int code, count;
    int msgsize;
    char param[MAXLINE], msg[MAXLINE];
    fd_set fds;
    struct timeval timeout;
    static ch_window=0, ch_menu=0;

    if (scr->setup_pipe_out==0) {
	return;
    }

    FD_ZERO(&fds);
    FD_SET(scr->setup_pipe_in, &fds);
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    if (select(scr->setup_pipe_in+1, &fds, NULL, NULL, &timeout)<=0) {
	return;
    }
    msgsize = read(scr->setup_pipe_in, msg, MAXLINE);
    if (msgsize<0) {
	wSetupClose(scr);
	wSysError("error reading data from setup panel");
	return;
    }
    if (msgsize==0) {
#ifdef DEBUG
	printf("read() from setup panel returned 0");
#endif
	wSetupClose(scr);
	return;
    }
    msg[msgsize] = 0;
#ifdef DEBUG
    printf("Received message: %s\n",msg);
#endif
    count = sscanf(msg, "%i %s", &code, param);
    if (count==1) {
	switch (code) {
	 case WDS_SEND_VERSION:
	    sprintf(msg, "%i.%i.%i\n", MAJ_VERSION, MIN_VERSION,SUBMIN_VERSION);
	    if (write(scr->setup_pipe_out, msg, strlen(msg))<0) {
		wSetupClose(scr);
		wSysError("error writing data to setup panel");
		return;
	    }
	    break;
	    
	 case WDS_REPAINT:
	    if (ch_menu) {
		realizeMenus(scr->root_menu);
		ch_menu =  0;
	    } 
	    if (ch_window) {
		WWindow *tmp;
		tmp = scr->focused_window;
		while (tmp) {
		    wWindowConfigureBorders(tmp);
		    XMoveWindow(dpy, tmp->client_win, 0, tmp->top_width);
		    tmp = tmp->prev;
		}
		ch_window = 0;
	    }
	    sprintf(msg, "%i\n", WDS_REQUEST_SUCCEED);
	    write(scr->setup_pipe_out, msg, strlen(msg));
	    break;

	 case WDS_SEND_FONTS: {
	     char *tfont, *mfont, *efont;

	     if (wPreferences.title_font)
	       tfont = wPreferences.title_font;
	     else
	       tfont = DEF_TITLE_FONT;
	     
	     if (wPreferences.menu_title_font)
	       mfont = wPreferences.menu_title_font;
	     else
	       mfont = DEF_MENU_TITLE_FONT;
	     
	     if (wPreferences.menu_entry_font)
	       efont = wPreferences.menu_entry_font;
	     else
	       efont = DEF_MENU_ENTRY_FONT;
	     
	     sprintf(msg, "%i %s\n%i %s\n%i %s\n",
		     WDS_WINDOW_TITLE_FONT, tfont,
		     WDS_MENU_TITLE_FONT, mfont,
		     WDS_MENU_ENTRY_FONT, efont);
	 
	    if (write(scr->setup_pipe_out, msg, strlen(msg))<0) {
		wSetupClose(scr);
		wSysError("error writing data to setup panel");
		return;
	    }
	 }
	    break;

	 case WDS_SEND_FORE_COLORS: {
	     char *ft, *ut, *mt, *mi, *di;

	     if (wPreferences.utitle_fcolor)
	       ut = wPreferences.utitle_fcolor;
	     else
	       ut = DEF_UNFOCUSED_TEXT_COLOR;

	     if (wPreferences.ftitle_fcolor)
	       ft = wPreferences.ftitle_fcolor;
	     else
	       ft = DEF_FOCUSED_TEXT_COLOR;

	     if (wPreferences.menu_title_fcolor)
	       mt = wPreferences.menu_title_fcolor;
	     else
	       mt = DEF_FOCUSED_TEXT_COLOR;

	     if (wPreferences.menu_text_fcolor)
	       mi = wPreferences.menu_text_fcolor;
	     else
	       mi = DEF_UNFOCUSED_TEXT_COLOR;

	     if (wPreferences.menu_text_dcolor)
	       di = wPreferences.menu_text_dcolor;
	     else
	       di = DEF_DISABLED_TEXT_COLOR;

	     sprintf(msg, "%i %s\n%i %s\n%i %s\n%i %s\n%i %s\n",
		     WDS_FTITLE_FORE_COLOR, ft,
		     WDS_UTITLE_FORE_COLOR, ut,
		     WDS_MENU_TITLE_FORE_COLOR, mt,
		     WDS_MENU_TEXT_FORE_COLOR, mi,
		     WDS_MENU_TEXT_DISABLED_COLOR, di);
	 
	    if (write(scr->setup_pipe_out, msg, strlen(msg))<0) {
		wSetupClose(scr);
		wSysError("error writing data to setup panel");
		return;
	    }
	 }
	    break;

	 case WDS_CLOSEDOWN:
	    wSetupClose(scr);
	    return;

	 default:
	    wWarning("received invalid request from setup panel");
	    return;
	}
    } else if (count==2) {
	if (code>=WDS_WINDOW_TITLE_FONT && code<=WDS_MENU_ENTRY_FONT) {
	    if (resetFont(scr, code, param)) {
		sprintf(msg, "%i\n", WDS_REQUEST_SUCCEED);
		if (code==WDS_WINDOW_TITLE_FONT)
		  ch_window=1;
		else
		  ch_menu=1;
	    } else {
		sprintf(msg, "%i\n", WDS_REQUEST_FAILED);
	    }
	    write(scr->setup_pipe_out, msg, strlen(msg));
	} else if (code>=WDS_FTITLE_FORE_COLOR 
		   && code<=WDS_MENU_TEXT_DISABLED_COLOR) {
	    if (resetColor(scr, code, param)) {
		sprintf(msg, "%i\n", WDS_REQUEST_SUCCEED);
		if (code<=WDS_UTITLE_FORE_COLOR)
		  ch_window=1;
		else
		  ch_menu=1;
	    } else {
		sprintf(msg, "%i\n", WDS_REQUEST_FAILED);
	    }
	    write(scr->setup_pipe_out, msg, strlen(msg));
	} else {
	    sprintf(msg, "%i\n", WDS_BAD_REQUEST);
	    write(scr->setup_pipe_out, msg, strlen(msg));
	}
    } else {
	wWarning("got invalid message from setup panel");
	return;
    } 
}


#endif /* DYNSETUP */
