/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "screen.h"
#include "window.h"
#include "funcs.h"
#include "actions.h"
#include "properties.h"
#include "stacking.h"

/*** Global Variables ***/
extern XContext wStackContext;




static int
levelToIndex(int level)
{
    switch (level) {
     case NSNormalWindowLevel:
	return 0;
     case NSFloatingWindowLevel:
	return 1;
     case NSDockWindowLevel:
	return 2;
     case NSSubmenuWindowLevel:
	return 3;
     case NSMainMenuWindowLevel:
	return 4;
     default:
	return 0;
    }
}

/*
 *---------------------------------------------------------------------- 
 * RemakeStackList--
 * 	Remakes the stacking_list for the screen, getting the real
 * stacking order from the server and reordering windows that are not
 * in the correct stacking.
 * 
 * Side effects:
 * 	The stacking order list and the actual window stacking 
 * may be changed (corrected)
 * 
 *---------------------------------------------------------------------- 
 */
#ifdef BUGGYSTACKING
void
RemakeStackList(WScreen *scr)
{
    Window *windows;
    unsigned int nwindows;
    Window junkr, junkp;
    WCoreWindow *frame;
    WCoreWindow *onbotw[MAX_WINDOW_LEVELS], *last[MAX_WINDOW_LEVELS];
    int level;
    int i, c;
    
    if (!XQueryTree(dpy, scr->root_win, &junkr, &junkp, &windows, &nwindows)) {
	wWarning("could not get window list!!"); 
	return;
    } else {
	c=0;
	for (i=0; i<MAX_WINDOW_LEVELS; i++) {
	    onbotw[i] = NULL;
	    last[i] = NULL;
	}
	/* this ensures the correctness of the window level order
	 * by using separate lists for each window level and then
	 * concatenating all lists together */
	for (i=0; i<nwindows; i++) {
	    if (XFindContext(dpy, windows[i], wStackContext, (XPointer*)&frame)
		==XCNOENT) {
		continue;
	    }
	    if (!frame) continue;
	    c++;
	    level = levelToIndex(frame->stacking->window_level);
	    if (onbotw[level])
	      onbotw[level]->stacking->above = frame;
	    else
	      last[level] = frame;
	    frame->stacking->under = onbotw[level];
	    frame->stacking->above = NULL;
	    onbotw[level] = frame;
	}
#ifdef DEBUG	
	if (c!=scr->window_count) {
	    puts("Found different number of windows than in window lists!!!");
	    scr->window_count = c;
	}
#endif
    }
    scr->stacking_list = NULL;
    /* now, just concatenate the lists */
    for (i=0; i<MAX_WINDOW_LEVELS; i++) {
	if (last[i]) {
	    last[i]->stacking->under = scr->stacking_list;
	    if (scr->stacking_list)
	      scr->stacking_list->stacking->above = last[i];
	    scr->stacking_list = onbotw[i];
	}
    }
    CommitStacking(scr);
}
#else /* BUGGYSTACKING */
void
RemakeStackList(WScreen *scr)
{
    Window *windows;
    unsigned int nwindows;
    Window junkr, junkp;
    WCoreWindow *frame;
    WCoreWindow *onbotw[MAX_WINDOW_LEVELS], *last[MAX_WINDOW_LEVELS];
    int level;
    int i, c;

    if (!XQueryTree(dpy, scr->root_win, &junkr, &junkp, &windows, &nwindows)) {
	wWarning("could not get window list!!"); 
	return;
    } else {
	for (i=0; i<MAX_WINDOW_LEVELS; i++) {
	    scr->stacking_list[i] = NULL;
	    onbotw[i] = NULL;
	    last[i] = NULL;
	    scr->window_level_count[i] = 0;
	}
	/* this verifies the integrity of the window level order
	 * and on_top window order, by placing !on_top windows in
	 * a separate list and then concatenating both lists */
	c=0;
	for (i=0; i<nwindows; i++) {
	    if (XFindContext(dpy, windows[i], wStackContext, (XPointer*)&frame)
		==XCNOENT) {
		continue;
	    }
	    if (!frame) continue;
	    c++;
	    level = levelToIndex(frame->stacking->window_level);
	    if (onbotw[level])
	      onbotw[level]->stacking->above = frame;
	    frame->stacking->under = onbotw[level];
	    frame->stacking->above = NULL;
	    onbotw[level] = frame;
	    scr->window_level_count[level]++;
	}
#ifdef DEBUG0
	if (c!=scr->window_count) {
	    puts("Found different number of windows than in window lists!!!");
	}
#endif
	scr->window_count = c;
    }
    /* now, just concatenate the lists */
    for (i=0; i<MAX_WINDOW_LEVELS; i++) {
	if (last[i]) {
	    /* there was at least one on_top window */
	    last[i]->stacking->under = onbotw[i];
	    if (onbotw[i])
	      onbotw[i]->stacking->above = last[i];
	} else {
	    scr->stacking_list[i] = onbotw[i];
	    if (onbotw[i])
	      onbotw[i]->stacking->above = NULL;
	}
    }
    CommitStacking(scr);
}
#endif /* !BUGGYSTACKING */

/*
 *----------------------------------------------------------------------
 * CommitStacking--
 * 	Reorders the actual window stacking, so that it has the stacking
 * order in the internal window stacking lists. It does the opposite
 * of RemakeStackList().
 * 
 * Side effects:
 * 	Windows may be restacked.
 *---------------------------------------------------------------------- 
 */
#ifdef BUGGYSTACKING 
void
CommitStacking(WScreen *scr)
{
    WCoreWindow *tmp;
    int nwindows;
    Window *windows;
    int i;

    nwindows = scr->window_count;
    windows = wmalloc(sizeof(Window)*nwindows);
    i=0;
    tmp = scr->stacking_list;
    while (tmp) {
	if (i>=nwindows) {
	    wWarning("Internal inconsistency! window_count incorrect.");
	    wfree(windows);
	    return;
	}
	windows[i++] = tmp->window;
	tmp = tmp->stacking->under;
    }
    XRestackWindows(dpy, windows, i);
    wfree(windows);
}
#else /* BUGGYSTACKING */
void
CommitStacking(WScreen *scr)
{
    WCoreWindow *tmp;
    int nwindows;
    Window *windows;
    int i, level;
    
    nwindows = scr->window_count;
    windows = wmalloc(sizeof(Window)*nwindows);
    i=0;
    for (level=MAX_WINDOW_LEVELS-1; level>=0; level--) {
	tmp = scr->stacking_list[level];
	while (tmp) {
#ifndef DEBUG
	    if (i>=nwindows) {
		puts("Internal inconsistency! window_count is incorrect!!!");
		printf("window_count says %i windows\n", nwindows);
		wfree(windows);
		return;
	    }
#endif
	    windows[i++] = tmp->window;
	    tmp = tmp->stacking->under;
	}
    }
    XRestackWindows(dpy, windows, i);
    wfree(windows);
}
#endif /* !BUGGYSTACKING */


#ifndef BUGGYSTACKING
/*
 *----------------------------------------------------------------------
 * moveFrameToUnder--
 * 	Reestacks windows so that "frame" is under "under".
 * 
 * Returns:
 *	None
 * 
 * Side effects:
 * 	Changes the stacking order of frame.
 *----------------------------------------------------------------------
 */
static void
moveFrameToUnder(WCoreWindow *under, WCoreWindow *frame)
{
    Window wins[2];

    wins[0] = under->window;
    wins[1] = frame->window;
    XRestackWindows(dpy, wins, 2);
}
#endif /* !BUGGYSTACKING */

/*
 *----------------------------------------------------------------------
 * wRaiseFrame--
 * 	Raises a frame taking the window level and the on_top flag
 * into account. 
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	Window stacking order and window list is changed.
 * 
 *---------------------------------------------------------------------- 
 */
#ifdef BUGGYSTACKING
void 
wRaiseFrame(WCoreWindow *frame)
{
    WScreen *scr=frame->screen_ptr;
    WCoreWindow *wlist=frame;

    /* already on top */
    if (frame->stacking->above == NULL
	|| (frame->stacking->window_level
	    !=wlist->stacking->above->stacking->window_level)) {
	return;
    }

    while (wlist && (frame->stacking->window_level
		     >= wlist->stacking->window_level)) {
	wlist = wlist->stacking->above;
    }
    /* window is inserted before the point */
    if (wlist==NULL) {
	/* topmost window (last on the list) */
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;

	frame->stacking->above = NULL;
	frame->stacking->under = scr->stacking_list;
	scr->stacking_list->stacking->above=frame;
	scr->stacking_list = frame;
    } else if (frame != wlist) {
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
	
	frame->stacking->above = wlist;
	frame->stacking->under = wlist->stacking->under;
	if (wlist->stacking->under)
	  wlist->stacking->under->stacking->above = frame;
	wlist->stacking->under = frame;
    }
    /* try to optimize things a little */
    if (frame->stacking->above == NULL) {
	XRaiseWindow(dpy, frame->window);
    } else {
	CommitStacking(frame->screen_ptr);
    }
}
#else /* BUGGYSTACKING */
void 
wRaiseFrame(WCoreWindow *frame)
{
    WCoreWindow *wlist=frame;
    int level = levelToIndex(frame->stacking->window_level);
    int i;

    /* already on top */
    if (frame->stacking->above == NULL) {
	return;
    }

    /* insert on top of other windows */
    while (wlist)
      wlist = wlist->stacking->above;
    
    /* window is inserted before the point found */
    if (wlist==NULL) {
	/* top most window (last on the list) */
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
	
	frame->stacking->above = NULL;
	frame->stacking->under = frame->screen_ptr->stacking_list[level];
	frame->screen_ptr->stacking_list[level]->stacking->above=frame;
	frame->screen_ptr->stacking_list[level] = frame;
    } else if (frame!=wlist) {
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
	
	frame->stacking->above = wlist;
	frame->stacking->under = wlist->stacking->under;
	if (wlist->stacking->under)
	  wlist->stacking->under->stacking->above = frame;
	wlist->stacking->under = frame;
    }
    /* try to optimize things a little */
    if (frame->stacking->above == NULL) {
	WCoreWindow *above=NULL;

	for (i=level+1; i<MAX_WINDOW_LEVELS; i++) {
	    if (frame->screen_ptr->stacking_list[i]!=NULL) {
		/* can't optimize */
		above = frame->screen_ptr->stacking_list[i];
		while (above->stacking->under)
		  above = above->stacking->under;
		break;
	    }
	}
	if (!above) {
	    XRaiseWindow(dpy, frame->window);
	} else {
	    moveFrameToUnder(above, frame);
	}
    } else {
	moveFrameToUnder(frame->stacking->above, frame);
    }
}
#endif /* !BUGGYSTACKING */

void
wRaiseLowerFrame(WCoreWindow *frame)
{
    if (!frame->stacking->above
	||(frame->stacking->window_level
	   !=frame->stacking->above->stacking->window_level))
      wLowerFrame(frame);
    else
      wRaiseFrame(frame);
}


#ifdef BUGGYSTACKING
void 
wLowerFrame(WCoreWindow *frame)
{
    WScreen *scr=frame->screen_ptr;
    WCoreWindow *prev, *wlist=frame;

    /* already in bottom */
    if (wlist->stacking->under==NULL
	|| (wlist->stacking->under->stacking->window_level
	    !=frame->stacking->window_level)) {
	return;
    }
    
    /* it was the top window */
    if (scr->stacking_list == frame) {
	scr->stacking_list = frame->stacking->under;
	scr->stacking_list->stacking->above = NULL;
    } else {
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
    }
    wlist = scr->stacking_list;
    prev = wlist;
    while (wlist && (frame->stacking->window_level
		     <= wlist->stacking->window_level)
	   && frame->stacking->child_of!=wlist) {
	prev = wlist;
	wlist = wlist->stacking->under;
    }
    frame->stacking->above = prev;
    frame->stacking->under = prev->stacking->under;
    if (prev->stacking->under)
      prev->stacking->under->stacking->above = frame;
    prev->stacking->under = frame;

    /* try to optimize things a little */
    if (frame->stacking->under == NULL) {
	XLowerWindow(dpy, frame->window);
    } else {
	CommitStacking(scr);
    }
}
#else /* BUGGYSTACKING */
void 
wLowerFrame(WCoreWindow *frame)
{
    WScreen *scr=frame->screen_ptr;
    WCoreWindow *prev, *wlist=frame;
    int level = levelToIndex(frame->stacking->window_level);
    int i;

    /* already in bottom */
    if (wlist->stacking->under==NULL) {
	return;
    }
    
    prev = wlist;
    /* it was the top window */
    if (scr->stacking_list[level] == frame) {
	scr->stacking_list[level] = frame->stacking->under;
	scr->stacking_list[level]->stacking->above = NULL;
    } else {
	if (frame->stacking->under)
	  frame->stacking->under->stacking->above = frame->stacking->above;
	if (frame->stacking->above)
	  frame->stacking->above->stacking->under = frame->stacking->under;
    }
    wlist = scr->stacking_list[level];
    while (wlist->stacking->under) {
	wlist = wlist->stacking->under;
    }
    frame->stacking->above = wlist;
    frame->stacking->under = NULL;
    wlist->stacking->under = frame;

    /* try to optimize things a little */
    if (frame->stacking->above == NULL) {
	WCoreWindow *above = NULL;
	for (i=level-1; i>=0; i--) {
	    if (scr->stacking_list[i]!=NULL) {
		/* can't optimize */
		above = scr->stacking_list[i];
		while (above->stacking->under)
		  above = above->stacking->under;
		break;
	    }
	}
	if (!above) {
	    XLowerWindow(dpy, frame->window);
	} else {
	    moveFrameToUnder(above, frame);
	}
    } else {
	moveFrameToUnder(frame->stacking->above, frame);
    }
}
#endif /* !BUGGYSTACKING */


/*
 *----------------------------------------------------------------------
 * AddToStackList--
 * 	Inserts the frame in the top of the stacking list. The
 * stacking precedence is obeyed.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	The frame is added to it's screen's window list.
 *---------------------------------------------------------------------- 
 */
#ifdef BUGGYSTACKING
void
AddToStackList(WCoreWindow *frame)
{
    WCoreWindow *prev, *tmpw;

    frame->screen_ptr->window_count++;
    tmpw = frame->screen_ptr->stacking_list;
    if (!tmpw) {
	frame->screen_ptr->stacking_list = frame;
	frame->stacking->above = NULL;
	frame->stacking->under = NULL;
	return;
    }
    prev = tmpw;
    while (tmpw && (frame->stacking->window_level
		    <tmpw->stacking->window_level)) {
	prev = tmpw;
	tmpw = tmpw->stacking->under;
    }
    /* add window after our place */
    if (tmpw == frame->screen_ptr->stacking_list) {
	/* last window on list (top) */
	tmpw->stacking->above = frame;
	frame->stacking->under = tmpw;
	frame->stacking->above = NULL;
	frame->screen_ptr->stacking_list = frame;
    } else if (tmpw == NULL) {
	/* first window on list (bottom) */
	prev->stacking->under = frame;
	frame->stacking->under = NULL;
	frame->stacking->above = prev;
    } else {
	/* other cases */
	frame->stacking->above = tmpw->stacking->above;
	frame->stacking->under = tmpw;
	tmpw->stacking->above->stacking->under = frame;
	tmpw->stacking->above = frame;
    }
    CommitStacking(frame->screen_ptr);
}
#else /* BUGGYSTACKING */
void
AddToStackList(WCoreWindow *frame)
{
    WCoreWindow *prev, *tmpw;
    int index = levelToIndex(frame->stacking->window_level);

    frame->screen_ptr->window_count++;
    frame->screen_ptr->window_level_count[index]++;
    XSaveContext(dpy, frame->window, wStackContext, (XPointer)frame);
    tmpw = frame->screen_ptr->stacking_list[index];
    if (!tmpw) {
	frame->screen_ptr->stacking_list[index] = frame;
	frame->stacking->above = NULL;
	frame->stacking->under = NULL;
	CommitStacking(frame->screen_ptr);	
	return;
    }
    prev = tmpw;
    /* put on top of the stacking list */
    frame->stacking->above = NULL;
    frame->stacking->under = tmpw;
    tmpw->stacking->above = frame;
    frame->screen_ptr->stacking_list[index] = frame;
    CommitStacking(frame->screen_ptr);
}
#endif /* !BUGGYSTACKING */

#ifndef BUGGYSTACKING
void
RemoveFromStackList(WCoreWindow *frame)
{
    int index = levelToIndex(frame->stacking->window_level);

    if (XDeleteContext(dpy, frame->window, wStackContext)==XCNOENT) {
#ifdef DEBUG0
	wWarning("RemoveFromStackingList(): window not in list ");
#endif
	return;
    }
    /* remove from the window stack list */
    if (frame->stacking->under)
      frame->stacking->under->stacking->above = frame->stacking->above;
    if (frame->stacking->above)
      frame->stacking->above->stacking->under = frame->stacking->under;
    else /* this was the first window on the list */
      frame->screen_ptr->stacking_list[index] = frame->stacking->under;

    frame->screen_ptr->window_count--;
    frame->screen_ptr->window_level_count[index]--;
}
#endif /* !BUGGYSTACKING */


void
ChangeStackingLevel(WCoreWindow *frame, int new_level)
{
    int old_level;
    if (frame->stacking->window_level == new_level)
      return;
    old_level = frame->stacking->window_level;

    RemoveFromStackList(frame);
    frame->stacking->window_level = new_level;
    AddToStackList(frame);
    if (old_level > new_level) {
	wRaiseFrame(frame);
    } else {
	wLowerFrame(frame);
    }
}
