/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdarg.h>
#include <string.h>
#include "gtkcontainer.h"
#include "gtkmain.h"
#include "gtkrc.h"
#include "gtksignal.h"
#include "gtkwidget.h"
#include "gtkwindow.h"


#define WIDGET_CLASS(w)  GTK_WIDGET_CLASS (GTK_OBJECT (w)->klass)


enum {
  SHOW,
  HIDE,
  MAP,
  UNMAP,
  REALIZE,
  UNREALIZE,
  DRAW,
  DRAW_FOCUS,
  DRAW_DEFAULT,
  SIZE_REQUEST,
  SIZE_ALLOCATE,
  STATE_CHANGED,
  INSTALL_ACCELERATOR,
  REMOVE_ACCELERATOR,
  EVENT,
  BUTTON_PRESS_EVENT,
  BUTTON_RELEASE_EVENT,
  MOTION_NOTIFY_EVENT,
  DELETE_EVENT,
  DESTROY_EVENT,
  EXPOSE_EVENT,
  KEY_PRESS_EVENT,
  KEY_RELEASE_EVENT,
  ENTER_NOTIFY_EVENT,
  LEAVE_NOTIFY_EVENT,
  CONFIGURE_EVENT,
  FOCUS_IN_EVENT,
  FOCUS_OUT_EVENT,
  MAP_EVENT,
  UNMAP_EVENT,
  PROPERTY_NOTIFY_EVENT,
  SELECTION_CLEAR_EVENT,
  SELECTION_REQUEST_EVENT,
  SELECTION_NOTIFY_EVENT,
  OTHER_EVENT,
  LAST_SIGNAL
};


typedef void (*GtkWidgetSignal1) (GtkObject *object,
				  gpointer   arg1,
				  gpointer   data);
typedef gint (*GtkWidgetSignal2) (GtkObject *object,
				  gpointer   arg1,
				  gchar      arg2,
				  gchar      arg3,
				  gpointer   data);
typedef void (*GtkWidgetSignal3) (GtkObject *object,
				  gpointer   arg1,
				  gpointer   data);
typedef gint (*GtkWidgetSignal4) (GtkObject *object,
				  gpointer   arg1,
				  gpointer   data);


static void gtk_widget_marshal_signal_1 (GtkObject      *object,
					 GtkSignalFunc   func,
					 gpointer        func_data,
					 GtkArg         *args);
static void gtk_widget_marshal_signal_2 (GtkObject      *object,
					 GtkSignalFunc   func,
					 gpointer        func_data,
					 GtkArg         *args);
static void gtk_widget_marshal_signal_3 (GtkObject      *object,
					 GtkSignalFunc   func,
					 gpointer        func_data,
					 GtkArg         *args);
static void gtk_widget_marshal_signal_4 (GtkObject      *object,
					 GtkSignalFunc   func,
					 gpointer        func_data,
					 GtkArg         *args);

static void gtk_widget_class_init                (GtkWidgetClass    *klass);
static void gtk_widget_init                      (GtkWidget         *widget);
static void gtk_widget_arg                       (GtkWidget         *widget,
						  GtkArg            *arg);
static void gtk_real_widget_destroy              (GtkObject         *object);
static void gtk_real_widget_show                 (GtkWidget         *widget);
static void gtk_real_widget_hide                 (GtkWidget         *widget);
static void gtk_real_widget_map                  (GtkWidget         *widget);
static void gtk_real_widget_unmap                (GtkWidget         *widget);
static void gtk_real_widget_realize              (GtkWidget         *widget);
static void gtk_real_widget_unrealize            (GtkWidget         *widget);
static void gtk_real_widget_draw                 (GtkWidget         *widget,
						  GdkRectangle      *area);
static void gtk_real_widget_size_allocate        (GtkWidget         *widget,
						  GtkAllocation     *allocation);

static GdkColormap* gtk_widget_peek_colormap (void);
static GdkVisual*   gtk_widget_peek_visual   (void);
static GtkStyle*    gtk_widget_peek_style    (void);

static void gtk_widget_set_parent_sensitive  (GtkWidget *widget,
					      gpointer   client_data);
static void gtk_widget_propagate_restore     (GtkWidget *widget,
					      gpointer   client_data);
static void gtk_widget_propagate_state       (GtkWidget *widget,
					      gpointer   client_data);
static void gtk_widget_draw_children_recurse (GtkWidget *widget,
					      gpointer   client_data);
static void gtk_widget_set_style_internal    (GtkWidget *widget,
					      GtkStyle  *style);
static void gtk_widget_set_style_recurse     (GtkWidget *widget,
					      gpointer   client_data);

extern GtkArg* gtk_object_collect_args (gint    *nargs,
					va_list  args1,
					va_list  args2);

static GtkWidgetAuxInfo* gtk_widget_aux_info_new     (void);
static void              gtk_widget_aux_info_destroy (GtkWidgetAuxInfo *aux_info);


static GtkObjectClass *parent_class = NULL;
static gint widget_signals[LAST_SIGNAL] = { 0 };

static GMemChunk *aux_info_mem_chunk = NULL;

static GdkColormap *default_colormap = NULL;
static GdkVisual *default_visual = NULL;
static GtkStyle *default_style = NULL;

static GSList *colormap_stack = NULL;
static GSList *visual_stack = NULL;
static GSList *style_stack = NULL;

static const char *aux_info_key = "aux_info";
static const char *colormap_key = "colormap";
static const char *visual_key = "visual";
static const char *event_key = "event_mask";



/*****************************************
 * gtk_widget_get_type:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

guint
gtk_widget_get_type ()
{
  static guint widget_type = 0;

  if (!widget_type)
    {
      GtkTypeInfo widget_info =
      {
	"GtkWidget",
	sizeof (GtkWidget),
	sizeof (GtkWidgetClass),
	(GtkClassInitFunc) gtk_widget_class_init,
	(GtkObjectInitFunc) gtk_widget_init,
	(GtkArgFunc) gtk_widget_arg,
      };

      widget_type = gtk_type_unique (gtk_object_get_type (), &widget_info);
    }

  return widget_type;
}

/*****************************************
 * gtk_widget_class_init:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_class_init (GtkWidgetClass *klass)
{
  GtkObjectClass *object_class;
  gint i;

  object_class = (GtkObjectClass*) klass;

  parent_class = gtk_type_class (gtk_object_get_type ());

  gtk_object_add_arg_type ("GtkWidget::x", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::y", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::width", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::height", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::visible", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::sensitive", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::events", GTK_ARG_INT);
  gtk_object_add_arg_type ("GtkWidget::name", GTK_ARG_POINTER);
  gtk_object_add_arg_type ("GtkWidget::style", GTK_ARG_POINTER);
  gtk_object_add_arg_type ("GtkWidget::parent", GTK_ARG_OBJECT);

  i = 0;
  widget_signals[i++] = gtk_signal_new ("show",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, show),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("hide",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, hide),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("map",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, map),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("unmap",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, unmap),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("realize",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, realize),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("unrealize",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, unrealize),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("draw",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, draw),
					gtk_widget_marshal_signal_1, GTK_ARG_NONE, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("draw_focus",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, draw_focus),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("draw_default",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, draw_default),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("size_request",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, size_request),
					gtk_widget_marshal_signal_1, GTK_ARG_NONE, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("size_allocate",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, size_allocate),
					gtk_widget_marshal_signal_1, GTK_ARG_NONE, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("state_changed",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, state_changed),
					gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  widget_signals[i++] = gtk_signal_new ("install_accelerator",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, install_accelerator),
					gtk_widget_marshal_signal_2, GTK_ARG_INT, 3,
					GTK_ARG_POINTER, GTK_ARG_CHAR, GTK_ARG_CHAR);
  widget_signals[i++] = gtk_signal_new ("remove_accelerator",
					GTK_RUN_FIRST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, remove_accelerator),
					gtk_widget_marshal_signal_3, GTK_ARG_NONE, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("button_press_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, button_press_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("button_release_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, button_release_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("motion_notify_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, motion_notify_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("delete_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, delete_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("destroy_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, destroy_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("expose_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, expose_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("key_press_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, key_press_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("key_release_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, key_release_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("enter_notify_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, enter_notify_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("leave_notify_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, leave_notify_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("configure_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, configure_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("focus_in_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, focus_in_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("focus_out_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, focus_out_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("map_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, map_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("unmap_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, unmap_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("property_notify_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, property_notify_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("selection_clear_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, selection_clear_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("selection_request_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, selection_request_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("selection_notify_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, selection_notify_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);
  widget_signals[i++] = gtk_signal_new ("other_event",
					GTK_RUN_LAST,
					object_class->type,
					GTK_SIGNAL_OFFSET (GtkWidgetClass, other_event),
					gtk_widget_marshal_signal_4, GTK_ARG_INT, 1,
					GTK_ARG_POINTER);

  gtk_object_class_add_signals (object_class, widget_signals, LAST_SIGNAL);

  object_class->destroy = gtk_real_widget_destroy;

  klass->activate_signal = 0;
  klass->show = gtk_real_widget_show;
  klass->hide = gtk_real_widget_hide;
  klass->map = gtk_real_widget_map;
  klass->unmap = gtk_real_widget_unmap;
  klass->realize = gtk_real_widget_realize;
  klass->unrealize = gtk_real_widget_unrealize;
  klass->draw = gtk_real_widget_draw;
  klass->draw_focus = NULL;
  klass->size_request = NULL;
  klass->size_allocate = gtk_real_widget_size_allocate;
  klass->state_changed = NULL;
  klass->install_accelerator = NULL;
  klass->remove_accelerator = NULL;
  klass->event = NULL;
  klass->button_press_event = NULL;
  klass->button_release_event = NULL;
  klass->motion_notify_event = NULL;
  klass->delete_event = NULL;
  klass->destroy_event = NULL;
  klass->expose_event = NULL;
  klass->key_press_event = NULL;
  klass->key_release_event = NULL;
  klass->enter_notify_event = NULL;
  klass->leave_notify_event = NULL;
  klass->configure_event = NULL;
  klass->focus_in_event = NULL;
  klass->focus_out_event = NULL;
  klass->map_event = NULL;
  klass->unmap_event = NULL;
  klass->property_notify_event = NULL;
  klass->selection_clear_event = NULL;
  klass->selection_request_event = NULL;
  klass->selection_notify_event = NULL;
  klass->other_event = NULL;
}

/*****************************************
 * gtk_widget_arg:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_arg (GtkWidget *widget,
		GtkArg    *arg)
{
  if (strcmp (arg->name, "x") == 0)
    {
      gtk_widget_set_uposition (widget, arg->d.i, -2);
    }
  else if (strcmp (arg->name, "y") == 0)
    {
      gtk_widget_set_uposition (widget, -2, arg->d.i);
    }
  else if (strcmp (arg->name, "width") == 0)
    {
      gtk_widget_set_usize (widget, arg->d.i, -1);
    }
  else if (strcmp (arg->name, "height") == 0)
    {
      gtk_widget_set_usize (widget, -1, arg->d.i);
    }
  else if (strcmp (arg->name, "visible") == 0)
    {
      if (arg->d.i)
	gtk_widget_show (widget);
      else
	gtk_widget_hide (widget);
    }
  else if (strcmp (arg->name, "sensitive") == 0)
    {
      gtk_widget_set_sensitive (widget, arg->d.i);
    }
  else if (strcmp (arg->name, "events") == 0)
    {
      gtk_widget_set_events (widget, arg->d.i);
    }
  else if (strcmp (arg->name, "name") == 0)
    {
      gtk_widget_set_name (widget, arg->d.p);
    }
  else if (strcmp (arg->name, "style") == 0)
    {
      gtk_widget_set_style (widget, arg->d.p);
    }
  else if (strcmp (arg->name, "parent") == 0)
    {
      gtk_container_add (GTK_CONTAINER (arg->d.o), widget);
    }
}

/*****************************************
 * gtk_widget_init:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_init (GtkWidget *widget)
{
  GdkColormap *colormap;
  GdkVisual *visual;

  GTK_OBJECT_FLAGS (widget) = GTK_SENSITIVE | GTK_PARENT_SENSITIVE;
  widget->state = GTK_STATE_NORMAL;
  widget->saved_state = GTK_STATE_NORMAL;
  widget->name = NULL;
  widget->requisition.width = 0;
  widget->requisition.height = 0;
  widget->allocation.x = -1;
  widget->allocation.y = -1;
  widget->allocation.width = 1;
  widget->allocation.height = 1;
  widget->window = NULL;
  widget->parent = NULL;

  widget->style = gtk_widget_peek_style ();
  gtk_style_ref (widget->style);

  colormap = gtk_widget_peek_colormap ();
  visual = gtk_widget_peek_visual ();

  if (colormap != gtk_widget_get_default_colormap ())
    gtk_object_set_data (GTK_OBJECT (widget), colormap_key, colormap);

  if (visual != gtk_widget_get_default_visual ())
    gtk_object_set_data (GTK_OBJECT (widget), visual_key, visual);
}

/*****************************************
 * gtk_widget_new:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkWidget*
gtk_widget_new (guint type,
		...)
{
  GtkObject *obj;
  GtkArg *args;
  gint nargs;
  va_list args1;
  va_list args2;

  g_return_val_if_fail (gtk_type_is_a (type, gtk_widget_get_type ()), NULL);

  obj = gtk_type_new (type);

  va_start (args1, type);
  va_start (args2, type);

  args = gtk_object_collect_args (&nargs, args1, args2);
  gtk_object_setv (obj, nargs, args);
  g_free (args);

  va_end (args1);
  va_end (args2);

  return GTK_WIDGET (obj);
}

/*****************************************
 * gtk_widget_newv:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkWidget*
gtk_widget_newv (guint   type,
		 gint    nargs,
		 GtkArg *args)
{
  g_return_val_if_fail (gtk_type_is_a (type, gtk_widget_get_type ()), NULL);

  return GTK_WIDGET (gtk_object_newv (type, nargs, args));
}

/*****************************************
 * gtk_widget_set:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set (GtkWidget *widget,
		...)
{
  GtkArg *args;
  gint nargs;
  va_list args1;
  va_list args2;

  g_return_if_fail (widget != NULL);

  va_start (args1, widget);
  va_start (args2, widget);

  args = gtk_object_collect_args (&nargs, args1, args2);
  gtk_object_setv (GTK_OBJECT (widget), nargs, args);
  g_free (args);

  va_end (args1);
  va_end (args2);
}

/*****************************************
 * gtk_widget_setv:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_setv (GtkWidget *widget,
		 gint       nargs,
		 GtkArg    *args)
{
  gtk_object_setv (GTK_OBJECT (widget), nargs, args);
}

/*****************************************
 * gtk_widget_destroy:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_destroy (GtkWidget *widget)
{
  if (widget->parent)
    {
      if (!GTK_OBJECT_BEING_DESTROYED (widget->parent))
	gtk_container_remove (GTK_CONTAINER (widget->parent), widget);
      else
	gtk_object_unref (GTK_OBJECT (widget));
    }
  gtk_object_destroy (GTK_OBJECT (widget));
}

/*****************************************
 * gtk_widget_unparent: do any cleanup necessary necessary before
 *                      setting parent = NULL. In particular, remove
 *                      the focus properly.
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_unparent (GtkWidget *widget)
{
  GtkWidget *toplevel;
  GtkWidget *child;

  g_return_if_fail (widget != NULL);

  toplevel = gtk_widget_get_toplevel (widget);
  if (GTK_IS_WINDOW (toplevel))
    {
      child = GTK_WINDOW (toplevel)->focus_widget;

      while (child && child != widget)
	child = child->parent;

      if (child == widget)
	gtk_window_set_focus (GTK_WINDOW (toplevel), NULL);
    }

  if (widget->window &&
      GTK_WIDGET_NO_WINDOW (widget) &&
      GTK_WIDGET_DRAWABLE (widget))
    gdk_window_clear_area (widget->window,
			   widget->allocation.x,
			   widget->allocation.y,
			   widget->allocation.width,
			   widget->allocation.height);

  widget->parent = NULL;

  gtk_object_unref (GTK_OBJECT (widget));
}

/*****************************************
 * gtk_widget_show:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_show (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (!GTK_WIDGET_VISIBLE (widget))
    gtk_signal_emit (GTK_OBJECT (widget), widget_signals[SHOW]);
}

/*****************************************
 * gtk_widget_hide:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_hide (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (GTK_WIDGET_VISIBLE (widget))
    gtk_signal_emit (GTK_OBJECT (widget), widget_signals[HIDE]);
}

/*****************************************
 * gtk_widget_map:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_map (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (!GTK_WIDGET_MAPPED (widget))
    {
      if (!GTK_WIDGET_REALIZED (widget))
	gtk_widget_realize (widget);

      gtk_signal_emit (GTK_OBJECT (widget), widget_signals[MAP]);
    }
}

/*****************************************
 * gtk_widget_unmap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_unmap (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (GTK_WIDGET_MAPPED (widget))
    gtk_signal_emit (GTK_OBJECT (widget), widget_signals[UNMAP]);
}

/*****************************************
 * gtk_widget_realize:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_realize (GtkWidget *widget)
{
  GtkStyle *new_style;

  g_return_if_fail (widget != NULL);

  if (!GTK_WIDGET_REALIZED (widget))
    {
      /*
      if (GTK_IS_CONTAINER (widget) && !GTK_WIDGET_NO_WINDOW (widget))
	g_print ("%s\n", gtk_type_name (GTK_WIDGET_TYPE (widget)));
	*/

      if (widget->parent && !GTK_WIDGET_REALIZED (widget->parent))
	gtk_widget_realize (widget->parent);

      if (!GTK_WIDGET_USER_STYLE (widget))
	{
	  new_style = gtk_rc_get_style (widget);
	  if (new_style != widget->style)
	    gtk_widget_set_style_internal (widget, new_style);
	}

      gtk_signal_emit (GTK_OBJECT (widget), widget_signals[REALIZE]);
    }
}

/*****************************************
 * gtk_widget_unrealize:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_unrealize (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (GTK_WIDGET_REALIZED (widget))
    gtk_signal_emit (GTK_OBJECT (widget), widget_signals[UNREALIZE]);
}

/*****************************************
 * gtk_widget_draw:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_draw (GtkWidget    *widget,
		 GdkRectangle *area)
{
  GdkRectangle temp_area;

  g_return_if_fail (widget != NULL);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      if (!area)
	{
	  if (GTK_WIDGET_NO_WINDOW (widget))
	    {
	      temp_area.x = widget->allocation.x;
	      temp_area.y = widget->allocation.y;
	    }
	  else
	    {
	      temp_area.x = 0;
	      temp_area.y = 0;
	    }

	  temp_area.width = widget->allocation.width;
	  temp_area.height = widget->allocation.height;
	  area = &temp_area;
	}

      gtk_signal_emit (GTK_OBJECT (widget), widget_signals[DRAW], area);
    }
}

/*****************************************
 * gtk_widget_draw_focus:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_draw_focus (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  gtk_signal_emit (GTK_OBJECT (widget), widget_signals[DRAW_FOCUS]);
}

/*****************************************
 * gtk_widget_draw_default:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_draw_default (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  gtk_signal_emit (GTK_OBJECT (widget), widget_signals[DRAW_DEFAULT]);
}

/*****************************************
 * gtk_widget_draw_children:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_draw_children (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  if (GTK_IS_CONTAINER (widget))
    gtk_container_foreach (GTK_CONTAINER (widget),
			   gtk_widget_draw_children_recurse,
			   NULL);
}

/*****************************************
 * gtk_widget_size_request:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_size_request (GtkWidget      *widget,
			 GtkRequisition *requisition)
{
  GtkWidgetAuxInfo *aux_info;

  g_return_if_fail (widget != NULL);

  if (gtk_signal_emit (GTK_OBJECT (widget), widget_signals[SIZE_REQUEST], requisition))
    {
      aux_info = gtk_object_get_data (GTK_OBJECT (widget), aux_info_key);
      if (aux_info)
	{
	  if (aux_info->width > 0)
	    requisition->width = aux_info->width;
	  if (aux_info->height > 0)
	    requisition->height = aux_info->height;
	}
    }
}

/*****************************************
 * gtk_widget_size_allocate:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_size_allocate (GtkWidget     *widget,
			  GtkAllocation *allocation)
{
  GtkWidgetAuxInfo *aux_info;
  GtkAllocation real_allocation;

  g_return_if_fail (widget != NULL);

  real_allocation = *allocation;
  aux_info = gtk_object_get_data (GTK_OBJECT (widget), aux_info_key);

  if (aux_info)
    {
      if (aux_info->x != -1)
	real_allocation.x = aux_info->x;
      if (aux_info->y != -1)
	real_allocation.y = aux_info->y;
    }

  gtk_signal_emit (GTK_OBJECT (widget), widget_signals[SIZE_ALLOCATE], allocation);
}

/*****************************************
 * gtk_widget_install_accelerator:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_install_accelerator (GtkWidget           *widget,
				GtkAcceleratorTable *table,
				gchar               *signal_name,
				gchar                key,
				guint8               modifiers)
{
  gint return_val;

  g_return_if_fail (widget != NULL);

  return_val = TRUE;
  if (gtk_signal_emit (GTK_OBJECT (widget), widget_signals[INSTALL_ACCELERATOR],
		       signal_name, key, modifiers, &return_val) && return_val)
    gtk_accelerator_table_install (table, GTK_OBJECT (widget), signal_name, key, modifiers);
}

/*****************************************
 * gtk_widget_remove_accelerator:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_remove_accelerator (GtkWidget           *widget,
			       GtkAcceleratorTable *table,
			       gchar               *signal_name)
{
  g_return_if_fail (widget != NULL);

  if (gtk_signal_emit (GTK_OBJECT (widget), widget_signals[REMOVE_ACCELERATOR], signal_name))
    gtk_accelerator_table_remove (table, GTK_OBJECT (widget), signal_name);
}

/*****************************************
 * gtk_widget_event:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

gint
gtk_widget_event (GtkWidget *widget,
		  GdkEvent  *event)
{
  gint return_val;
  gint signal_num;

  g_return_val_if_fail (widget != NULL, FALSE);

  return_val = FALSE;
  if (gtk_signal_emit (GTK_OBJECT (widget), widget_signals[EVENT], event, &return_val))
    {
      if (return_val)
	return TRUE;

      switch (event->type)
	{
	case GDK_NOTHING:
	  signal_num = -1;
	  break;
	case GDK_BUTTON_PRESS:
	case GDK_2BUTTON_PRESS:
	case GDK_3BUTTON_PRESS:
	  signal_num = BUTTON_PRESS_EVENT;
	  break;
	case GDK_BUTTON_RELEASE:
	  signal_num = BUTTON_RELEASE_EVENT;
	  break;
	case GDK_MOTION_NOTIFY:
	  signal_num = MOTION_NOTIFY_EVENT;
	  break;
	case GDK_DELETE:
	  signal_num = DELETE_EVENT;
	  break;
	case GDK_DESTROY:
	  signal_num = DESTROY_EVENT;
	  break;
	case GDK_EXPOSE:
	  signal_num = EXPOSE_EVENT;
	  break;
	case GDK_KEY_PRESS:
	  signal_num = KEY_PRESS_EVENT;
	  break;
	case GDK_KEY_RELEASE:
	  signal_num = KEY_RELEASE_EVENT;
	  break;
	case GDK_ENTER_NOTIFY:
	  signal_num = ENTER_NOTIFY_EVENT;
	  break;
	case GDK_LEAVE_NOTIFY:
	  signal_num = LEAVE_NOTIFY_EVENT;
	  break;
	case GDK_FOCUS_CHANGE:
	  if (event->focus_change.in)
	    signal_num = FOCUS_IN_EVENT;
	  else
	    signal_num = FOCUS_OUT_EVENT;
	  break;
	case GDK_CONFIGURE:
	  signal_num = CONFIGURE_EVENT;
	  break;
	case GDK_MAP:
	  signal_num = MAP_EVENT;
	  break;
	case GDK_UNMAP:
	  signal_num = UNMAP_EVENT;
	  break;
	case GDK_PROPERTY_NOTIFY:
	  signal_num = PROPERTY_NOTIFY_EVENT;
	  break;
	case GDK_SELECTION_CLEAR:
	  signal_num = SELECTION_CLEAR_EVENT;
	  break;
	case GDK_SELECTION_REQUEST:
	  signal_num = SELECTION_REQUEST_EVENT;
	  break;
	case GDK_SELECTION_NOTIFY:
	  signal_num = SELECTION_NOTIFY_EVENT;
	  break;
	case GDK_OTHER_EVENT:
	  signal_num = OTHER_EVENT;
	  break;
	default:
	  g_warning ("could not determine signal number for event: %d", event->type);
	  return return_val;
	}

      if (signal_num != -1)
	gtk_signal_emit (GTK_OBJECT (widget), widget_signals[signal_num], event, &return_val);
    }

  return return_val;
}

/*****************************************
 * gtk_widget_activate:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_activate (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (WIDGET_CLASS (widget)->activate_signal)
    gtk_signal_emit (GTK_OBJECT (widget), WIDGET_CLASS (widget)->activate_signal);
}

/*****************************************
 * gtk_widget_reparent:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_reparent (GtkWidget *widget,
		     GtkWidget *new_parent)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (new_parent != NULL);
  g_return_if_fail (GTK_IS_CONTAINER (new_parent));

  if (widget->parent != new_parent)
    {
      gtk_container_remove (GTK_CONTAINER (widget->parent), widget);
      gtk_container_add (GTK_CONTAINER (new_parent), widget);

      if (GTK_WIDGET_REALIZED (widget))
	{
	  if (GTK_WIDGET_REALIZED (new_parent) && !GTK_WIDGET_NO_WINDOW (widget))
	    {
	      gdk_window_reparent (widget->window, widget->parent->window, 0, 0);
	    }
	  else
	    {
	      GTK_WIDGET_UNSET_FLAGS (widget, GTK_REALIZED | GTK_MAPPED);
	      if (!GTK_WIDGET_NO_WINDOW (widget))
		gdk_window_destroy (widget->window);
	      widget->window = NULL;

	      if (GTK_WIDGET_REALIZED (new_parent))
		gtk_widget_realize (widget);
	      if (GTK_WIDGET_MAPPED (new_parent))
		gtk_widget_map (widget);
	    }
	}

      if (!GTK_WIDGET_REALIZED (widget) && GTK_WIDGET_REALIZED (new_parent))
	gtk_widget_realize (widget);
      if (!GTK_WIDGET_MAPPED (widget) && GTK_WIDGET_MAPPED (new_parent))
	gtk_widget_map (widget);

      gtk_container_check_resize (GTK_CONTAINER (widget->parent), widget);
    }
}

/*****************************************
 * gtk_widget_popup:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_popup (GtkWidget *widget,
		  gint       x,
		  gint       y)
{
  g_return_if_fail (widget != NULL);

  if (!GTK_WIDGET_VISIBLE (widget))
    {
      if (!GTK_WIDGET_REALIZED (widget))
	gtk_widget_realize (widget);
      if (!GTK_WIDGET_NO_WINDOW (widget))
	gdk_window_move (widget->window, x, y);
      gtk_widget_show (widget);
    }
}

/*****************************************
 * gtk_widget_intersect:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

gint
gtk_widget_intersect (GtkWidget    *widget,
		      GdkRectangle *area,
		      GdkRectangle *intersection)
{
  GdkRectangle *dest;
  GdkRectangle tmp;
  gint return_val;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (area != NULL, FALSE);

  if (intersection)
    dest = intersection;
  else
    dest = &tmp;

  return_val = gdk_rectangle_intersect ((GdkRectangle*) &widget->allocation, area, dest);

  if (return_val && intersection && !GTK_WIDGET_NO_WINDOW (widget))
    {
      intersection->x -= widget->allocation.x;
      intersection->y -= widget->allocation.y;
    }

  return return_val;
}


gint
gtk_widget_basic (GtkWidget *widget)
{
  GList *children;
  GList *tmp_list;
  gint return_val;

  g_return_val_if_fail (widget != NULL, FALSE);

  if (!GTK_WIDGET_BASIC (widget))
    return FALSE;
  else if (GTK_IS_CONTAINER (widget))
    {
      children = gtk_container_children (GTK_CONTAINER (widget));
      if (children)
	{
	  return_val = TRUE;
	  tmp_list = children;

	  while (tmp_list)
	    {
	      if (!gtk_widget_basic (GTK_WIDGET (tmp_list->data)))
		{
		  return_val = FALSE;
		  break;
		}

	      tmp_list = tmp_list->next;
	    }

	  g_list_free (children);
	  return return_val;
	}
    }

  return TRUE;
}


/*****************************************
 * gtk_widget_grab_focus:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_grab_focus (GtkWidget *widget)
{
  GtkWidget *window;
  GtkWidget *child;
  gint window_type;

  g_return_if_fail (widget != NULL);

  window_type = gtk_window_get_type ();
  window = widget->parent;
  child = widget;

  while (window && !gtk_type_is_a (GTK_WIDGET_TYPE (window), window_type))
    {
      GTK_CONTAINER (window)->focus_child = child;
      child = window;
      window = window->parent;
    }

  if (window && gtk_type_is_a (GTK_WIDGET_TYPE (window), window_type))
    {
      GTK_CONTAINER (window)->focus_child = child;
      gtk_window_set_focus (GTK_WINDOW (window), widget);
    }
}

/*****************************************
 * gtk_widget_grab_default:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_grab_default (GtkWidget *widget)
{
  GtkWidget *window;
  gint window_type;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_WIDGET_CAN_DEFAULT (widget));

  window_type = gtk_window_get_type ();
  window = widget->parent;

  while (window && !gtk_type_is_a (GTK_WIDGET_TYPE (window), window_type))
    window = window->parent;

  if (window && gtk_type_is_a (GTK_WIDGET_TYPE (window), window_type))
    gtk_window_set_default (GTK_WINDOW (window), widget);
}

/*****************************************
 * gtk_widget_restore_state:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_restore_state (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);

  widget->state = widget->saved_state;
  if (gtk_signal_emit (GTK_OBJECT (widget), widget_signals[STATE_CHANGED]))
    {
      if (GTK_IS_CONTAINER (widget))
	gtk_container_foreach (GTK_CONTAINER (widget),
			       gtk_widget_propagate_restore,
			       NULL);
    }
}

/*****************************************
 * gtk_widget_set_name:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_name (GtkWidget *widget,
		     gchar     *name)
{
  GtkStyle *new_style;

  g_return_if_fail (widget != NULL);

  if (widget->name)
    g_free (widget->name);
  widget->name = g_strdup (name);

  if (!GTK_WIDGET_USER_STYLE (widget))
    {
      new_style = gtk_rc_get_style (widget);
      gtk_widget_set_style_internal (widget, new_style);
    }
}

/*****************************************
 * gtk_widget_get_name:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

gchar*
gtk_widget_get_name (GtkWidget *widget)
{
  g_return_val_if_fail (widget != NULL, NULL);

  if (widget->name)
    return widget->name;
  return gtk_type_name (GTK_WIDGET_TYPE (widget));
}

/*****************************************
 * gtk_widget_set_state:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_state (GtkWidget    *widget,
		      GtkStateType  state)
{
  g_return_if_fail (widget != NULL);

  if (widget->state != state)
    {
      widget->saved_state = widget->state;
      widget->state = state;

      if (!gtk_signal_emit (GTK_OBJECT (widget), widget_signals[STATE_CHANGED]))
	return;
    }

  if (GTK_IS_CONTAINER (widget))
    gtk_container_foreach (GTK_CONTAINER (widget),
			   gtk_widget_propagate_state,
			   (gpointer) &state);
}

/*****************************************
 * gtk_widget_set_sensitive:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_sensitive (GtkWidget *widget,
			  gint       sensitive)
{
  GtkWidget *window;
  gint old_val;

  g_return_if_fail (widget != NULL);

  old_val = GTK_WIDGET_IS_SENSITIVE (widget);

  if (sensitive)
    {
      GTK_WIDGET_SET_FLAGS (widget, GTK_SENSITIVE);
    }
  else
    {
      GTK_WIDGET_UNSET_FLAGS (widget, GTK_SENSITIVE);

      if (GTK_WIDGET_HAS_FOCUS (widget))
	{
	  window = gtk_widget_get_ancestor (widget, gtk_window_get_type ());
	  if (window)
	    gtk_window_set_focus (GTK_WINDOW (window), NULL);
	}
    }

  if (GTK_IS_CONTAINER (widget))
    gtk_container_foreach (GTK_CONTAINER (widget),
			   gtk_widget_set_parent_sensitive,
			   &sensitive);

  if (old_val != GTK_WIDGET_IS_SENSITIVE (widget))
    gtk_widget_draw (widget, NULL);
}

/*****************************************
 * gtk_widget_set_parent:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_parent (GtkWidget *widget,
		       GtkWidget *parent)
{
  GtkStyle *style;
  gint sensitive;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (parent != NULL);

  gtk_object_ref (GTK_OBJECT (widget));

  widget->parent = parent;

  sensitive = GTK_WIDGET_IS_SENSITIVE (parent);
  gtk_widget_set_parent_sensitive (widget, &sensitive);

  if ((widget->parent->state != GTK_STATE_NORMAL) &&
      (widget->parent->state != widget->state))
    gtk_widget_set_state (widget, widget->parent->state);

  while (parent->parent != NULL)
    parent = parent->parent;

  if (GTK_WIDGET_ANCHORED (parent))
    {
      if (!GTK_WIDGET_USER_STYLE (widget))
	{
	  style = gtk_rc_get_style (widget);
	  if (style != widget->style)
	    gtk_widget_set_style_internal (widget, style);
	}

      if (GTK_IS_CONTAINER (widget))
	gtk_container_foreach (GTK_CONTAINER (widget),
			       gtk_widget_set_style_recurse,
			       NULL);
    }
}

/*****************************************
 * gtk_widget_set_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_style (GtkWidget *widget,
		      GtkStyle  *style)
{
  g_return_if_fail (widget != NULL);

  GTK_WIDGET_SET_FLAGS (widget, GTK_USER_STYLE);
  gtk_widget_set_style_internal (widget, style);
}

/*****************************************
 * gtk_widget_set_uposition:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_uposition (GtkWidget *widget,
			  gint       x,
			  gint       y)
{
  GtkWidgetAuxInfo *aux_info;

  g_return_if_fail (widget != NULL);

  aux_info = gtk_object_get_data (GTK_OBJECT (widget), aux_info_key);
  if (!aux_info)
    {
      aux_info = gtk_widget_aux_info_new ();
      gtk_object_set_data (GTK_OBJECT (widget), aux_info_key, aux_info);
    }

  if (x > -2)
    aux_info->x = x;
  if (y > -2)
    aux_info->y = y;

  if (GTK_WIDGET_REALIZED (widget) && GTK_IS_WINDOW (widget) &&
      (aux_info->x != -1) && (aux_info->y != -1))
    {
      gdk_window_set_hints (widget->window, aux_info->x, aux_info->y, 0, 0, 0, 0, GDK_HINT_POS);
      gdk_window_move (widget->window, aux_info->x, aux_info->y);
    }

  if (GTK_WIDGET_VISIBLE (widget) && widget->parent)
    gtk_widget_size_allocate (widget, &widget->allocation);
}

/*****************************************
 * gtk_widget_set_usize:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_usize (GtkWidget *widget,
		      gint       width,
		      gint       height)
{
  GtkWidgetAuxInfo *aux_info;

  g_return_if_fail (widget != NULL);

  aux_info = gtk_object_get_data (GTK_OBJECT (widget), aux_info_key);
  if (!aux_info)
    {
      aux_info = gtk_widget_aux_info_new ();
      gtk_object_set_data (GTK_OBJECT (widget), aux_info_key, aux_info);
    }

  if (width > -1)
    aux_info->width = width;
  if (height > -1)
    aux_info->height = height;

  if (GTK_WIDGET_VISIBLE (widget))
    {
      if (widget->parent && GTK_WIDGET_VISIBLE (widget->parent))
	gtk_container_check_resize (GTK_CONTAINER (widget->parent), NULL);
      else if (GTK_IS_CONTAINER (widget))
	gtk_container_check_resize (GTK_CONTAINER (widget), NULL);
    }
}

/*****************************************
 * gtk_widget_set_events:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_events (GtkWidget *widget,
		       gint       events)
{
  gint *eventp;

  g_return_if_fail (widget != NULL);

  eventp = gtk_object_get_data (GTK_OBJECT (widget), event_key);

  if (events)
    {
      if (!eventp)
	eventp = g_new (gint, 1);

      *eventp = events;
      gtk_object_set_data (GTK_OBJECT (widget), event_key, eventp);
    }
  else
    {
      if (eventp)
	g_free (eventp);

      gtk_object_remove_data (GTK_OBJECT (widget), event_key);
    }
}


/*****************************************
 * gtk_widget_get_toplevel:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkWidget*
gtk_widget_get_toplevel (GtkWidget *widget)
{
  g_return_val_if_fail (widget != NULL, NULL);

  while (widget->parent)
    widget = widget->parent;

  return widget;
}

/*****************************************
 * gtk_widget_get_ancestor:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkWidget*
gtk_widget_get_ancestor (GtkWidget *widget,
			 gint       type)
{
  g_return_val_if_fail (widget != NULL, NULL);

  while (widget && !gtk_type_is_a (GTK_WIDGET_TYPE (widget), type))
    widget = widget->parent;

  if (!(widget && gtk_type_is_a (GTK_WIDGET_TYPE (widget), type)))
    return NULL;

  return widget;
}

/*****************************************
 * gtk_widget_get_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GdkColormap*
gtk_widget_get_colormap (GtkWidget *widget)
{
  GdkColormap *colormap;

  g_return_val_if_fail (widget != NULL, NULL);

  if (!widget->window)
    {
      colormap = gtk_object_get_data (GTK_OBJECT (widget), colormap_key);
      if (colormap)
	return colormap;
      return gtk_widget_get_default_colormap ();
    }

  return gdk_window_get_colormap (widget->window);
}

/*****************************************
 * gtk_widget_get_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GdkVisual*
gtk_widget_get_visual (GtkWidget *widget)
{
  GdkVisual *visual;

  g_return_val_if_fail (widget != NULL, NULL);

  if (!widget->window)
    {
      visual = gtk_object_get_data (GTK_OBJECT (widget), visual_key);
      if (visual)
	return visual;
      return gtk_widget_get_default_visual ();
    }

  return gdk_window_get_visual (widget->window);
}

/*****************************************
 * gtk_widget_get_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkStyle*
gtk_widget_get_style (GtkWidget *widget)
{
  g_return_val_if_fail (widget != NULL, NULL);

  return widget->style;
}

/*****************************************
 * gtk_widget_get_events:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

gint
gtk_widget_get_events (GtkWidget *widget)
{
  gint *events;

  g_return_val_if_fail (widget != NULL, 0);

  events = gtk_object_get_data (GTK_OBJECT (widget), event_key);
  if (events)
    return *events;

  return 0;
}

/*****************************************
 * gtk_widget_get_pointer:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_get_pointer (GtkWidget *widget,
			gint      *x,
			gint      *y)
{
  g_return_if_fail (widget != NULL);

  if (x)
    *x = -1;
  if (y)
    *y = -1;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_get_pointer (widget->window, x, y, NULL);

      if (GTK_WIDGET_NO_WINDOW (widget))
	{
	  if (x)
	    *x -= widget->allocation.x;
	  if (y)
	    *y -= widget->allocation.y;
	}
    }
}

/*****************************************
 * gtk_widget_push_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_push_colormap (GdkColormap *cmap)
{
  colormap_stack = g_slist_prepend (colormap_stack, cmap);
}

/*****************************************
 * gtk_widget_push_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_push_visual (GdkVisual *visual)
{
  visual_stack = g_slist_prepend (visual_stack, visual);
}

/*****************************************
 * gtk_widget_push_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_push_style (GtkStyle *style)
{
  gtk_style_ref (style);
  style_stack = g_slist_prepend (style_stack, style);
}

/*****************************************
 * gtk_widget_pop_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_pop_colormap ()
{
  GSList *tmp;

  if (colormap_stack)
    {
      tmp = colormap_stack;
      colormap_stack = colormap_stack->next;
      g_slist_free_1 (tmp);
    }
}

/*****************************************
 * gtk_widget_pop_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_pop_visual ()
{
  GSList *tmp;

  if (visual_stack)
    {
      tmp = visual_stack;
      visual_stack = visual_stack->next;
      g_slist_free_1 (tmp);
    }
}

/*****************************************
 * gtk_widget_pop_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_pop_style ()
{
  GSList *tmp;

  if (style_stack)
    {
      tmp = style_stack;
      style_stack = style_stack->next;
      gtk_style_unref ((GtkStyle*) tmp->data);
      g_slist_free_1 (tmp);
    }
}

/*****************************************
 * gtk_widget_set_default_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_default_colormap (GdkColormap *colormap)
{
  if (default_colormap && (default_colormap != colormap))
    gdk_colormap_destroy (default_colormap);
  default_colormap = colormap;
}

/*****************************************
 * gtk_widget_set_default_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_default_visual (GdkVisual *visual)
{
  default_visual = visual;
}

/*****************************************
 * gtk_widget_set_default_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

void
gtk_widget_set_default_style (GtkStyle *style)
{
  if (default_style)
    gtk_style_unref (default_style);

  default_style = style;
  gtk_style_ref (default_style);
}

/*****************************************
 * gtk_widget_get_default_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GdkColormap*
gtk_widget_get_default_colormap ()
{
  if (!default_colormap)
    default_colormap = gdk_colormap_get_system ();

  return default_colormap;
}

/*****************************************
 * gtk_widget_get_default_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GdkVisual*
gtk_widget_get_default_visual ()
{
  if (!default_visual)
    default_visual = gdk_visual_get_system ();

  return default_visual;
}

/*****************************************
 * gtk_widget_get_default_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

GtkStyle*
gtk_widget_get_default_style ()
{
  if (!default_style)
    {
      default_style = gtk_style_new ();
      gtk_style_ref (default_style);
    }

  return default_style;
}


/*****************************************
 * gtk_widget_marshal_signal_1:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_marshal_signal_1 (GtkObject      *object,
			     GtkSignalFunc   func,
			     gpointer        func_data,
			     GtkArg         *args)
{
  GtkWidgetSignal1 rfunc;

  rfunc = (GtkWidgetSignal1) func;

  (* rfunc) (object, args[0].d.p, func_data);
}

/*****************************************
 * gtk_widget_marshal_signal_2:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_marshal_signal_2 (GtkObject      *object,
			     GtkSignalFunc   func,
			     gpointer        func_data,
			     GtkArg         *args)
{
  GtkWidgetSignal2 rfunc;
  gint *return_val;

  rfunc = (GtkWidgetSignal2) func;
  return_val = (gint*) args[3].d.p;

  *return_val = (* rfunc) (object, args[0].d.p, args[1].d.c, args[2].d.c, func_data);
}

/*****************************************
 * gtk_widget_marshal_signal_3:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_marshal_signal_3 (GtkObject      *object,
			     GtkSignalFunc   func,
			     gpointer        func_data,
			     GtkArg         *args)
{
  GtkWidgetSignal3 rfunc;

  rfunc = (GtkWidgetSignal3) func;

  (* rfunc) (object, args[0].d.p, func_data);
}

/*****************************************
 * gtk_widget_marshal_signal_4:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_marshal_signal_4 (GtkObject      *object,
			     GtkSignalFunc   func,
			     gpointer        func_data,
			     GtkArg         *args)
{
  GtkWidgetSignal4 rfunc;
  gint *return_val;

  rfunc = (GtkWidgetSignal4) func;
  return_val = (gint*) args[1].d.p;

  *return_val = (* rfunc) (object, args[0].d.p, func_data);
}

/*****************************************
 * gtk_real_widget_destroy:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_destroy (GtkObject *object)
{
  GtkWidget *widget;
  GtkWidgetAuxInfo *aux_info;
  gint *events;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_WIDGET (object));

  widget = GTK_WIDGET (object);

  gtk_grab_remove (widget);

  if (widget->name)
    g_free (widget->name);

  aux_info = gtk_object_get_data (GTK_OBJECT (widget), aux_info_key);
  if (aux_info)
    {
      gtk_widget_aux_info_destroy (aux_info);
      gtk_object_remove_data (GTK_OBJECT (widget), aux_info_key);
    }

  events = gtk_object_get_data (GTK_OBJECT (object), event_key);
  if (events)
    g_free (events);

  if (GTK_WIDGET_REALIZED (widget))
    gtk_widget_unrealize (widget);
  gtk_style_unref (widget->style);

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/*****************************************
 * gtk_real_widget_show:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_show (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (!GTK_WIDGET_VISIBLE (widget))
    {
      GTK_WIDGET_SET_FLAGS (widget, GTK_VISIBLE);

      if (widget->parent)
	{
	  gtk_container_check_resize (GTK_CONTAINER (widget->parent), widget);

	  if (GTK_WIDGET_MAPPED (widget->parent))
	    gtk_widget_map (widget);
	}
    }
}

/*****************************************
 * gtk_real_widget_hide:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_hide (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (GTK_WIDGET_VISIBLE (widget))
    {
      GTK_WIDGET_UNSET_FLAGS (widget, GTK_VISIBLE);

      if (GTK_WIDGET_MAPPED (widget))
	gtk_widget_unmap (widget);

      if (widget->parent)
	gtk_container_check_resize (GTK_CONTAINER (widget->parent), widget);
    }
}

/*****************************************
 * gtk_real_widget_map:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_map (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (GTK_WIDGET_REALIZED (widget) && !GTK_WIDGET_MAPPED (widget))
    {
      GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

      if (!GTK_WIDGET_NO_WINDOW (widget))
	gdk_window_show (widget->window);
      else
	gtk_widget_draw (widget, NULL);
    }
}

/*****************************************
 * gtk_real_widget_unmap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_unmap (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (GTK_WIDGET_MAPPED (widget))
    {
      GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);

      if (GTK_WIDGET_NO_WINDOW (widget))
	gdk_window_clear_area (widget->window,
			       widget->allocation.x,
			       widget->allocation.y,
			       widget->allocation.width,
			       widget->allocation.height);
      else
	gdk_window_hide (widget->window);
    }
}

/*****************************************
 * gtk_real_widget_realize:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_realize (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  widget->window = widget->parent->window;
  widget->style = gtk_style_attach (widget->style, widget->window);
}

/*****************************************
 * gtk_real_widget_unrealize:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_unrealize (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  GTK_WIDGET_UNSET_FLAGS (widget, GTK_REALIZED | GTK_MAPPED);

  gtk_style_detach (widget->style);
  if (!GTK_WIDGET_NO_WINDOW (widget))
    {
      gdk_window_set_user_data (widget->window, NULL);
      gdk_window_destroy (widget->window);
    }
  widget->window = NULL;
}

/*****************************************
 * gtk_real_widget_draw:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_draw (GtkWidget    *widget,
		      GdkRectangle *area)
{
  GdkEventExpose event;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));
  g_return_if_fail (area != NULL);

  if (GTK_WIDGET_DRAWABLE (widget))
    {
      event.type = GDK_EXPOSE;
      event.window = widget->window;
      event.area = *area;

      gtk_widget_event (widget, (GdkEvent*) &event);
    }
}

/*****************************************
 * gtk_real_widget_size_allocate:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_real_widget_size_allocate (GtkWidget     *widget,
			       GtkAllocation *allocation)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (GTK_WIDGET_NO_WINDOW (widget) &&
      GTK_WIDGET_MAPPED (widget) &&
      ((widget->allocation.x != allocation->x) ||
       (widget->allocation.y != allocation->y) ||
       (widget->allocation.width != allocation->width) ||
       (widget->allocation.height != allocation->height)) &&
      (widget->allocation.width != 0) &&
      (widget->allocation.height != 0))
    gdk_window_clear_area (widget->window,
			   widget->allocation.x,
			   widget->allocation.y,
			   widget->allocation.width,
			   widget->allocation.height);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget) &&
      !GTK_WIDGET_NO_WINDOW (widget))
    gdk_window_move_resize (widget->window,
			    allocation->x, allocation->y,
			    allocation->width, allocation->height);
}

/*****************************************
 * gtk_widget_peek_colormap:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static GdkColormap*
gtk_widget_peek_colormap ()
{
  if (colormap_stack)
    return (GdkColormap*) colormap_stack->data;
  return gtk_widget_get_default_colormap ();
}

/*****************************************
 * gtk_widget_peek_visual:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static GdkVisual*
gtk_widget_peek_visual ()
{
  if (visual_stack)
    return (GdkVisual*) visual_stack->data;
  return gtk_widget_get_default_visual ();
}

/*****************************************
 * gtk_widget_peek_style:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static GtkStyle*
gtk_widget_peek_style ()
{
  if (style_stack)
    return (GtkStyle*) style_stack->data;
  return gtk_widget_get_default_style ();
}


/*****************************************
 * gtk_widget_set_parent_sensitive:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_set_parent_sensitive (GtkWidget *widget,
				 gpointer   client_data)
{
  GtkWidget *window;
  gint *sensitive;

  sensitive = client_data;
  if (*sensitive)
    {
      GTK_WIDGET_SET_FLAGS (widget, GTK_PARENT_SENSITIVE);

      if (GTK_IS_CONTAINER (widget) && GTK_WIDGET_SENSITIVE (widget))
	gtk_container_foreach (GTK_CONTAINER (widget),
			       gtk_widget_set_parent_sensitive,
			       client_data);
    }
  else
    {
      GTK_WIDGET_UNSET_FLAGS (widget, GTK_PARENT_SENSITIVE);

      if (GTK_WIDGET_HAS_FOCUS (widget))
	{
	  window = gtk_widget_get_ancestor (widget, gtk_window_get_type ());
	  if (window)
	    gtk_window_set_focus (GTK_WINDOW (window), NULL);
	}

      if (GTK_IS_CONTAINER (widget))
	gtk_container_foreach (GTK_CONTAINER (widget),
			       gtk_widget_set_parent_sensitive,
			       client_data);
    }
}

/*****************************************
 * gtk_widget_propagate_restore:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_propagate_restore (GtkWidget *widget,
			      gpointer   client_data)
{
  gtk_widget_restore_state (widget);
}

/*****************************************
 * gtk_widget_propagate_state:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_propagate_state (GtkWidget *widget,
			    gpointer   client_data)
{
  GtkStateType *state;

  state = (GtkStateType*) client_data;
  gtk_widget_set_state (widget, *state);
}

/*****************************************
 * gtk_widget_draw_children_recurse:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_draw_children_recurse (GtkWidget *widget,
				  gpointer   client_data)
{
  gtk_widget_draw (widget, NULL);
  gtk_widget_draw_children (widget);
}

/*****************************************
 * gtk_widget_set_style_internal:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_set_style_internal (GtkWidget *widget,
			       GtkStyle  *style)
{
  GtkRequisition old_requisition;

  g_return_if_fail (widget != NULL);

  if (widget->style != style)
    {
      if (GTK_WIDGET_REALIZED (widget))
	gtk_style_detach (widget->style);

      gtk_style_unref (widget->style);

      widget->style = style;
      gtk_style_ref (widget->style);

      if (GTK_WIDGET_REALIZED (widget))
	widget->style = gtk_style_attach (widget->style, widget->window);

      if (widget->parent)
	{
	  old_requisition = widget->requisition;
	  gtk_widget_size_request (widget, &widget->requisition);

	  if ((old_requisition.width != widget->requisition.width) ||
	      (old_requisition.height != widget->requisition.height))
	    gtk_container_check_resize (GTK_CONTAINER (widget->parent), widget);
	  else if (GTK_WIDGET_DRAWABLE (widget))
	    gtk_widget_draw (widget, NULL);
	}
    }
}

/*****************************************
 * gtk_widget_set_style_recurse:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_set_style_recurse (GtkWidget *widget,
			      gpointer   client_data)
{
  GtkStyle *style;

  style = gtk_rc_get_style (widget);
  if (style != widget->style)
    gtk_widget_set_style (widget, style);

  if (GTK_IS_CONTAINER (widget))
    gtk_container_foreach (GTK_CONTAINER (widget),
			   gtk_widget_set_style_recurse,
			   NULL);
}

/*****************************************
 * gtk_widget_aux_info_new:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static GtkWidgetAuxInfo*
gtk_widget_aux_info_new ()
{
  GtkWidgetAuxInfo *aux_info;

  if (!aux_info_mem_chunk)
    aux_info_mem_chunk = g_mem_chunk_new ("widget aux info mem chunk",
					  sizeof (GtkWidgetAuxInfo),
					  1024, G_ALLOC_AND_FREE);

  aux_info = g_chunk_new (GtkWidgetAuxInfo, aux_info_mem_chunk);

  aux_info->x = -1;
  aux_info->y = -1;
  aux_info->width = 0;
  aux_info->height = 0;

  return aux_info;
}

/*****************************************
 * gtk_widget_aux_info_destroy:
 *
 *   arguments:
 *
 *   results:
 *****************************************/

static void
gtk_widget_aux_info_destroy (GtkWidgetAuxInfo *aux_info)
{
  g_return_if_fail (aux_info != NULL);

  g_mem_chunk_free (aux_info_mem_chunk, aux_info);
}
