/*
 * This is a plugin for the GIMP v1.0.
 *
 * gradmap.c -- produce Gradient Map
 * Copyright (C) 1997 Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * $Id: gradmap.c 1.3 Wed, 04 Jun 1997 22:56:59 -0700 pmattis $
 *
 * This plug-in requires GIMP v0.99.7 (or above).
 *
 * This plug-in maps the image with the custom gradient (specified in
 * the gradient editor).
 *
 * This version is very alpha, since I believe there's a memory leak
 * somewhere in the code.
 *
 * The trick is very cheap that this plug-in creates new image and
 * layer, blends it and then remap the contents of specified drawable
 * using it.  I don't know how to clean up when the temporary image is
 * no longer needed. :-(
 *
 * Feel free about to correct my WRONG English, or to modify Plug-in Path,
 * and so on. ;-)
 *
 *	Eiichi Takamori <taka@ma1.seikyou.ne.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "libgimp/gimp.h"

/* Some useful macros */

#define TILE_CACHE_SIZE 16
#define BOUNDS(a,x,y)  ((a < x) ? x : ((a > y) ? y : a))

/* Declare a local function.
 */
static void	query	(void);
static void	run	(gchar   *name,
			 gint    nparams,
			 GParam  *param,
			 gint    *nreturn_vals,
			 GParam  **return_vals);

static void	gradmap	(GDrawable *drawable);
static guchar  *get_custom_gradient( GDrawable *drawable );


GPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

MAIN ();

static void
query()
{
  static GParamDef args[]=
    {
      { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
      { PARAM_IMAGE, "image", "Input image (unused)" },
      { PARAM_DRAWABLE, "drawable", "Input drawable" },
   };
  static GParamDef *return_vals = NULL;
  static gint nargs = sizeof (args) / sizeof (args[0]);
  static gint nreturn_vals = 0;

  gimp_install_procedure ("plug_in_gradmap",
			  "Map the contents of the specified drawable with the cusom gradient",
			  "This plug-in maps the contents of the specified drawable with the cusom gradient. Black becomes the leftmost color of the gradient, and white becomes the rightmost. Other intermediate colors become the related color of the gradient. Works on grayscale and RGB image with or without alpha channel.",
			  "Eiichi Takamori",
			  "Eiichi Takamori",
			  "1997",
			  "<Image>/Filters/Image/Gradient Map",
			  "RGB*, GRAY*",
			  PROC_PLUG_IN,
			  nargs, nreturn_vals,
			  args, return_vals);
}

static void
run (gchar   *name,
     gint    nparams,
     GParam  *param,
     gint    *nreturn_vals,
     GParam  **return_vals)
{
  static GParam values[1];
  GDrawable *drawable;
  GRunModeType run_mode;
  GStatusType status = STATUS_SUCCESS;

  g_debug ("gradmap");

  run_mode = param[0].data.d_int32;

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = status;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  /*  Make sure that the drawable is gray or RGB color  */
  if (gimp_drawable_color (drawable->id) || gimp_drawable_gray (drawable->id))
	{
	  gimp_progress_init ("Gradient Map...");
	  gimp_tile_cache_ntiles (TILE_CACHE_SIZE);

	  gradmap (drawable);

	  if (run_mode != RUN_NONINTERACTIVE)
	    gimp_displays_flush ();
	}
      else
	{
	  /* gimp_message ("gradmap: cannot operate on indexed color images"); */
	  status = STATUS_EXECUTION_ERROR;
	}

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}

static void
gradmap (GDrawable *drawable)
{
  GPixelRgn src_rgn, dest_rgn;
  gpointer pr;
  guchar *src_row, *dest_row;
  guchar *src, *dest;
  guchar *gradient, *grad;
  gint x1, y1, x2, y2;
  gint row, col;
  gint alpha, has_alpha, bpp, b;
  gint is_color;
  gint progress, max_progress;
  gint grayscale;

  gimp_drawable_mask_bounds (drawable->id, &x1, &y1, &x2, &y2);

  bpp = alpha = gimp_drawable_bpp( drawable->id );
  has_alpha = gimp_drawable_has_alpha( drawable->id );
  if( has_alpha )
    alpha--;
  is_color = gimp_drawable_color( drawable->id );

  gradient = get_custom_gradient(drawable);

  gimp_pixel_rgn_init (&src_rgn, drawable, x1, y1, x2-x1, y2-y1, FALSE, FALSE);
  gimp_pixel_rgn_init (&dest_rgn, drawable, x1, y1, x2-x1, y2-y1, TRUE, TRUE);

  /* Initialize progress */
  progress = 0;
  max_progress = (x2 - x1) * (y2 - y1);

  /*  g_debug("gradmap"); */

  for (pr = gimp_pixel_rgns_register (2, &src_rgn, &dest_rgn);
       pr != NULL; pr = gimp_pixel_rgns_process (pr))
    {
      src_row = src_rgn.data;
      dest_row = dest_rgn.data;

      for ( row = 0; row < src_rgn.h; row++ )
	{
	  src = src_row;
	  dest = dest_row;

	  for ( col = 0; col < src_rgn.w; col++ )
	    {
	      if( is_color )
		grayscale = ( src[0] + src[1] + src[2] ) / 3;
	      else
		grayscale = src[0];

	      grad = &gradient[ grayscale * bpp ];
	      for( b = 0; b<bpp; b++ )
		dest[b] = grad[b];

	      src += src_rgn.bpp;
	      dest += dest_rgn.bpp;
	    }
	  src_row += src_rgn.rowstride;
	  dest_row += dest_rgn.rowstride;
	}
      /* Update progress */
      progress += src_rgn.w * src_rgn.h;
      gimp_progress_update ((double) progress / (double) max_progress);
    }

  g_free( gradient );

  /*  update the blurred region  */
  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, x1, y1, (x2 - x1), (y2 - y1));
}

#define NUM_GRAD  256

static guchar *
get_custom_gradient( GDrawable *drawable )
{
  GImageType    image_type;
  gint          image_id;
  GDrawableType layer_type;
  gint          layer_id;
  GDrawable     *layer_drawable;
  guchar        *gradient;
  GPixelRgn     srcrgn;
  GParam        *return_vals;
  gint          nreturn_vals;
  gint          bpp;

  layer_type = gimp_drawable_type( drawable->id );
  switch( layer_type )
    {
    case RGB_IMAGE:
    case RGBA_IMAGE:
      image_type = RGB;
      break;
    case GRAY_IMAGE:
    case GRAYA_IMAGE:
      image_type = GRAY;
      break;
    default:
      printf("This plug-in doesn't support index image.\n");
      return NULL;
      break;
    }

  image_id = gimp_image_new ( NUM_GRAD, 1, image_type );
  if( image_id < 0 ) {
    printf("image_new failed\n");
    return NULL;
  }
  layer_id = gimp_layer_new ( image_id, "Layer 1", NUM_GRAD, 1, layer_type,
			      100, NORMAL_MODE );
  if( layer_id < 0 ) {
    printf("layer_new failed\n");
    return NULL;
  }
  gimp_image_add_layer ( image_id, layer_id, 0 );


  return_vals =
    gimp_run_procedure ( "gimp_blend", &nreturn_vals,
			 PARAM_IMAGE, image_id,
			 PARAM_DRAWABLE, layer_id,
			 PARAM_INT32, 3,       /* blend_mode = CUSTOM */
			 PARAM_INT32, 0,       /* paint_mode = NORMAL */
			 PARAM_INT32, 0,       /* gradient_type = LINEAR */
			 PARAM_FLOAT, 100.0,   /* opacity    = 100 */
			 PARAM_FLOAT, 0.0,     /* offset     = 0 */
			 PARAM_INT32, 0,       /* repeat     = 0 */
			 PARAM_INT32, FALSE,   /* supersample */
			 PARAM_INT32, 0,       /* max depth */
			 PARAM_FLOAT, 0.0,     /* threshold */
			 PARAM_FLOAT, 0.0,       /* x1 */
			 PARAM_FLOAT, 0.0,       /* y1 */
			 PARAM_FLOAT, (double) NUM_GRAD, /* x2 */
			 PARAM_FLOAT, 0.0,       /* y2 */
			 PARAM_END);
  gimp_destroy_params (return_vals, nreturn_vals);

  layer_drawable = gimp_drawable_get( layer_id );
  bpp = gimp_drawable_bpp ( layer_id );

  gradient = g_new ( guchar, NUM_GRAD * bpp );

  gimp_pixel_rgn_init ( &srcrgn, layer_drawable, 0, 0, NUM_GRAD, 0, FALSE, FALSE );
  gimp_pixel_rgn_get_row ( &srcrgn, gradient, 0, 0, NUM_GRAD );

  /* THE CLEAN UP HERE IS TERRIBLE since the below line causes segv when uncommented */
  /* gimp_drawable_detach( layer_drawable ); */

  gimp_image_delete( image_id );
  return gradient;
}
