//
// FILE:
// parseOfx.C
//
// FUNCTION:
// Parses OFX dtd's
//
// HISTORY:
// Written by Linas Vepstas March 1998

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "DtdParser.h"
#include "OutputCharStream.h"


// output messages is used to set up the SP parser
// to dump it's error messages to stderr
#define OUTPUT_MESSAGES

#ifdef OUTPUT_MESSAGES
#include "sptchar.h"
#include "MessageReporter.h"
#include "MessageTable.h"
#endif /* OUTPUT_MESSAGES */

#ifdef SP_NAMESPACE
namespace SP_NAMESPACE {
#endif

#ifdef OUTPUT_MESSAGES
#ifdef SP_MANUAL_INST

#define SP_DEFINE_TEMPLATES
#include "Owner.h"
#undef SP_DEFINE_TEMPLATES

#include "Message.h"

#ifdef SP_ANSI_CLASS_INST
template class Owner< Messenger>;
#else
typedef Owner< Messenger> Dummy_0;
#endif

#endif /* SP_MANUAL_INST */
#endif /* OUTPUT_MESSAGES */


static FileOutputByteStream standardOutput(1, 0);
static FileOutputByteStream standardError(2, 0);

#ifdef OUTPUT_MESSAGES

class MyMessageReporter
: public MessageReporter
{
  public:
    MyMessageReporter( const InputCodingSystem *, OutputCharStream *);
  private:
    Boolean getMessageText( const MessageFragment &, StringC &);
    const InputCodingSystem * codingSystem_;
};

MyMessageReporter::
MyMessageReporter( const InputCodingSystem * codesys,
                   OutputCharStream * errorStream)
: MessageReporter( errorStream),
  codingSystem_( codesys)
{
}

Boolean
MyMessageReporter::
getMessageText( const MessageFragment & frag, StringC & text)
{
  String<SP_TCHAR> str;
  if (!MessageTable::instance()->getText(frag, str))
    return 0;
#ifdef SP_WIDE_SYSTEM
  text.assign((const Char *)str.data(), str.size());
#else
  str += 0;
  text = codingSystem_->convertIn(str.data());
#endif
  return 1;
}

#endif /* OUTPUT_MESSAGES */

// ===========================================================
// convert the string type to an ordinary char *

char *
pfxCharify (const String<short unsigned int> &s)
{
   int len = s.size ();

   char * str = new char [len+1];

   for (int i=0; i<len; i++) {
      str[i] = s[i];
   }
   str[len] = 0x0;

   return str;
}

char * pfxToLower (char * str)
{
   int len = strlen (str);

   char * low = new char [len+1];
   
   for (int i=0; i<len; i++) {
      low[i] = tolower (str[i]);
   }
   low[len] = 0x0;

   return low;
}

char * pfxCapLower (char * str)
{
   int len = strlen (str);

   char * low = new char [len+1];
   
   for (int i=0; i<len; i++) {
      low[i] = tolower (str[i]);
   }

   if (1 < len) {
      low[0] = str[0];
   }

   low[len] = 0x0;

   return low;
}

#define DENT(dent) { for (int bonk=0; bonk<(dent); bonk++) printf ("    "); }

// ===========================================================
// class pfxBaseType stores & handles base types

// AddBaseVar adds a new variable with a base type.  
//    returns pointer to variable name if a base type, else returns null.

#define MAXVARS 3000

class pfxBaseType {
   public:
      pfxBaseType (void);

      char * AddBaseVar (const ElementType *);
      char * GetBaseType (char *varname);
      void PrintBaseTypes (int dent);
      void PrintBaseDecl (int dent, const ContentToken & token);


   private:
      char * varname[MAXVARS];
      char * vartype[MAXVARS];
      int nvars;

};

pfxBaseType :: pfxBaseType (void)
{
    nvars = 0;
    varname[0] = 0x0;
    vartype[0] = 0x0;
}

char * 
pfxBaseType :: AddBaseVar (const ElementType *type)
{

    // its not a base type unless we have a model group.
    if (ElementDefinition::modelGroup != 
        type->definition()->declaredContent()) return 0x0;

    const ModelGroup * mg;
    mg = type->definition()->compiledModelGroup()->modelGroup();

    // to be a base type, it must have exactly one member
    if (1 != mg->nMembers()) return 0x0;

    const ContentToken & token = mg->member (0);

    // anything with a sub-model cannot be a base type
    const ModelGroup * subModel = token.asModelGroup();
    if (subModel)  return 0x0;

    const LeafContentToken * leaf = token.asLeafContentToken();
    const ElementType * elementType = leaf->elementType();

    if (!elementType) {
       char * name = pfxCharify (type->name());

       printf ("Unkonwon, unexpected PCDATA: %s \n", name);
       return 0x0;
    }

    // now look to see if we have already processed this one 
    char *vname = pfxCharify (type->name());

    // if its already in out database, just return
    for (int i = 0; i<nvars; i++) {
        if (!strcmp (vname, varname[i])) {
           delete vname;
           return varname[i];
        }
    }

    // look to see if this is a typ[e that we recognize
    char * vtype = pfxCharify (elementType->name());
    if (strcmp ("XACC-AMOUNT",   vtype) &&
        strcmp ("XACC-BOOL",     vtype) &&
        strcmp ("XACC-DTTM",     vtype) &&
        strcmp ("XACC-ID",       vtype) &&
        strcmp ("XACC-INT",      vtype) &&
        strcmp ("XACC-PRICE",    vtype) &&
        strcmp ("XACC-RATE",     vtype) &&
        strcmp ("XACC-SRVRID",   vtype) &&
        strcmp ("XACC-STR",      vtype) &&
        strcmp ("XACC-URL",      vtype) &&
        strcmp ("XACC-UUID",     vtype)) {

        delete vtype;
        delete vname;
        return 0x0;
    } 

    // we haven't encounterd this before. Add it to our list.
    varname[nvars] = vname;
    vartype[nvars] = vtype;
    nvars ++;

    return vname;
}

void 
pfxBaseType :: PrintBaseTypes (int dent)
{

    DENT (dent); printf ("// Basic Types \n");
    DENT (dent); printf ("typedef double bfxAmount; \n");
    DENT (dent); printf ("typedef int    bfxBool; \n");
    DENT (dent); printf ("typedef Date   bfxDate; \n");
    DENT (dent); printf ("typedef char * bfxId; \n");
    DENT (dent); printf ("typedef int    bfxInt; \n");
    DENT (dent); printf ("typedef double bfxPrice; \n");
    DENT (dent); printf ("typedef double bfxRate; \n");
    DENT (dent); printf ("typedef char * bfxStr; \n");
    DENT (dent); printf ("typedef char * bfxSrvrid; \n");
    DENT (dent); printf ("typedef char * bfxUrl; \n");
    DENT (dent); printf ("typedef char   bfxUuid [36]; \n");
    printf ("\n\n");
}

void 
pfxBaseType :: PrintBaseDecl (int dent, const ContentToken & token)
{
    // print an individual member
    const LeafContentToken * leaf = token.asLeafContentToken();
    const ElementType * elementType = leaf->elementType();
    if (!elementType) return;

    char * varname = pfxCharify (elementType->name());
                
    if (!varname) return;

    char * vtype = GetBaseType (varname);
    if (!vtype) return;

    DENT (dent);

    if (!strcmp ("XACC-AMOUNT", vtype)) {
       printf ("bfxAmount ");
    } else

    if (!strcmp ("XACC-BOOL", vtype)) {
       printf ("bfxBool ");
    } else

    if (!strcmp ("XACC-DTTM", vtype)) {
       printf ("bfxDttm ");
    } else

    if (!strcmp ("XACC-ID", vtype)) {
       printf ("bfxId ");
    } else

    if (!strcmp ("XACC-INT", vtype)) {
       printf ("bfxInt ");
    } else

    if (!strcmp ("XACC-PRICE", vtype)) {
       printf ("bfxPrice ");
    } else

    if (!strcmp ("XACC-RATE", vtype)) {
       printf ("bfxRate ");
    } else

    if (!strcmp ("XACC-STR", vtype)) {
       printf ("bfxStr ");
    } else
    
    if (!strcmp ("XACC-SRVRID", vtype)) {
       printf ("bfxSrvrid ");
    } else
    
    if (!strcmp ("XACC-URL", vtype)) {
       printf ("bfxUrl ");
    } else
    
    if (!strcmp ("XACC-UUID", vtype)) {
       printf ("bfxUuid ");
    } else
    
    printf ("---> unknown type >%s<\n", vtype);


    switch ( token.occurrenceIndicator()) {

      // element *must* occur
      case ContentToken::none:
        break;

      // element may or may not occur
      case ContentToken::opt:
        printf ("*");
        break;

      // element must occur at least once, maybe more times.
      case ContentToken::plus:
        printf ("**");
        break;

      // element may occur zero or more times
      case ContentToken::rep:
        printf ("**");
        break;
    }

    // printed var names will be lower case
    char * vname = pfxToLower (varname);
    printf ("%s;\n", vname);
    delete vname;
    delete varname;
}


char *
pfxBaseType :: GetBaseType (char * vname)
{

    for (int i = 0; i<nvars; i++) {
        if (!strcmp (vname, varname[i])) return vartype[i];
    }

    return 0x0;
}

// ===========================================================
// handles compund types
// IsUnion returns true if its pure-or 

class pfxCompoundType : 
   public pfxBaseType
{
    public:
        pfxCompoundType (void);
        char * AddCompoundType (const ElementType *);
        void  PrintClassDecl (const ElementType *);

        void  PrintMemberDecl (int dent, int onion, const ContentToken &);

        int   IsUnion (const ModelGroup *);
        void  PrintUnionDecl (int dent, const ContentToken &);

        void  PrintGroupMemberDecls (int dent, const ModelGroup *);
        

        char * IsClass (char *);

   private:
      char * classname[MAXVARS];
      int nclasses;
};

pfxCompoundType :: pfxCompoundType (void)
{
   nclasses = 0;
   classname[0] = 0x0;
}

char *
pfxCompoundType :: IsClass (char * vname)
{
    for (int i = 0; i<nclasses; i++) {
        if (!strcmp (vname, classname[i])) return classname[i];
    }

    return 0x0;
}

char * 
pfxCompoundType :: AddCompoundType (const ElementType *type)
{

    // its not a compound type unless we have a model group.
    if (ElementDefinition::modelGroup != 
        type->definition()->declaredContent()) return 0x0;

    // make sure its not a base type
    char * basename = AddBaseVar (type);
    if (basename) return 0x0;

    const ModelGroup * mg;
    mg = type->definition()->compiledModelGroup()->modelGroup();

    char * cname = pfxCharify (type->name());

    classname [nclasses] = cname;
    nclasses ++;

    char * pname = pfxCapLower (cname);
    printf ("class cfx%s;\n", pname);
    delete pname;

    return cname;
}

void 
pfxCompoundType :: PrintMemberDecl (int dent, int onion, 
                                    const ContentToken & token)
{
    // print an individual member
    const LeafContentToken * leaf = token.asLeafContentToken();
    const ElementType * elementType = leaf->elementType();
    if (!elementType) return;

    char * varname = pfxCharify (elementType->name());
                
    // see if its a base type -- e.g. int, double, etc.
    char * basetype = GetBaseType (varname);
    if (basetype) {
        PrintBaseDecl (dent, token);
    } else {

        // Not base type.  Must be a class
        char * classtype = IsClass (varname);
        if (classtype) {
            char * decl = pfxCapLower (varname);
            char * var  = pfxToLower (varname);

            // look to see if the element occurs
            // once, more than once, etc.
            switch ( token.occurrenceIndicator()) {

              // element *must* occur
              case ContentToken::none:
                if (!onion) {
                    printf ("\n");
                    DENT (dent);
                    printf ("// %s *must* be present \n", decl);
                }
                DENT (dent);
                printf ("cfx%s %s; \n", decl, var);
                break;

              // element may or may not occur
              case ContentToken::opt:
                if (!onion) {
                    printf ("\n");
                    DENT (dent);
                    printf ("// %s may be a null pointer \n", decl);
                }
                DENT (dent);
                printf ("cfx%s *%s; \n", decl, var);
                break;

              // element must occur at least once, maybe more times.
              case ContentToken::plus:
                if (!onion) {
                    printf ("\n");
                    DENT (dent);
                    printf ("// %s is a list of one or more items \n", decl);
                }
                DENT (dent);
                printf ("cfx%s **%s; \n", decl, var);
                break;

              // element may occur zero or more times
              case ContentToken::rep:
                if (!onion) {
                    printf ("\n");
                    DENT (dent);
                    printf ("// %s is a list of zero or more items \n", decl);
                }
                DENT (dent);
                printf ("cfx%s **%s; \n", decl, var);
                break;
            }

            delete decl;
            delete var;
        } else {
            printf (" this variable has no type -----> %s \n", varname);
        }
    }
    delete varname;
}

int  
pfxCompoundType :: IsUnion (const ModelGroup * mg)
{
    if (ModelGroup::orConnector == mg->connector()) return 1;
    return 0;
}

void 
pfxCompoundType :: PrintUnionDecl (int dent,  const ContentToken & untoken)
{

    const ModelGroup * mg = untoken.asModelGroup();

    if (!mg) {
       printf ("Error: PrintUnionDecl: unexpected non model \n");
       return;
    }

    // The connector *must* be logical-OR type in order 
    // for this to be a union.
    if (ModelGroup::orConnector != mg->connector()) return;

    DENT (dent);
    printf ("enum { \n");
    int i = 0;
    while (i < mg->nMembers()) {
        const ContentToken & token = mg->member( i);
        const ModelGroup * subModel = token.asModelGroup();

        if ( subModel) {
             printf ("-----> we've got submodels duuude \n");
        } else {
            const LeafContentToken * leaf = token.asLeafContentToken();
            const ElementType * elementType = leaf->elementType();

            char * varname = pfxCharify (elementType->name());
            DENT (dent+1);
            printf ("T_%s,\n", varname);
        }

        i++;
    }
    DENT (dent);
    printf ("}; \n");

    DENT (dent);
    printf ("int u_type; \n");
    DENT (dent);
    printf ("typedef union { \n");

    i = 0;
    while (i < mg->nMembers()) {
        const ContentToken & token = mg->member( i);
        const ModelGroup * subModel = token.asModelGroup();
        if ( subModel) {
             printf ("-----> we've got submodels duuude \n");
        }

        else {

            // put the member elements into the class declaration
            PrintMemberDecl (dent+1, 1, token);
        }

        i++;
    }

    DENT (dent);
    printf ("} Onion;\n");
    
    // now handle repeated lists
    switch ( untoken.occurrenceIndicator()) {

      // element *must* occur
      case ContentToken::none:
        DENT (dent);
        printf ("Onion u;   // must occur \n");
        break;

      // element may or may not occur
      case ContentToken::opt:
        DENT (dent);
        printf ("Onion *u; // optionally present \n");
        break;

      // element must occur at least once, maybe more times.
      case ContentToken::plus:
        DENT (dent);
        printf ("Onion **u; // list of one or more\n");
        break;

      // element may occur zero or more times
      case ContentToken::rep:
        DENT (dent);
        printf ("Onion **u; // list of zero or more \n");
        break;
    }
}

void 
pfxCompoundType :: PrintGroupMemberDecls (int dent, const ModelGroup * mg)
{

    int i = 0;
    while (i < mg->nMembers()) {
        const ContentToken & token = mg->member( i);
        const ModelGroup * subModel = token.asModelGroup();

        if ( subModel) {
             if (ModelGroup::orConnector == subModel->connector()) {
                 PrintUnionDecl (dent, token);

             } else {
                 PrintGroupMemberDecls (dent, subModel);
                 if (ContentToken::none != token.occurrenceIndicator()) {
printf ("------> unhandled occurance indicator for subgroup!! \n");
                 }

             }
        } else {

            // put the member elements into the class declaration
            PrintMemberDecl (dent, 0, token);
        }

        i++;
    }
}


void 
pfxCompoundType :: PrintClassDecl (const ElementType *type)
{

    // its not a compound type unless we have a model group.
    if (ElementDefinition::modelGroup != 
        type->definition()->declaredContent()) return;

    const ModelGroup * mg;
    mg = type->definition()->compiledModelGroup()->modelGroup();

    // open the class declaration
    char * cname = pfxCharify (type->name());
    char * pname = pfxCapLower (cname);

    printf ("\n\n");
    printf ("class cfx%s \n", pname);
    printf ("{ \n");
    printf ("    public:\n");
    printf ("        cfx%s (void);      // constructor \n", pname);
    printf ("        virtual ~cfx%s (); // destructor \n", pname);
    printf ("\n");

    PrintGroupMemberDecls (2, mg);

    printf ("};\n");
}

// ===========================================================

extern "C"

void Usage (char *progname)
{
    printf ("Usage: %s <dtdfile> \n", progname);
    exit (1);
}

int
main( int argc, char ** argv)
{
    if ( argc < 2) Usage (argv[0]);

    DtdParser parser;

    pfxCompoundType compoundTypeHandler;

#ifdef OUTPUT_MESSAGES
    Owner< Messenger> mgr;
    if ( argv[ 1][ 0] != '-' || argv[ 1][ 1] != 's' || argv[ 1][ 2] != '\0') {
        mgr = new MyMessageReporter( parser.codingSystem(),
                                     new EncodeOutputCharStream( &standardError,
                                                                 parser.codingSystem()));
        parser.setMessenger( mgr.pointer());
    }
#endif

    StringC sysid;
    parser.makeSystemId( argc - 1, argv + 1, sysid);
    ConstPtr< Dtd> dtd = parser.parseDtd( sysid);

    if ( dtd.isNull()) {
        printf ("Error: specified dtd %s was null \n", argv[1]);
        return 1;
    }

    // ----------------------------------------------------
    // print everything
    time_t now = time (0);
    char * cnow = ctime (&now);

    printf ("//\n");
    printf ("// This file automatically generated by %s \n", argv[0]);
    printf ("// Do not edit -- your changes will be lost! \n");
    printf ("// Generated on %s \n", cnow);
    printf ("//\n");
    printf ("//\n");
    compoundTypeHandler.PrintBaseTypes (0);


    printf ("//----------------------------------------------\n");
    printf ("//\n");
    printf ("// Forward declarations of classes\n");
    printf ("//\n");
    printf ("//\n");
    const ElementType * type;
    Dtd::ConstElementTypeIter e( dtd->elementTypeIter());
    type = e.next();
    while ( type) {

        // see if is a base type
        char * varname = compoundTypeHandler.AddBaseVar (type);

        if (!varname) { 
            compoundTypeHandler.AddCompoundType (type);

            switch ( type->definition()->declaredContent()) {
              case ElementDefinition::modelGroup:
                break;
              case ElementDefinition::any: 
                break;
              default: {
                 char * vname = pfxCharify (type->name());
                 printf ("unexpected duuuuuuuuuuuuuuuuuuude: %s \n", vname);
                 delete vname;
              }
           }
        } 

        const AttributeDefinitionList * attlist = type->attributeDefTemp();
        if ( attlist) {
           printf (" attribute lists duuuuuuuuuuuuuuuuuuuuuuuuuude \n");
        }

        if ( !( type = e.next()))
          break;
        continue;
    }

    // ----------------------------------------------------
    // print  classes
    printf ("//----------------------------------------------\n");
    printf ("//\n");
    printf ("// Class declarations\n");
    printf ("//\n");
    printf ("//\n");

    Dtd::ConstElementTypeIter ee( dtd->elementTypeIter());
    type = ee.next();
    while ( type) {

        // see if is a base type
        char * varname = compoundTypeHandler.AddBaseVar (type);

        if (!varname) { 
            compoundTypeHandler.PrintClassDecl (type);
        }

        type = ee.next();
    }
    return 0;
}

#ifdef SP_NAMESPACE
}
#endif
