/***********************************************************************
 *
 * Module: fl_edit.c
 * Author: Marc van Kempen (marc@bowtie.nl)
 * Desc:   implementation of edit object 
 *
 * Copyright (c) 1996, Marc van Kempen
 *
 * All rights reserved.
 *
 * This software may be used, modified, copied, distributed, and
 * sold, in both source and binary form provided that the above
 * copyright and these terms are retained, verbatim, as the first
 * lines of this file.  Under no circumstances is the author
 * responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with
 * its use.
 *
 ***********************************************************************/

/*       $Id: fl_edit.c,v 1.43 1996/12/22 13:56:42 gena Exp $
 */

#include <stdlib.h>
#include <sys/param.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <forms.h>
#include "fl_edit.h"
#include "textbuf.h"
#include "fl_error.h"
#include "fl_scrollbar.h"

/***********************************************************************
 *
 * Some definitions 
 *
 ***********************************************************************/

/* a little macro for control characters */
#define ctrl(a) (a-'A'+1)

/***********************************************************************
 *
 * Textarea and coordinates
 * ------------------------
 *                             ob->x + ob->w - 
 *                          2 * ob->bw - sb_width
 *                                           |    +-> ob->x + ob->w
 *  ob->x                                    v    |
 *  +----------------------------------------+----+
 *  | x --+                                  |^^^^|
 *  |     |                                  |  | |
 *  .     v                                  .  +---> TEXT_SB_WIDTH
 *  .  (TEXT_X_OFFSET, TEXT_Y_OFFSET)        .    .
 *
 *  .                                        .    .
 *  |                                        |    |
 *  +----------------------------------------+----+
 *
 * Scrollbar     
 * ---------
 *             +-----> TEXT_SB_WIDTH = ob->bw + sb_box_width + ob->bw
 *             |
 *       <--------->
 *  ..--+-----------+
 *      |           |
 *      |           |
 *      .           .
 *      .           .
 *
 ***********************************************************************/

/* define the offset of the text from the coordinates of its box */
#define TEXT_X_OFFSET		2
#define TEXT_Y_OFFSET		0
#define TEXT_SB_WIDTH		25

/* this does not seem to be defined in forms.h */
void fl_draw_object_label();

/* local prototypes */
static void fl_textedit_pageup(FL_OBJECT *ob);
static void fl_textedit_pagedown(FL_OBJECT *ob);
static int  fl_textedit_set_cursor(FL_OBJECT *ob, int mx, int my, int draw);
static void fl_textedit_remove_selection(FL_OBJECT *ob);
static int  fl_textedit_selected(FL_OBJECT *ob);
static char *fl_get_textedit_seltext(FL_OBJECT *ob);
static int  fl_textedit_movecursor(FL_OBJECT *ob, int r, int c);
static void fl_textedit_inssel(FL_OBJECT *ob, char *buf);
static void fl_textedit_set_vscrollbar(FL_OBJECT *ob);
static void fl_textedit_set_hscrollbar(FL_OBJECT *ob);

static int
fl_textedit_get_wsize(FL_OBJECT *ob)
/*
 * Desc: return the textwindow size (in lines)
 */
{
    int nlines, asc, desc, lh;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the lineheight */
    lh = fl_get_char_height(spec->text_style, spec->text_size,
			    &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / lh;

    return(nlines);
} /* fl_textedit_get_wsize() */

static int
fl_textedit_get_csize(FL_OBJECT *ob)
/*
 * Desc: returns screen width in characters
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int nchars, cw;

    /* get character width */
    cw = fl_get_char_width(spec->text_style, spec->text_size);
    nchars = (ob->w - 2*abs(ob->bw) - 2*TEXT_X_OFFSET) / cw;

    return(nchars);
}

/***********************************************************************
 *
 * Text rendering routines
 * 
 ***********************************************************************/

static void
fl_textedit_get_textbb(FL_OBJECT *ob, int *x, int *y, int *w, int *h)
/*
 * Desc: return the coordinates for the textarea
 */
{
    *x = ob->x + abs(ob->bw) + TEXT_X_OFFSET;
    *y = ob->y + abs(ob->bw) + TEXT_Y_OFFSET;
    *w = ob->w - 2*abs(ob->bw) - TEXT_X_OFFSET;
    *h = ob->h - 2*abs(ob->bw) - TEXT_Y_OFFSET;

    return;
} /* fl_textedit_get_textbb() */

static void
fl_textedit_draw_textline(int align, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
			  int fgcol, int bgcol, int size, int style, char *line, 
			  int ccol, int cpos, int sels, int sele, int leftcol)
/*
 * Desc: Draw the line at the specified position with the alignment <align>
 *	 etc.
 *       Draw the selection if needed
 */
{
int swidth1, swidth2, slinelen, slen, seltype;
char *sline;

    slen = strlen(line);

    /* check how the selection should be drawed */
    if ((sels == 0) && ((sele == -1) || (sele >= slen)))
	seltype = 1;  /* the whole line is selected */
    else
    if ((sels == -1) || ((sele >= 0) && (sele <= sels)))
	seltype = -1; /* line is not selected */
    else
	seltype = 0;  /* line is partially selected */

    /* line is invisible */
    if (leftcol >= slen) {
	if (seltype == 1)
		fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, fgcol);
	else
		fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
        /* only the cursor is visible */
        if (leftcol == cpos)
	 fl_drw_text_cursor(align, x, y, w, h, fgcol, size, style, "", ccol, 0);
	return;
			}

    line += leftcol;
    slen = strlen(line);
    cpos -= leftcol;
    if (cpos < 0)
	cpos = -1;
    if (cpos > slen)
	cpos = slen;

    /* no selection */
    if (seltype == -1) {
    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
    fl_drw_text_cursor(align, x, y, w, h, fgcol, size, style, line, ccol, cpos);
    return;	    					}

    /* the whole line is selected */
    if (seltype == 1)	{
    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, fgcol);
    fl_drw_text_cursor(align, x, y, w, h, bgcol, size, style, line, ccol, cpos);
    return;	    						}

    if (sels >= 0) {
	sels -= leftcol;
	if (sels < 0)
	    sels = 0;
		   }

    if (sele >= 0) {
	sele -= leftcol;
	if (sele < 0)
	    sele = sels;
	if (sele > slen)
		sele = slen;
		   }
    else
	sele = slen;

    /* draw the whole non-selected line */
    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, bgcol);
    fl_drw_text_cursor(align, x, y, w, h, fgcol, size, style, line, ccol, cpos);

    /* draw the selected part */
    if (sele <= sels)
	return;

    slinelen = sele - sels;
    sline = (char *)malloc(slinelen + 1);
    strncpy(sline, line + sels, slinelen);
    sline[slinelen] = '\0';

    swidth1 = fl_get_string_width(style, size, line, sels);
    if (swidth1 > w)
	return;

    swidth2 = fl_get_string_width(style, size, sline, slinelen);
    if ((swidth1 + swidth2) > w)
	swidth2 = w - swidth1;

    if (sels)
    fl_rectf(x + swidth1, y, swidth2, h, fgcol);
    else
    fl_rectf(x + swidth1 - TEXT_X_OFFSET, y, swidth2, h, fgcol);
    fl_drw_text_cursor(align, x + swidth1, y, swidth2, h, bgcol, size, style, sline, ccol,((cpos >= sels) && (cpos <= (sels + slinelen))) ? (cpos - sels) : -1);

    free(sline);
} /* fl_textedit_draw_textline() */

static void
fl_textedit_draw_screen(FL_OBJECT *ob)
/*
 * Desc: refresh the screen
 */
{
    int i, w, h, eof, pos, nlines, asc, desc, lh, sels, sele, linenum;
    int tx, ty, tw, th;		/* textarea coordinates */
    char *line;
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;

    if (ob->form->frozen)
	return;

    /* get the lineheight */
    h = lh = fl_get_char_height(spec->text_style, spec->text_size,
			    &asc, &desc);
    w = fl_get_char_width(spec->text_style, spec->text_size);
    nlines = (ob->h - 2*abs(ob->bw)) / lh;

    fl_textedit_get_textbb(ob, &tx, &ty, &tw, &th);

    /* draw the box on the screen */
    fl_set_text_clipping(tx, ty, tw, th);

    fl_drw_frame(ob->boxtype, ob->x + abs(ob->bw), ob->y + abs(ob->bw), tw + TEXT_X_OFFSET, 
		 ob->h - 2*abs(ob->bw), ob->col2, abs(ob->bw));

    /* draw the text */
    i = 0;
    tb_set_current_line(tb, spec->topline);
    eof = (tb->currentline) == NULL;
    if (eof && ob->focus) {
	/* draw an empty line, so the cursor shows */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, h, 
				  ob->col1, ob->col2,
				  spec->text_size, spec->text_style, "", 
				  spec->ccol, 0, -1, -1, spec->leftcol);
	i++;
    }
    while (!eof & (i<nlines)) {
	linenum = i + spec->topline;
	tb_get_line(tb, &line);
	fl_get_string_dimension(spec->text_style, spec->text_size,
				line, strlen(line), &w, &h);
	if ((linenum == spec->r) && ob->focus) {
	    /* draw the cursor at the right position */
	    pos = spec->c;
	} else {
	    /* draw no cursor */
	    pos = -1;
	}

	sels = -1;
	sele = -1;
	/* calculate selection */
	if (fl_textedit_selected(ob) &&
		(linenum >= spec->sselr) &&
		(linenum <= spec->eselr))	{
	  sels = 0;
	  if (linenum == spec->sselr)
		sels = spec->sselc;

	  if (linenum == spec->eselr)
		sele = spec->eselc;
						}

	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty + i*h, tw, h, 
				  tb_get_linefgcolor(tb),
				  tb_get_linebgcolor(tb),
				  spec->text_size, spec->text_style, line, 
				  spec->ccol, pos, sels, sele, spec->leftcol);
	eof = !tb_set_next_line(tb);
	i++;
    }

    lh = i*h;
    fl_rectf(tx - TEXT_X_OFFSET, ty + lh, 
	 tw+TEXT_X_OFFSET, ob->h - 2*abs(ob->bw) - lh , ob->col2);
    /* fill in the little remaining bar under the text 
    fl_rectf(tx - TEXT_X_OFFSET, ty+nlines*h, 
	     tw + TEXT_X_OFFSET, ob->h - 2*abs(ob->bw) -(nlines*h), ob->col2);
	*/
    fl_unset_text_clipping();
    
    /* restore the current linenr */
    tb_set_current_line(tb, spec->r);

    tl = tb->firstline;
    while (tl)	{
	tl->flags &= ~TLINE_MODIFIED;
	tl = tl->next;
		}

    return;
} /* fl_textedit_draw_screen() */

void
fl_textedit_draw_line(FL_OBJECT *ob, int n)
/*
 * Desc: redraw the n-th line
 * Pre:  n<tb_get_lines(ob->spec->tb)
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;
    int	nlines, tx, ty, tw, th;		/* bounding box of line */
    int x, y, w, h, sels, sele;
    int  asc, desc;

    if (ob->form->frozen)
	return;

    /* check if n is valid */
    if (n<0 || n >= tb_get_nlines(tb)) {
	return;
    }
    
    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;
    if ((n < spec->topline) || (n >= spec->topline+nlines)) {
	/* line is out of visible range */
	return;
    }

    /* get the line */
    if ((tl = tb_get_lineptr_by_num(tb, n)) == NULL)
	return;

    /* get bounding box of line */
    fl_textedit_get_textbb(ob, &x, &y, &w, &h);
    tw = w;
    tx = x;
    ty = y + (n - spec->topline) * th;

    /* enable text clipping */
    fl_set_text_clipping(tx, ty, tw, th);

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, ob->col2);

    sels = -1;
    sele = -1;
    /* calculate selection */
    if (fl_textedit_selected(ob) &&
     (n >= spec->sselr) &&
     (n <= spec->eselr))	{
	sels = 0;
	if (n == spec->sselr)
		sels = spec->sselc;

	if (n == spec->eselr) {
		sele = spec->eselc;
		if (tl->strlen && (sels == sele)) {
		 sels = -1;
		 sele = -1;	  }
			      }
				}

    if (spec->r == n) {
	/* draw a cursor */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, th, 
				  tl->fgcolor,
				  tl->bgcolor,
				  spec->text_size, 
				  spec->text_style,tl->buf, spec->ccol, spec->c,
				  sels, sele, spec->leftcol);
    } else {
	/* just draw the text */
	fl_textedit_draw_textline(FL_ALIGN_LEFT, tx, ty, tw, th,
				  tl->fgcolor,
				  tl->bgcolor,
				  spec->text_size, 
				  spec->text_style,tl->buf, spec->ccol, -1,
				  sels, sele, spec->leftcol);
    }
    
    fl_unset_text_clipping();
    tl->flags &= ~TLINE_MODIFIED;

    return;
} /* fl_textedit_draw_line() */

void
fl_textedit_refresh_screen(FL_OBJECT *ob, int bclear)
/*
 * Desc: update screen. Draw only lines that require update
 *       bclear != 0 means that if screen is not filled with text
 *       then we should "erase" the empty bottom part
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    TextLine *tl;
    int line, asc, desc, nlines, th;
    int x, y, w, h;

    tl = tb->firstline;
    line = 0;
    while (tl) 	{
     if (!(tl->flags & TLINE_MODIFIED))	{
	tl = tl->next;
	line++;
	continue;			}

     fl_textedit_draw_line(ob, line);
     tl->flags &= ~TLINE_MODIFIED;

     tl = tl->next;
     line++;
		}

   if (bclear)	{
    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;
    if (tb->n >= (spec->topline + nlines))
	return;
    line = tb->n - spec->topline;
    fl_textedit_get_textbb(ob, &x, &y, &w, &h);
    y += (th * line);
    h -= (th * line);
    fl_rectf(x - TEXT_X_OFFSET, y, w + TEXT_X_OFFSET, h, ob->col2);
    if ((line == 0) && (spec->leftcol == 0)) 	{
     w = fl_get_char_width(spec->text_style, spec->text_size);
     fl_drw_text_cursor(FL_ALIGN_LEFT, x, y, w, th,ob->col1, spec->text_size, spec->text_style, "", spec->ccol, 0);
		   				}
		}

   return;
}

void
fl_textedit_draw_selection(FL_OBJECT *ob)
/*
 * Desc: redraw the selected text
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int i;

    if ((spec->sselr < 0) ||
	(spec->eselr < 0))
	return;

    if (spec->sselr == spec->eselr)
	fl_textedit_draw_line(ob, spec->sselr);
    else {
     for (i = spec->sselr; i <= spec->eselr; i++)
	fl_textedit_draw_line(ob, i);
	 }
}

void
fl_textedit_search(FL_OBJECT *ob, int again)
/*
 * Desc: search for text
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf	 *tb = &spec->tb;
    char *p;
    int line, pos, n;

    if (again)	{
     if (!spec->exp)	{
	fl_edit_error("No previous search pattern");
	fl_winset(ob->form->window);
	return;		}
		}
    else	{
	p = (char *)fl_show_input("Search for", spec->exp ? spec->exp : "");
	fl_winset(ob->form->window);
	if (spec->exp)
		free(spec->exp);
	if (!p || !strlen(p))	{
		spec->exp = NULL;
		return;		}
	spec->exp = strdup(p);
		}

    if ((line = tb_find_text(tb, spec->exp, spec->r, spec->c + 1, &pos)) == -1){
	fl_edit_error("Text not found");
	fl_winset(ob->form->window);
	return; 							       }

    fl_textedit_remove_selection(ob);
    spec->sselr = spec->eselr = line;
    spec->sselc = pos;
    spec->eselc = pos + strlen(spec->exp);
    n = spec->r;
    spec->r = line;
    spec->c = pos;

    if (fl_textedit_line_visible(ob, line)) {
	fl_textedit_draw_line(ob, n);
	fl_textedit_draw_selection(ob);	    }
    else {
	spec->topline = line;
	fl_textedit_draw_screen(ob);
	fl_textedit_set_vscrollbar(ob);
	 }
}

void
fl_textedit_hscrollbar_dim(FL_OBJECT *ob)
{
SPEC *spec = (SPEC *)ob->spec;

 spec->hsb->x = ob->x;
 spec->hsb->y = ob->y + ob->h;
 if (spec->sb->visible)
  spec->hsb->w = ob->w + TEXT_SB_WIDTH;
 else
  spec->hsb->w = ob->w;
 spec->hsb->h = TEXT_SB_WIDTH;
 spec->hsb->resize = ob->resize;
 spec->hsb->resize &= ~FL_RESIZE_Y;
 spec->hsb->segravity = ob->segravity;
 spec->hsb->nwgravity = ob->nwgravity;

 return;
}

void
fl_textedit_vscrollbar_dim(FL_OBJECT *ob)
{
SPEC *spec = (SPEC *)ob->spec;

 spec->sb->x = ob->x + ob->w;
 spec->sb->y = ob->y;
 spec->sb->w = TEXT_SB_WIDTH;
 spec->sb->h = ob->h;
 spec->sb->resize = ob->resize;
 spec->sb->resize &= ~FL_RESIZE_X;
 spec->sb->segravity = ob->segravity;
 spec->sb->nwgravity = ob->nwgravity;

 return;
}

int
fl_textedit_get_llength(FL_OBJECT *ob)
/*
 * Desc: returns length of longest displayed line
 *  screen size in characters is returned if every line fits on the screen
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;
TextLine *tl;
int csize, wsize, llength;

  csize = fl_textedit_get_csize(ob);
  wsize = fl_textedit_get_wsize(ob);

  llength = csize;
  tl = tb->firstline;
  while (tl) {
   if (tl->strlen > llength)
	llength = tl->strlen;
   tl = tl->next;
	     }

  return llength;
}

int
fl_textedit_switch_hscrollbar(FL_OBJECT *ob)
/*
 * Desc: turn the scrollbar on/off according to the length of the
 *       longest line on the screen and screen width
 *       Returns 1 when scrollbar does not require redrawing
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;
TextLine *tl;
int csize, wsize, llength;

  if (!(spec->flags & FL_TEXTEDIT_HSBAR))
	return 1;

  if (!(spec->flags & FL_TEXTEDIT_HSBAR_AUTO))
	return 0;

  csize = fl_textedit_get_csize(ob);
  wsize = fl_textedit_get_wsize(ob);

  llength = csize;
  tl = tb->firstline;
  while (tl) {
   if (tl->strlen > llength)
	llength = tl->strlen;
   tl = tl->next;
	     }

  if (spec->hsb->visible) {
	if (llength <= csize) 	{
		spec->hsb->visible = 0;
		ob->h += TEXT_SB_WIDTH;
		if (spec->sb->visible)       {
		 fl_textedit_vscrollbar_dim(ob);
		 fl_redraw_object(spec->sb); }
		fl_redraw_object(ob);
		return 1;
				}
			  }
  else	{
	if (llength > csize)	{
		fl_set_scrollbar_parameters(spec->hsb, 0, llength, spec->leftcol, csize);
		ob->h -= TEXT_SB_WIDTH;
		fl_textedit_hscrollbar_dim(ob);
		fl_show_object(spec->hsb);
		fl_redraw_object(ob);
		if (spec->sb->visible)       {
		 fl_textedit_vscrollbar_dim(ob);
		 fl_redraw_object(spec->sb); }
		return 1;
				}
	return 1;
	}

  return 0;
}

int
fl_textedit_switch_vscrollbar(FL_OBJECT *ob)
/*
 * Desc: turn the scrollbar on/off according to the number of lines 
 *       and screen size
 *       Returns 1 when scrollbar does not require redrawing
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;
int wsize;

  if (!(spec->flags & FL_TEXTEDIT_VSBAR))
	return 1;

  if (!(spec->flags & FL_TEXTEDIT_VSBAR_AUTO))
	return 0;

  wsize = fl_textedit_get_wsize(ob);
  if (spec->sb->visible) {
	if (tb->n <= wsize) 	{
		ob->w += TEXT_SB_WIDTH;
		/* fl_hide_object(spec->sb); */
		spec->sb->visible = 0;
		fl_redraw_object(ob);
		return 1;
				}
			  }
  else	{
	if (tb->n > wsize)	{
		fl_set_scrollbar_parameters(spec->sb, 0, tb->n, spec->topline, wsize);
		ob->w -= TEXT_SB_WIDTH;
		fl_textedit_vscrollbar_dim(ob);
		fl_show_object(spec->sb);
		fl_redraw_object(ob);
		return 1;
				}
	return 1;
	}

  return 0;
}

static void
fl_textedit_set_vscrollbar(FL_OBJECT *ob)
/*
 * Desc: set scrollbar value
 */
{
SPEC *spec = (SPEC *)ob->spec;

  if (!(spec->flags & FL_TEXTEDIT_VSBAR))
	return;

  if (fl_textedit_switch_vscrollbar(ob) != 0)
	return;

  fl_set_scrollbar_val(spec->sb, spec->topline);

  return;
}

static void
fl_textedit_set_hscrollbar(FL_OBJECT *ob)
/*
 * Desc: set hscrollbar value
 */
{
SPEC *spec = (SPEC *)ob->spec;

  if (!(spec->flags & FL_TEXTEDIT_HSBAR))
	return;

  if (fl_textedit_switch_hscrollbar(ob) != 0)
	return;

  fl_set_scrollbar_val(spec->hsb, spec->leftcol);

  return;
}

void
fl_textedit_set_vscrollbar_max(FL_OBJECT *ob)
/*
 * Desc: set scrollbar max value
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;

  if (!(spec->flags & FL_TEXTEDIT_VSBAR))
	return;

  if (fl_textedit_switch_vscrollbar(ob) != 0)
	return;

  fl_set_scrollbar_max(spec->sb, tb->n);

  return;
}

void
fl_textedit_set_hscrollbar_max(FL_OBJECT *ob)
/*
 * Desc: set hscrollbar max value
 */
{
SPEC *spec = (SPEC *)ob->spec;
int llength;

  if (!(spec->flags & FL_TEXTEDIT_HSBAR))
	return;

  if (fl_textedit_switch_hscrollbar(ob) != 0)
	return;

  llength = fl_textedit_get_llength(ob);
  fl_set_scrollbar_max(spec->hsb, llength);

  return;
}

void
fl_textedit_set_vscrollbar_wsize(FL_OBJECT *ob)
/*
 * Desc: set scrollbar wsize value
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;

  if (!(spec->flags & FL_TEXTEDIT_VSBAR))
	return;

  fl_set_scrollbar_parameters(spec->sb, 0, tb->n, spec->topline, fl_textedit_get_wsize(ob));

  fl_textedit_switch_vscrollbar(ob);

  return;
}

void
fl_textedit_set_hscrollbar_wsize(FL_OBJECT *ob)
/*
 * Desc: set hscrollbar wsize value
 */
{
SPEC *spec = (SPEC *)ob->spec;
TextBuf	 *tb = &spec->tb;
TextLine *tl;
int llength, csize, wsize;

  if (!(spec->flags & FL_TEXTEDIT_HSBAR))
	return;

  wsize = fl_textedit_get_wsize(ob);
  csize = fl_textedit_get_csize(ob);

  llength = csize;
  tl = tb->firstline;
  while (tl) {
   if (tl->strlen > llength)
	llength = tl->strlen;
   tl = tl->next;
	     }

  fl_set_scrollbar_parameters(spec->hsb, 0, llength, spec->leftcol, csize);

  fl_textedit_switch_hscrollbar(ob);

  return;
}

/***********************************************************************
 *
 * Event handling
 *
 ***********************************************************************/

void
fl_textedit_scroll_with_mouse(FL_OBJECT *ob, int key)
/*
 * Desc: scroll the text while the mouse button <key> is being
 *	 held down
 */
{
    int		 mx, my, oldmx, oldmy, lastmy;
    unsigned int keymask;
    int 	 button, nlines, maxlines;
    SPEC	 *spec = (SPEC *)ob->spec;
    TextBuf	 *tb = &spec->tb;
    int 	 oldtopline = spec->topline;
    int		 lineskip = 4;

    button = Button1Mask;
    switch (key) {
    case 1: button = Button1Mask; break;
    case 2: button = Button2Mask; break;
    case 3: button = Button3Mask; break;
    }

    /* Query the pointer and move the text around */
    fl_get_mouse(&oldmx, &oldmy, &keymask);
    lastmy = oldmy;
    while (keymask & button) {
	fl_get_mouse(&mx, &my, &keymask);
	if (abs(lastmy-my) >= lineskip) {
	    nlines = (oldmy-my) / lineskip;
	    spec->topline = oldtopline + nlines;
	    maxlines = tb_get_nlines(&spec->tb);
	    if (spec->topline < 0) spec->topline = 0;
	    if (spec->topline >= maxlines) spec->topline = maxlines-1;
	    spec->r = spec->topline;
	    tb_set_current_line(tb, spec->r);
	    if (spec->c > tb_get_linelen(tb)) {
		spec->c = tb_get_linelen(tb);
	    }
	    fl_textedit_set_vscrollbar(ob);
	    fl_textedit_set_cursor(ob, mx, my, 1);
	    fl_textedit_draw_screen(ob);
	    lastmy = my;
	}
	else
	    fl_textedit_set_cursor(ob, mx, my, 1);
    }

    return;
} /* fl_textedit_scroll_with_mouse() */

void
fl_textedit_cleft(FL_OBJECT *ob)
/*
 * Desc: move the whole page one character left
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (spec->leftcol == 0)
	return;

    spec->leftcol--;
    spec->c = spec->leftcol;
    fl_textedit_draw_screen(ob);
    fl_textedit_set_hscrollbar(ob);
}

void
fl_textedit_cright(FL_OBJECT *ob)
/*
 * Desc: move the whole page one character right
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (spec->leftcol >= fl_textedit_get_llength(ob))
	return;

    spec->leftcol++;
    spec->c = spec->leftcol + fl_textedit_get_csize(ob);
    fl_textedit_draw_screen(ob);
    fl_textedit_set_hscrollbar(ob);
}

static void
fl_textedit_pagedown(FL_OBJECT *ob)
/*
 * Desc: move the cursor one page down
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines, asc, desc;

    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;
    
    if (tb_get_nlines(tb) <= (nlines + spec->topline)) 	{
	tb_set_current_line(tb, tb_get_nlines(tb) - 1);
	if (spec->cpos > tb_get_linelen(tb))
		spec->cpos = tb_get_linelen(tb);
	fl_textedit_movecursor(ob, tb_get_nlines(tb) - 1, spec->cpos);
	return;						}

    spec->c = spec->cpos;
    spec->r = spec->r + nlines;
    spec->topline = spec->topline + nlines;

    if (tb_get_nlines(tb) <= (nlines + spec->topline))
	spec->topline = tb_get_nlines(tb) - nlines;

    if (spec->r >= tb_get_nlines(tb))
	spec->r = tb_get_nlines(tb) - 1;

    tb_set_current_line(tb, spec->r);
    if (spec->c > tb_get_linelen(tb))
	spec->c = tb_get_linelen(tb);

    fl_textedit_draw_screen(ob);
    fl_textedit_set_vscrollbar(ob);

    return;
} /* fl_textedit_pagedown() */

static void
fl_textedit_pageup(FL_OBJECT *ob)
/*
 * Desc; move the cursor one page up
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines, asc, desc;

    if (spec->topline == 0) 	{
	tb_set_current_line(tb, 0);
	if (spec->cpos > tb_get_linelen(tb))
		spec->cpos = tb_get_linelen(tb);
	fl_textedit_movecursor(ob, 0, spec->cpos);
	return;
				}

    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;

    spec->c = spec->cpos;
    if (spec->r > nlines-1) {
	spec->r = spec->r - nlines;
    } else {
	spec->r = 0;
    }
    if (spec->topline > spec->r) {
	spec->topline = spec->r;
    }
    tb_set_current_line(tb, spec->r);
    if (spec->c > tb_get_linelen(tb))
	spec->c = tb_get_linelen(tb);

    fl_textedit_draw_screen(ob);
    fl_textedit_set_vscrollbar(ob);

    return;
} /* fl_textedit_pageup() */

void 
fl_textedit_linedown(FL_OBJECT *ob)
/*
 * Desc: move the cursor one line down
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int th, nlines, asc, desc;

    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;

    if ((spec->r < spec->topline+nlines-1) && (spec->r < tb_get_nlines(tb)-1)) {
	tb_set_next_line(tb);
	fl_textedit_movecursor(ob, spec->r + 1,
		spec->cpos >= tb_get_linelen(tb) ? tb_get_linelen(tb) : spec->cpos);
    } else {
	/* scroll the text down by one line */
	spec->c = spec->cpos;
	if (spec->r < tb_get_nlines(tb)-1) {
	    spec->topline++;
	    spec->r++;
	    tb_set_next_line(tb);
	    if (spec->c >= tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	    fl_textedit_set_vscrollbar(ob);
	    fl_textedit_draw_screen(ob);
	}
	else	{
	if (spec->c >= tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
		}
    }
    
    return;
} /* fl_textedit_linedown() */

void
fl_textedit_lineup(FL_OBJECT *ob)
/*
 * Desc: move the cursor one line up
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    if (spec->r > spec->topline) {
	tb_set_prev_line(tb);
	fl_textedit_movecursor(ob, spec->r - 1,
		spec->cpos >= tb_get_linelen(tb) ? tb_get_linelen(tb) : spec->cpos);
    } else {
	/* scroll the text up by one line */
	if (spec->topline > 0) 	{
	    spec->topline--;
	    spec->r--;
	    tb_set_prev_line(tb);
	    spec->c = spec->cpos;
	    if (spec->c >= tb_get_linelen(tb)) {
		spec->c = tb_get_linelen(tb);
	    }
	fl_textedit_set_vscrollbar(ob);
	fl_textedit_draw_screen(ob);
				}
    }

    return;
} /* fl_textedit_lineup() */

void
fl_textedit_handle_keyboard(FL_OBJECT *ob, int key, void *xev)
/*
 * Desc: handle the keyboard input 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int wsize, nlines, th, asc, desc, n;
    XKeyEvent *xkev = (XKeyEvent *)xev;

#ifdef DEBUG
    printf("key character = %c (%d)\n", key, key);
    printf("ctrl is pressed = %d\n", ControlMask & xkev->state);
#endif

    /* check if read-only */
    if (spec->flags & FL_TEXTEDIT_READONLY)	{
	if ((key > 26) && (key < 256))
		return;

	switch (key)	{
	 case '\b':
	 case '\n':
	 case '\r':
	 case '\t':
	 case  127:
		return;
	 break;
			}
						}

    /* check selection */
    if (fl_textedit_selected(ob))	{
	switch (key)	{
	 case '\b':	/* delete selected block */
		if (tb_del_block(tb, spec->sselr, spec->sselc,
			spec->eselr, spec->eselc)) {
			while (spec->sselr &&
				!tb_set_current_line(tb, spec->sselr))
				spec->sselr--;

			spec->r = spec->sselr;
			spec->c = spec->sselc;
			if (spec->c > tb_get_linelen(tb))
				spec->c = tb_get_linelen(tb);
			spec->sselr = -1;
			spec->eselr = -1;
			n = tb_get_nlines(tb);
			if (spec->topline > spec->r)
				spec->topline = spec->r;
			fl_textedit_refresh_screen(ob, 1);
			fl_textedit_set_vscrollbar_max(ob);
			fl_textedit_set_hscrollbar_max(ob);
						   }
		else	{
			spec->c = spec->sselc;
			if (spec->c > tb_get_linelen(tb))
				spec->c = tb_get_linelen(tb);
			fl_textedit_remove_selection(ob);
			}

		return;
	 break;

	 case ctrl('L'): /* leave the selection alone */
	 break;

	 /* deselect everything unless we try to type something */
	 default:
		if ((key == '\t') || ((key > 26) && (key < 256))) {
			if (tb_del_block(tb, spec->sselr, spec->sselc,
				spec->eselr, spec->eselc)) {
				while (spec->sselr &&
					!tb_set_current_line(tb, spec->sselr))
					spec->sselr--;

				spec->r = spec->sselr;
				spec->c = spec->sselc;
				if (spec->c > tb_get_linelen(tb))
					spec->c = tb_get_linelen(tb);
				spec->sselr = -1;
				spec->eselr = -1;
				n = tb_get_nlines(tb);
				if (spec->topline > spec->r)
					spec->topline = spec->r;
				fl_textedit_set_vscrollbar(ob);
				fl_textedit_set_vscrollbar_max(ob);
				fl_textedit_set_hscrollbar_max(ob);
				fl_textedit_refresh_screen(ob, 1);
							   }
			else	{
				spec->c = spec->sselc;
				if (spec->c > tb_get_linelen(tb))
					spec->c = tb_get_linelen(tb);
				fl_textedit_remove_selection(ob);
				}
								  }
		else
			fl_textedit_remove_selection(ob);
	 break;
			}
					}

    /* first calculate the # of lines on the screen */
    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (int) (ob->h - 2*abs(ob->bw)) / th; /* # of lines on the screen */
    
    /* handle ctrl keys */
    if (key == '\n' || key == '\r') {
	tb_insert_cr(tb, spec->c);	/* insert a newline at 
					   position spec->c */
	n = tb_get_nlines(tb);
	spec->r++;
	spec->c = 0;
	spec->cpos = 0;
	tb_set_next_line(tb);
	if (spec->r >= (spec->topline + nlines)) {
	    spec->leftcol = 0;
	    spec->topline++;
	    fl_textedit_set_vscrollbar(ob);
	    fl_textedit_draw_screen(ob);
	    fl_textedit_set_vscrollbar_max(ob);
	}
	if (spec->leftcol) {
	    spec->leftcol = 0;
	    fl_textedit_draw_screen(ob); }
	else
		fl_textedit_refresh_screen(ob, 0);
	fl_textedit_set_hscrollbar_wsize(ob);
	return;
    }

    if (key == '\b') {
	/* backspace */
	if (spec->c > 0) {
	    spec->c--;
	    tb_del_char(tb, spec->c);
	    fl_textedit_refresh_screen(ob, 1);
	} else {
	    if (spec->r > 0) {
		spec->r--;
		tb_set_prev_line(tb);
		spec->c = tb_get_linelen(tb);
		tb_del_char(tb, spec->c);
		if (spec->c > tb_get_linelen(tb))
			spec->c = tb_get_linelen(tb);
		if (spec->r < spec->topline) {
		    spec->topline--;
		    n = tb_get_nlines(tb);
		fl_textedit_set_vscrollbar(ob);
		fl_textedit_set_vscrollbar_max(ob);
		}
		fl_textedit_refresh_screen(ob, 1);
		fl_textedit_set_hscrollbar_max(ob);
	    }
	}
	spec->cpos = spec->c;
	return;
    }
    if (key == 127) {
	/* del key */
	n = tb_get_nlines(tb);
	if (tb_del_char(tb, spec->c)) {
	    if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	    fl_textedit_refresh_screen(ob, 1);
	    /* check if a line got deleted */
	    if (n != tb_get_nlines(tb))
		fl_textedit_set_vscrollbar_max(ob);
	    fl_textedit_set_hscrollbar_max(ob);
	} else {
	    fl_textedit_draw_line(ob, spec->r);
	}
	return;
    }

    /* handle character inserts */
    if ((key == '\t') || ((key > 26) && (key < 256))) {
	n = tb_insert_char(tb, spec->c, key);
	/* line was wrapped */
	if (n == -1)   {
	 /* cursor is at the end of line */
	 nlines = tb_get_linelen(tb);
	 if (spec->c >= nlines) {
	  n = spec->c;
	  fl_textedit_linedown(ob);
	  spec->c = n - nlines ? n - nlines : 1; }
	 else
	  spec->c++;

	 fl_textedit_set_vscrollbar_max(ob);
	 spec->cpos = spec->c;
	 if (spec->r >= spec->topline + nlines) {
	    spec->topline++;
	    fl_textedit_set_vscrollbar(ob);
	 }
	 fl_textedit_refresh_screen(ob, 0);
						}
	else 	{
	 /* advance the cursor */
	 spec->c += n;
	 spec->cpos = spec->c;
	 if (spec->c > (int)fl_get_scrollbar_max(spec->hsb))
		fl_textedit_set_hscrollbar_max(ob);
	 fl_textedit_draw_line(ob, spec->r);
		}
	return;
    }
    
    /* handle function keys */
    switch(key) {
    case XK_Home:
	if (xkev->state & ControlMask) {
	    /* Control-Home was pressed, go to the beginning of the buffer */
	    if (fl_textedit_movecursor(ob, 0, 0) < 0) {
		spec->topline = 0;
		spec->leftcol = 0;
		spec->r = 0;
		spec->c = 0;
		fl_textedit_draw_screen(ob);		    }

	    fl_textedit_set_vscrollbar(ob);
	    fl_textedit_set_hscrollbar(ob);
	    spec->cpos = 0;
	} else {
	    /* Just Home */
	    if (spec->leftcol) {
		spec->leftcol = 0;
		spec->c = 0;
		spec->cpos = 0;
		fl_textedit_draw_screen(ob);
		fl_textedit_set_hscrollbar(ob); }
	    else {
	        fl_textedit_movecursor(ob, spec->r, 0);
	        spec->cpos = 0;
		 }
	}
	return;
	break;
    case ctrl('A'):
	if (spec->leftcol) {
		spec->leftcol = 0;
		fl_textedit_draw_screen(ob);
		fl_textedit_set_hscrollbar(ob); }
	else 	{
	        fl_textedit_movecursor(ob, spec->r, 0);
        	spec->cpos = 0;
	 	}
	return;
	break;
    case XK_End:
	if (xkev->state & ControlMask) {
	    /* Control-End was pressed, go to the end of the buffer */
	    wsize = fl_textedit_get_wsize(ob);
	    /* check if the last line is visible */
	    if (tb_get_nlines(tb) <= (wsize + spec->topline)) 	{
		fl_textedit_movecursor(ob, tb_get_nlines(tb) - 1, spec->c);
		return;						}
	    spec->r = tb_get_nlines(tb) - 1;
	    if (spec->r < 0)
		spec->r = 0;
	    spec->c = 0;
	    spec->cpos = 0;
	    spec->topline = tb_get_nlines(tb) - wsize;
	    spec->leftcol = 0;
	    fl_textedit_set_vscrollbar(ob);
	    fl_textedit_set_hscrollbar(ob);
	    fl_textedit_draw_screen(ob);
	} else {
	    if (fl_textedit_movecursor(ob, spec->r, tb_get_linelen(tb)) == -1) {
		spec->leftcol = tb_get_linelen(tb) - fl_textedit_get_csize(ob) + 1;
		fl_textedit_set_hscrollbar(ob);
		fl_textedit_draw_screen(ob);	}
	    spec->cpos = spec->c;
	} 
	return;
	break;
    case ctrl('E'):
	if (fl_textedit_movecursor(ob, spec->r, tb_get_linelen(tb)) == -1) {
		spec->leftcol = tb_get_linelen(tb) - fl_textedit_get_csize(ob) + 1;
		fl_textedit_set_hscrollbar(ob);
		fl_textedit_draw_screen(ob);	}
	spec->cpos = spec->c;
	return;
	break;
    case XK_Left:
	if (spec->c > 0) {
	    fl_textedit_movecursor(ob, spec->r, spec->c - 1);
	} else {
	    if (spec->r > 0) {
		fl_textedit_lineup(ob);
		fl_textedit_movecursor(ob, spec->r, tb_get_linelen(tb));
	    }
	}
	spec->cpos = spec->c;
	return;
	break;
    case XK_Right:
	if (spec->c < tb_get_linelen(tb)) {
	    fl_textedit_movecursor(ob, spec->r, spec->c + 1);
	} else {
	    if (spec->r < tb_get_nlines(tb)-1) {
		fl_textedit_linedown(ob);
		fl_textedit_movecursor(ob, spec->r, 0);
	    }
	}
	spec->cpos = spec->c;
	return;
	break;
    case XK_Up:
	/* move line up */
	fl_textedit_lineup(ob);
	return;
	break;
    case XK_Down:
	/* move line down */
	fl_textedit_linedown(ob);
	return;
	break;
    case XK_Next:
    case ctrl('V'):
	/* pagedown */
	fl_textedit_pagedown(ob);
	return;
	break;
    case XK_Prior:
    case ctrl('R'):
	/* pageup */
	fl_textedit_pageup(ob);
	return;
	break;
    case ctrl('K'):
	if (spec->flags & FL_TEXTEDIT_READONLY)
		return;
	/* delete to end of line */
	/* if the line is empty delete it */
	if (tb_get_linelen(tb)) {
		tb_set_linelen(tb, spec->c);
		fl_textedit_refresh_screen(ob, 0);
		return;		}
	else	{
	tb_del_line(tb);
	n = tb_get_nlines(tb);
	if (n == 0)
		spec->r = 0;
	else
	if (spec->r >= n)
		spec->r = n - 1;
	if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	fl_textedit_set_vscrollbar_max(ob);
	spec->cpos = spec->c;
	fl_textedit_refresh_screen(ob, 1);
	return;
		}
	break;
    case ctrl('L'):
	/* refresh screen */
	fl_textedit_draw_screen(ob);
	fl_textedit_set_vscrollbar(ob);
	fl_textedit_set_hscrollbar(ob);
	return;
	break;
    case ctrl('U'):
	if (spec->flags & FL_TEXTEDIT_READONLY)
		return;
	/* clear line */
	tb_set_linelen(tb, 0);
	spec->c = 0;
	spec->cpos = 0;
	fl_textedit_refresh_screen(ob, 0);
	return;
	break;
    case ctrl('T'):
	if (spec->flags & FL_TEXTEDIT_READONLY)
		return;
	fl_textedit_paste(ob);
	return;
	break;
    case ctrl('Y'):
	if (spec->flags & FL_TEXTEDIT_READONLY)
		return;
	/* delete line */
	tb_del_line(tb);
	n = tb_get_nlines(tb);
	if (n == 0)
		spec->r = 0;
	else
	if (spec->r >= n)
		spec->r = n - 1;
	if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	fl_textedit_set_vscrollbar_max(ob);
	spec->cpos = spec->c;
	fl_textedit_refresh_screen(ob, 1);
	return;
	break;
   case ctrl('Z'):
	if (spec->flags & FL_TEXTEDIT_READONLY)
		return;
	/* clear screen */
	tb_clear(tb);
	tb->bufchanged = TRUE;
	fl_textedit_set_hscrollbar_max(ob);
	fl_textedit_set_vscrollbar_max(ob);
	fl_textedit_draw_screen(ob);
	spec->c = 0;
	spec->r = 0;
	spec->cpos = 0;
	return;
	break;
  case ctrl('F'):
	fl_textedit_search(ob, 0);
	return;
	break;
  case ctrl('G'):
	fl_textedit_search(ob, 1);
	return;
	break;
    }
    
    return;
} /* fl_textedit_handle_keyboard() */

void
fl_textedit_paste(FL_OBJECT *ob)
/*
 * Paste selection into the textarea
 */
{
Window win;
Atom prop;
char *buf;
int buflen;

 buf = NULL;
 buflen = 0;

 if (XGetSelectionOwner(fl_display,XA_PRIMARY) == None) {
	buf = XFetchBytes(fl_display, &buflen);
	if (!buf || !buflen)
		return;
	fl_textedit_inssel(ob, buf);
	return;						}

 win = XGetSelectionOwner(fl_display,XA_PRIMARY);
 if (win == ob->form->window)	{
   if ((buf = fl_get_textedit_seltext(ob)) == NULL) 	{
	buf = XFetchBytes(fl_display, &buflen);
	if (!buf || !buflen)
		return;
	fl_textedit_inssel(ob, buf);
	return;						}
   else	{
	fl_textedit_inssel(ob, buf);
	return;
	}
				}
 else	{
    prop = XInternAtom(fl_display, "XF_SELECTION", FALSE );
    XConvertSelection(fl_display, XA_PRIMARY, XA_STRING, prop, ob->form->window, CurrentTime);
    XFlush(fl_display);
	}

 return;
}

void
fl_textedit_handle_selnotify(FL_OBJECT *ob, XEvent *xevent)
/*
 * Desc: process SelectionNotify event, extract text buffer and paste it
 */
{
char *buf;
int buflen, actual_format, nread = 0;
Window win;
Atom prop, actual_type;
u_long nitems, bytes_after;
unsigned char *back;

    prop = xevent->xselection.property;
    win  = xevent->xselection.requestor;

    buflen = 0;
    buf = NULL;

    do {
        int r = XGetWindowProperty(fl_display, win, prop,
                                    nread/4, 1024, TRUE,
                                    AnyPropertyType, &actual_type,
                                    &actual_format, &nitems,
                                    &bytes_after, &back );

        if ( r != Success || actual_type != XA_STRING )
            break;

        while ((nread + nitems) >= buflen) {
	    if (buflen) {
	     buflen *= 2;
	     buf = (char *)realloc(buf, buflen);
			}
	    else	{
	     buflen = 256;
	     buf = (char *)malloc(buflen);
			}
					   }
        memcpy(buf + nread, back, nitems );
        nread += nitems;
        XFree((char *)back);
    } while ( bytes_after > 0 );

    if (buflen == 0)
	return;

    buf[nread] = 0;

    fl_textedit_inssel(ob, buf);
    free(buf);

    return;
}

static void
fl_textedit_inssel(FL_OBJECT *ob, char *buf)
/*
 * Desc: insert buf into text area
 *       buf is freed after insertion is done
 */
{
TextBuf *tb;
SPEC    *spec;
int lines, n;
u_long offt;

 spec = (SPEC *)ob->spec;
 tb = &spec->tb;

 fl_get_textedit_cursorpos(ob, NULL, NULL, &offt);

 lines = tb_get_nlines(tb);
 tb_insert_block(tb, spec->r, spec->c, buf);
 if (strchr(buf, '\n') || (lines != tb_get_nlines(tb)))	{
	if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
	spec->sselr = -1;
	spec->eselr = -1;
	fl_textedit_refresh_screen(ob, 0);
	fl_textedit_set_vscrollbar_max(ob);
	fl_textedit_set_hscrollbar_max(ob);
							}
 else	{
  offt += strlen(buf);
  n = spec->r;
  fl_calc_cursorpos(ob, offt, &spec->c, &spec->r);
  fl_textedit_remove_selection(ob);
  fl_textedit_refresh_screen(ob, 0);
  fl_textedit_set_hscrollbar_max(ob);
	}
}

static char *
fl_get_textedit_seltext(FL_OBJECT *ob)
/*
 * Desc: returns pointer to selected text
 *       The text should be freed after using
 */
{
TextBuf *tb;
SPEC    *spec;
char *buf;

 spec = (SPEC *)ob->spec;
 tb = &spec->tb;

 /* check if we have any text selected */
 if ((spec->sselr < 0) || (spec->eselr < 0) ||
	((spec->sselr == spec->eselr) && (spec->sselc == spec->eselc)))
	return NULL;

 tb_get_block(tb, spec->sselr, spec->sselc, spec->eselr, spec->eselc, &buf);

 return buf;
}

static void
fl_textedit_copy(FL_OBJECT *ob)
/*
 * Copy selected text from textarea to X-servers buffer
 */
{
char *buf;

 if (fl_textedit_selected(ob)) 	{
  XSetSelectionOwner(fl_display, XA_PRIMARY, ob->form->window, CurrentTime);
  buf = fl_get_textedit_seltext(ob);
  if (buf) 		{
	XStoreBytes(fl_display, buf, strlen(buf));
	free(buf);	}
				}
}

static int
fl_textedit_set_cursor(FL_OBJECT *ob, int mx, int my, int draw)
/*
 * Desc: set the textcursor to the textcoordinates 
 *	  corresponding with (mx,my)
 * Returns: -1 if cursor is below the text area
 *          -2 at the left
 *           0 in the text area
 *           2 at the right
 *           1 above the text area
 */
{
    TextBuf *tb;
    SPEC    *spec;
    char    *line, ch[2];
    int	    i, lh, asc, desc, x, oldline, nlines, r, c;

    /* correct the mx and my offset to be relative to the 
       start of the text */
    mx = mx - ob->x - TEXT_X_OFFSET;
    my = my - ob->y - TEXT_Y_OFFSET;

    if (my < 0)
	return -1;

    if (mx < 0)
	return -2;

    if (mx > (ob->w - 2*abs(ob->bw) - 2* TEXT_X_OFFSET))
	return 2;

    spec = (SPEC *)ob->spec;
    tb = &spec->tb;

    /* get the height of the font */
    lh = fl_get_char_height(spec->text_style, spec->text_size,
			    &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / lh;

    /* save the old line with the cursor on it */
    oldline = spec->r;

    /* set the current line to the right one, but check if it's not
       out of range */
    if (spec->topline + (my/lh) >= tb_get_nlines(tb)) {
	return 0;
    }

    if ((my/lh) >= nlines)
	return 1;

    r = spec->topline + (my/lh);
    tb_set_current_line(tb, r);

    /* now calculate which character the x position points to */
    /* first get the current line */
    tb_get_line(tb, &line);
    if (!line)
	return 0;

    lh = strlen(line);
    x = 0;
    i = spec->leftcol;
    if (i > lh)
	i = lh;
    ch[1] = '\0';
    while ((x < mx) && (i < lh) && line[i]) {
	/* get the width of the i-th character */
	ch[0] = line[i];
	x = x + fl_get_string_width(spec->text_style, spec->text_size, ch, 1);
	i++;
    } /* x>=mx || line[i]==NULL */
    if (x>=mx && i>0) {
	i--;
    }

    c = i;
    fl_textedit_movecursor(ob, r, c);
    spec->cpos = spec->c;

    return 0;
} /* fl_textedit_set_cursor() */

static int
fl_textedit_selected(FL_OBJECT *ob)
{
    SPEC *spec = (SPEC *)ob->spec;

    /* check if there is any text selected */
    if ((spec->sselr < 0) ||
	(spec->eselr < 0))
	return 0;

    /* check if length of the selected part is bigger then zero */
    if ((spec->sselr == spec->eselr) &&
	(spec->sselc == spec->eselc))
	return 0;

    return 1;
}

static void
fl_textedit_remove_selection(FL_OBJECT *ob)
/*
 * Desc: remove text selection
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int i, ssel, esel;

    if ((spec->sselr < 0) ||
        (spec->eselr < 0))
        return;

    ssel = spec->sselr;
    esel = spec->eselr;

    spec->sselr = -1;
    spec->eselr = -1;

    for (i = ssel; i <= esel; i++)
	fl_textedit_draw_line(ob, i);
}

void
fl_textedit_extend_selection(FL_OBJECT *ob, int oldr, int oldc)
/*
 * Desc: extend selection up to cursor location
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int sselr, eselr, sselc, eselc, i;

 /* no selection */
 if ((spec->sselr < 0) ||
	(spec->sselc < 0))
	return;

 eselr = spec->eselr;
 sselr = spec->sselr;
 eselc = spec->eselc;
 sselc = spec->sselc;

 /* the selection now starts/ends at cursor position */
 if (spec->sselr > spec->r)	{
	spec->sselr = spec->r;
	spec->sselc = spec->c;
				}
 else
 if (spec->sselr == spec->r)	{
	if (spec->c < spec->sselc) {
		spec->sselr = spec->r;
		spec->sselc = spec->c;
				   }
	else
	if (spec->c > spec->sselc) {
		spec->eselr = spec->r;
		spec->eselc = spec->c;
				   }	
	else	{
		if (oldr != spec->r)	{
			fl_textedit_draw_line(ob, oldr);
			fl_textedit_draw_line(ob, spec->r);
					}
		return;
		}
				}
 else				{
	spec->eselr = spec->r;
	spec->eselc = spec->c;
				}

 if (eselr < spec->eselr)	{
   for (i = eselr; i <= spec->eselr; i++)
	fl_textedit_draw_line(ob, i);
				}
 else
 if (eselr > spec->eselr)	{
   for (i = spec->eselr; i <= eselr; i++)
	fl_textedit_draw_line(ob, i);
				}
 else
 if (sselr < spec->sselr)	{
   for (i = sselr; i <= spec->sselr; i++)
	fl_textedit_draw_line(ob, i);
				}
 else
 if (sselr > spec->sselr)	{
   for (i = spec->sselr; i <= sselr; i++)
	fl_textedit_draw_line(ob, i);
				}
 else {
  if ((eselc != spec->eselc) || (sselc != spec->sselc) || (oldr != spec->r))
        fl_textedit_draw_line(ob, spec->r);
      }

 if (oldr != spec->r)
	fl_textedit_draw_line(ob, oldr);
}

/***********************************************************************
 *
 * Main event handling 
 *
 ***********************************************************************/

static int
handle_textedit(FL_OBJECT *ob, int event, FL_Coord mx, FL_Coord my, 
		int key, void *xev)
/*
 * Desc: the event handler for the textedit object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int oldc, oldr, hsbh;
    char *line;
    XEvent *txev;
     
    txev = (XEvent *)xev;

#ifdef DEBUG
    if (xev && ((XEvent *)xev)->type == KeyPress) {
	XKeyEvent *xkev = (XKeyEvent *)xev;
	printf("keycode = %d, state = %d\n", xkev->keycode, xkev->state);
	printf("Control = %d, Shift = %d, Alt = %d\n",
	       xkev->state & ControlMask, xkev->state & ShiftMask, 
	       xkev->state & Mod1Mask);
    }
#endif

    switch(event) {
    case FL_DRAW:
	if (spec->flags & FL_TEXTEDIT_NOTDRAWN)	{
		spec->flags &= ~FL_TEXTEDIT_NOTDRAWN;
		fl_textedit_switch_vscrollbar(ob);
		fl_textedit_switch_hscrollbar(ob);
		spec->sb->x = ob->x;
		spec->hsb->x = ob->y;
		  				}

	if (spec->sb->visible) 	{
	if ((spec->sb->x != ob->x + ob->w) ||
	    (spec->sb->y != ob->y) ||
	    (spec->sb->w != TEXT_SB_WIDTH) ||
	    (spec->sb->h != ob->h) ||
	    ((spec->sb->resize & FL_RESIZE_X) != (ob->resize & FL_RESIZE_X)) ||
	    (spec->sb->segravity != ob->segravity) ||
	    (spec->sb->nwgravity != ob->nwgravity)) {
		fl_textedit_vscrollbar_dim(ob);
		if (spec->oldh != ob->h)
			fl_textedit_set_vscrollbar_wsize(ob);
		fl_scrollbar_draw(spec->sb);
				    		    }
		      		}
	else	{
	if (spec->oldh != ob->h)
		fl_textedit_set_vscrollbar_wsize(ob);
		}

	if (spec->hsb->visible) 	{
	hsbh = spec->sb->visible ? ob->w + TEXT_SB_WIDTH : ob->w;
	if ((spec->hsb->y != ob->y + ob->h) ||
	    (spec->hsb->x != ob->x) ||
	    (spec->hsb->h != TEXT_SB_WIDTH) ||
	    (spec->hsb->w != hsbh) ||
	    ((spec->hsb->resize & FL_RESIZE_Y) != (ob->resize & FL_RESIZE_Y)) ||
	    (spec->hsb->segravity != ob->segravity) ||
	    (spec->hsb->nwgravity != ob->nwgravity)) {
		fl_textedit_hscrollbar_dim(ob);
		if (spec->oldw != ob->w)
			fl_textedit_set_hscrollbar_wsize(ob);
		fl_scrollbar_draw(spec->hsb);
		      				     }
					}
	else	{
	if (spec->oldw != ob->w)
		fl_textedit_set_hscrollbar_wsize(ob);
		}

	fl_textedit_draw_screen(ob);
	spec->oldw = ob->w;
	spec->oldh = ob->h;
	/* fall through */
    case FL_DRAWLABEL:
	fl_draw_object_label(ob);
	break;
    case FL_PUSH:
	switch (key)  {
	 case 1:
	    /* set the cursor at the correct position */
	    fl_textedit_set_cursor(ob, mx, my, 1);

	    /* start new selection */
	    fl_textedit_remove_selection(ob);
	    spec->sselr = spec->eselr = spec->r;
	    spec->sselc = spec->eselc = spec->c;
	 break;

	 case 2:
	    /* paste selection into the text area */
	    if (spec->flags & FL_TEXTEDIT_READONLY)
		break;
	    fl_textedit_paste(ob);
	 break;

	 case 3:
	    /* call callback function */
	    return(1);
	 break;
			}
	break;
    case FL_MOUSE:
	if (key != 1)
		break;
	/* store the old cursor position */
	oldc = spec->c;
	oldr = spec->r;

	/* set the cursor at the correct position */
	switch (fl_textedit_set_cursor(ob, mx, my, 0))	{
	 case -1:
		fl_textedit_lineup(ob);
	 break;

	 case 1:
		fl_textedit_linedown(ob);
	 break;

	 case -2:
		fl_textedit_cleft(ob);
	 break;

	 case 2:
		fl_textedit_cright(ob);
	 break;
							}

	/* select everything up to the cursor position */
	fl_textedit_extend_selection(ob, oldr, oldc);
	break;
    case FL_RELEASE:
	/* copy selected text */
	if (key == 1)
		fl_textedit_copy(ob);
	break;
    case FL_KEYBOARD:
	fl_textedit_handle_keyboard(ob, key, xev);
	break;
    case FL_UNFOCUS:
	/* deselect everything */
	fl_textedit_remove_selection(ob);
	/* hide cursor */
	if (ob->form->visible)
		fl_textedit_movecursor(ob, -1, -1);
	break;
    case FL_FOCUS:
	/* show cursor */
	if (ob->form->visible)
		fl_textedit_movecursor(ob, spec->r, spec->c);
	break;
    case FL_FREEMEM:
	/* free the memory used by the spec member of ob */
	{
	    tb_clear(tb);
	    if (spec->exp)
		free(spec->exp);
	    free(spec);
	}

	break;
    case FL_DBLCLICK:
	/* a double click event was received, the timeout must be 
	   set in ob->click_timeout */
	/* mark the current word */
	tb_get_line(tb, &line);
	if (line == NULL)
		break;

	fl_textedit_remove_selection(ob);
	spec->sselr = spec->r;
	spec->eselr = spec->r;

	spec->sselc = spec->c;
	spec->eselc = spec->c;

	while ((spec->sselc > 0) && (line[spec->sselc - 1] != ' '))
		spec->sselc--;

	while ((line[spec->eselc] != '\0') && (line[spec->eselc] != ' '))
		spec->eselc++;

	fl_textedit_draw_selection(ob);
	fl_textedit_copy(ob);
	break;
    case FL_TRPLCLICK:
	/* a triple click event was received, the timeout must be 
	   set in ob->click_timeout */
	/* mark the current line */
	fl_textedit_remove_selection(ob);
	spec->sselr = spec->r;
	spec->eselr = spec->r;

	spec->sselc = 0;
	spec->eselc = -1;
	fl_textedit_draw_selection(ob);
	fl_textedit_copy(ob);
	break;
   case FL_OTHER:
	switch (txev->type) {
	  case SelectionClear:
		fl_textedit_remove_selection(ob);
	  break;

	  case SelectionNotify:
		if (!ob->focus)
			return(0);
		fl_textedit_remove_selection(ob);
		fl_textedit_handle_selnotify(ob, txev);
	  break;

	  case SelectionRequest:	{
            XSelectionRequestEvent *req = &txev->xselectionrequest;
            XEvent evt;
	    char *buf;
            evt.xselection.type = SelectionNotify;
            evt.xselection.display      = req->display;
            evt.xselection.requestor    = req->requestor;
            evt.xselection.selection    = req->selection;
            evt.xselection.target       = req->target;
            evt.xselection.property     = None;
            evt.xselection.time = req->time;

	    if (!fl_textedit_selected(ob))
		break;

            if ( req->target == XA_STRING ) {
		buf = fl_get_textedit_seltext(ob);
		if (buf) 	{
                XChangeProperty ( fl_display, req->requestor, req->property,
                                  XA_STRING, 8,
                                  PropModeReplace,
                                  (unsigned char *)buf, strlen(buf) );
                evt.xselection.property = req->property;
		free(buf);	}
            }
            XSendEvent( fl_display, req->requestor, False, 0, &evt );
            }
	  break;
				}	
	break;
    }
    
    return(0);
} /* handle_textedit() */

void
fl_textedit_sb_cb(FL_OBJECT *ob, long data)
/*
 * Desc: the callback for the scrollbar object to call to redraw
 *	  the screen
 */
{
    FL_OBJECT *obj = (FL_OBJECT *)data;
    SPEC *spec = (SPEC *)obj->spec;
    TextBuf *tb = &spec->tb;

    spec->r = fl_get_scrollbar_val(ob);
    spec->c = spec->cpos = 0;
    spec->topline = spec->r;

    if (spec->r >= tb->n) 	 {
	spec->r = tb->n - 1;
	spec->topline = spec->r; }

    fl_textedit_draw_screen((FL_OBJECT *) data);

    return;
} /* fl_textedit_sb_cb() */

void
fl_textedit_hsb_cb(FL_OBJECT *ob, long data)
/*
 * Desc: the callback for the horizontal scrollbar object to call to redraw
 *	  the screen
 */
{
    FL_OBJECT *obj = (FL_OBJECT *)data;
    SPEC *spec = (SPEC *)obj->spec;

    spec->leftcol = fl_get_scrollbar_val(ob);
    fl_textedit_draw_screen((FL_OBJECT *) data);

    return;
}


FL_OBJECT *
fl_create_textedit(int type, FL_Coord x, FL_Coord y, 
		   FL_Coord w, FL_Coord h, const char *label)
/*
 * Desc: create the textedit object
 */
{
    FL_OBJECT *ob;
    SPEC *spec;

    ob = fl_make_object(FL_TEXTEDIT, type, x, y, w, h,
			label, handle_textedit);

    ob->col1 = FL_BLACK;
    ob->col2 = FL_MCOL;
    ob->boxtype = FL_DOWN_BOX;
    ob->wantkey = FL_KEY_TAB;
    ob->click_timeout = 300;
    ob->input = TRUE;

    /* initialize the object specific user data */
    ob->spec = (SPEC *) fl_calloc(1, sizeof(SPEC));
    spec = (SPEC *)ob->spec;
    if (!ob->spec) {
	fl_edit_error("Could not malloc space for SPEC structure in object:%s", 
		 label);
	exit(1);
    }
    tb_init(&spec->tb);
    spec->tb.fgcolor_def = ob->col1;
    spec->tb.bgcolor_def = ob->col2;
    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->name[0] = '\0';
    spec->text_style = FL_FIXED_STYLE;
    spec->text_size = FL_NORMAL_SIZE;
    spec->ccol = FL_YELLOW;
    spec->sselr = -1;
    spec->sselc = 0;
    spec->eselr = -1;
    spec->eselc = 0;
    spec->flags = FL_TEXTEDIT_NOTDRAWN | FL_TEXTEDIT_VSBAR | FL_TEXTEDIT_VSBAR_AUTO | FL_TEXTEDIT_HSBAR | FL_TEXTEDIT_HSBAR_AUTO;
    spec->exp = NULL;
    spec->oldw = ob->w;
    spec->oldh = ob->h;

    /* scrollbar section */
    spec->sb = fl_add_scrollbar(FL_VERTICAL_SCROLLBAR, x+w-TEXT_SB_WIDTH,
				y, TEXT_SB_WIDTH, h, "");
    fl_set_object_callback(spec->sb, fl_textedit_sb_cb, (long) ob);
    spec->sb->visible = 0;
    fl_set_scrollbar_parameters(spec->sb, 0.0, 1.0, 0.0, fl_textedit_get_wsize(ob));
    fl_textedit_vscrollbar_dim(ob);

    spec->hsb = fl_add_scrollbar(FL_HORIZONTAL_SCROLLBAR, x, y+h-TEXT_SB_WIDTH,
				w, TEXT_SB_WIDTH, "");
    fl_set_object_callback(spec->hsb, fl_textedit_hsb_cb, (long) ob);
    spec->hsb->visible = 0;
    fl_set_scrollbar_parameters(spec->hsb, 0.0, fl_textedit_get_csize(ob), 0.0, fl_textedit_get_csize(ob));
    fl_textedit_hscrollbar_dim(ob);

    return(ob);
} /* fl_create_textedit() */

/***********************************************************************
 *
 * Exported API
 *
 ***********************************************************************/

FL_OBJECT *
fl_add_textedit(int type, FL_Coord x, FL_Coord y, 
		   FL_Coord w, FL_Coord h, const char *label)
{
    FL_OBJECT *ob = fl_create_textedit(type, x, y, w, h, label);

    fl_add_object(fl_current_form, ob);

    return(ob);
} /* fl_add_textedit() */

void
fl_set_textedit(FL_OBJECT *ob, char *buf, long len)
/*
 * Desc: insert text from <buf> with length <len> into the textedit object <ob>
 * Pre:  The textedit object has been created
 *
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n;

    tb_set_text(tb, buf, len);
    strcpy(spec->name, "noname");
    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->sselr = spec->eselr = -1;
    n = tb_get_nlines(tb);

    fl_textedit_set_vscrollbar_wsize(ob);
    fl_textedit_set_hscrollbar_wsize(ob);

    fl_redraw_object(ob);

    return;
} /* fl_set_textedit() */

void
fl_load_textedit(FL_OBJECT *ob, char *fname)
/*
 * Desc: load a file into the textedit object <ob>
 * Pre:  The textedit object has been created
 *
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    tb_load_file(tb, fname);
    strcpy(spec->name, fname);
    spec->r = 0;
    spec->c = 0;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    spec->sselr = spec->eselr = -1;
 
    fl_textedit_set_vscrollbar_wsize(ob);
    fl_textedit_set_hscrollbar_wsize(ob);

    fl_redraw_object(ob);

    return;
} /* fl_load_textedit() */

void
fl_save_textedit(FL_OBJECT *ob, char *fname)
/*
 * Desc: Save the textedit file under the name <fname>
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    tb_save_file(tb, fname);
    strcpy(spec->name, fname);

    return;  
} /* fl_save_textedit() */

char *
fl_get_textedit(FL_OBJECT *ob, u_long *len)
/*
 * Desc: return pointer to buffer with text of object <ob>
 * length of the buffer will be returned in <len>
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    return tb_get_text(tb, len);

} /* fl_get_textedit() */

void 
fl_set_textedit_fontsize(FL_OBJECT *ob, int size)
/*
 * Desc: set the font size of the object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    if (spec->text_size != size)		{
	spec->text_size = size;
	fl_redraw_object(ob);
	fl_textedit_set_vscrollbar_wsize(ob);
	fl_textedit_set_hscrollbar_wsize(ob);	}

    spec->text_size = size;
    
    return;
} /* fl_set_textedit_fontsize() */

int
fl_get_textedit_fontsize(FL_OBJECT *ob)
/*
 * Desc: get the font size of the object
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    return(spec->text_size);
} /* fl_get_textedit_fontsize() */

void
fl_set_textedit_fontstyle(FL_OBJECT *ob, int style)
/*
 * Desc: set the font style of the object 
 */
{
    SPEC *spec = (SPEC *) ob->spec;
    
    if (style != spec->text_style) 		{
	spec->text_style = style;
	fl_redraw_object(ob);
	fl_textedit_set_vscrollbar_wsize(ob);
	fl_textedit_set_hscrollbar_wsize(ob);   }

    spec->text_style = style;

    return;
} /* fl_set_textedit_fontstyle() */

int
fl_get_textedit_fontstyle(FL_OBJECT *ob)
/*
 * Desc: get the font style of the object
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return(spec->text_style);
} /* fl_get_textedit_fontstyle() */

void
fl_set_textedit_color(FL_OBJECT *ob, int bgcol, int fgcol, int ccol, int all)
/*
 * Desc: set the color of the object 
 * <bgcol>, <fgcol> and <ccol> are background , foreground and cursort colors
 *	respectively
 * if <all> is not 0 then color of every line is set to given colors
 */
{
int redraw;
SPEC *spec = (SPEC *)ob->spec;
TextBuf *tb = &spec->tb;

    redraw = all;
    if (spec->ccol != ccol)
	redraw = 1;
    spec->ccol = ccol;
    ob->col1 = fgcol;
    ob->col2 = bgcol;
    tb->fgcolor_def = ob->col1;
    tb->bgcolor_def = ob->col2;

    if (all && (tb_get_nlines(tb) > 0)) {
        tb_set_current_line(tb, 0);
        tb_set_linefgcolor(tb, fgcol);
        tb_set_linebgcolor(tb, bgcol);
        while (tb_set_next_line(tb)) {
            tb_set_linefgcolor(tb, fgcol);
            tb_set_linebgcolor(tb, bgcol);
        }
    }

    if (redraw)
	fl_redraw_object(ob);
}

void
fl_get_textedit_color(FL_OBJECT *ob, int *bgcol, int *fgcol, int *ccol)
/*
 * Desc: get the color of the object 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    if (ccol)
	*ccol = spec->ccol;
    if (bgcol)
	*bgcol = ob->col2;
    if (fgcol)
	*fgcol = ob->col1;
}

char *
fl_get_textedit_bufname(FL_OBJECT *ob)
/* 
 * Desc: return the name of the buffer 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    return(spec->name);
} /* fl_get_textedit_bufname() */

void
fl_set_textedit_bufname(FL_OBJECT *ob, char *fname)
/*
 * Desc: set the name of the buffer
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    strcpy(spec->name, fname);
    
    return;
} /* fl_set_textedit_bufname() */

int
fl_get_textedit_bufchanged(FL_OBJECT *ob)
/*
 * Desc: return the change status of the textbuffer 
 *	  TRUE if it needs saving, FALSE if not
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;

    return(tb_bufchanged(tb));
} /* fl_textedit_bufchanged() */

void
fl_set_textedit_bufchanged(FL_OBJECT *ob, int status)
/*
 * Desc: set the bufferchanged status 
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    
    tb_set_bufchanged(tb, status);
    return;
} /* fl_set_textedit_bufchanged() */

FL_OBJECT *
fl_get_textedit_vscrollbar(FL_OBJECT *ob)
/*
 * return pointer to vertical scrollbar associated with textedit 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->sb;
}

FL_OBJECT *
fl_get_textedit_hscrollbar(FL_OBJECT *ob)
/*
 * return pointer to horizontal scrollbar associated with textedit 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->hsb;
}

void
fl_set_textedit_vscrollbar(FL_OBJECT *ob, int set)
/*
 * toggle textedit's vertical scrollbar
 * set = 0 - scrollbar off
 * set = 1 - scrollbar on
 * set = 2 - scrollbar auto
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    TextBuf *tb = &spec->tb;
    int n;

  switch (set) {
   case 0:
    /* disable the scrollbar */
    if (spec->sb->visible) 	{
	fl_hide_object(spec->sb);
	ob->w += TEXT_SB_WIDTH;
	fl_redraw_object(ob); 	}
    spec->flags &= ~FL_TEXTEDIT_VSBAR;
    spec->flags &= ~FL_TEXTEDIT_VSBAR_AUTO;
   break;

   case 1:
   /* enable the scrollbar */
   if (!spec->sb->visible) {
    n = tb_get_nlines(tb);
    fl_set_scrollbar_parameters(spec->sb, 0, n, spec->topline, fl_textedit_get_wsize(ob));
    ob->w -= TEXT_SB_WIDTH;
    fl_redraw_object(ob);
    fl_textedit_vscrollbar_dim(ob);
    fl_show_object(spec->sb);
			   }
    spec->flags |= FL_TEXTEDIT_VSBAR;
    spec->flags &= ~FL_TEXTEDIT_VSBAR_AUTO;
   break;

   case 2:
   /* auto switch scrollbar */
    spec->flags |= FL_TEXTEDIT_VSBAR;
    spec->flags |= FL_TEXTEDIT_VSBAR_AUTO;
    fl_textedit_switch_vscrollbar(ob);
   break;
				}

   return;
}

void
fl_set_textedit_hscrollbar(FL_OBJECT *ob, int set)
/*
 * toggle textedit's horizontal scrollbar
 * set = 0 - scrollbar off
 * set = 1 - scrollbar on
 * set = 2 - scrollbar auto
 */
{
  SPEC *spec = (SPEC *)ob->spec;

  switch (set) {
   case 0:
    /* disable the scrollbar */
    if (spec->hsb->visible) 	{
	fl_hide_object(spec->hsb);
	ob->h += TEXT_SB_WIDTH;
	fl_redraw_object(ob); 	}
    spec->flags &= ~FL_TEXTEDIT_HSBAR;
    spec->flags &= ~FL_TEXTEDIT_HSBAR_AUTO;
   break;

   case 1:
   /* enable the scrollbar */
   if (!spec->hsb->visible) {
    fl_set_scrollbar_parameters(spec->hsb, 0, fl_textedit_get_llength(ob), spec->leftcol, fl_textedit_get_csize(ob));
    ob->h -= TEXT_SB_WIDTH;
    fl_redraw_object(ob);
    fl_textedit_hscrollbar_dim(ob);
    fl_show_object(spec->hsb);
    if (spec->sb->visible) {
	 fl_textedit_vscrollbar_dim(ob);
	 fl_redraw_object(spec->sb); }
			   }
    spec->flags |= FL_TEXTEDIT_HSBAR;
    spec->flags &= ~FL_TEXTEDIT_HSBAR_AUTO;
   break;

   case 2:
   /* auto switch scrollbar */
    spec->flags |= FL_TEXTEDIT_HSBAR;
    spec->flags |= FL_TEXTEDIT_HSBAR_AUTO;
    fl_textedit_switch_hscrollbar(ob);
   break;
				}

   return;
}

void
fl_clear_textedit(FL_OBJECT *ob)
/*
 * clear text area
 */
{
SPEC *spec = (SPEC *)ob->spec;
 
    tb_clear(&spec->tb);
    spec->r = 0;
    spec->c = 0;
    spec->sselr = -1;
    spec->eselr = -1;
    spec->cpos = 0;
    spec->topline = 0;
    spec->leftcol = 0;
    if (spec->exp)	  {
	free(spec->exp);
	spec->exp = NULL; }

    fl_set_scrollbar_max(spec->sb, 0);
    fl_set_scrollbar_max(spec->hsb, fl_textedit_get_csize(ob));
    fl_redraw_object(ob);
}

static int
fl_textedit_movecursor(FL_OBJECT *ob, int r, int c)
/*
 * move cursor to specified position
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb = &spec->tb;
int	nlines, tx, ty, tw, th;		/* bounding box of line */
int x, y, w, h, swidth, rc, invcur;
char onechar[2];
int  asc, desc;
TextLine *tl;

    fl_textedit_get_textbb(ob, &x, &y, &w, &h);
    th = fl_get_char_height(spec->text_style, spec->text_size, &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / th;


    tl = tb_get_lineptr_by_num(tb, spec->r);
    if (!tl) 		{
      if (!spec->r && !spec->c) 	      {
	swidth = fl_get_string_width(spec->text_style, spec->text_size, " ", 1);
	fl_rectf(x, y, swidth, th, ob->col2); }
	goto showcur;   }
    
    /* line is out of visible range */
    if ((spec->r < spec->topline) || (spec->r >= spec->topline+nlines))
	goto showcur;

    if (spec->c > tl->strlen)
	spec->c = tl->strlen;
    rc = spec->c - spec->leftcol;

    /* cursor is beyond the left margin */
    if (rc < 0)
	goto showcur;

    onechar[0] = (spec->c == tl->strlen) ? ' ' : tl->buf[spec->c];
    onechar[1] = '\0';

    /* get string width up to cursor position */
    swidth = fl_get_string_width(spec->text_style, spec->text_size, tl->buf + spec->leftcol, rc);

    /* cursor is beyond the right margin */
    if (swidth > (ob->w - 2*abs(ob->bw) - 2*TEXT_X_OFFSET))
	goto showcur;

    tw = fl_get_string_width(spec->text_style, spec->text_size,onechar, 1);
    tx = x + swidth;
    ty = y + (spec->r - spec->topline) * th;

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, tl->bgcolor);

    fl_drw_text_cursor(FL_ALIGN_LEFT, tx, ty, tw, th, tl->fgcolor, spec->text_size, spec->text_style, onechar, spec->ccol, -1);

showcur:

    tl = tb_get_lineptr_by_num(tb, r);
    if (!tl) {
	if (!r && !c) {
	 swidth = fl_get_string_width(spec->text_style,spec->text_size, " ", 1);
	 fl_drw_text_cursor(FL_ALIGN_LEFT, x, y, swidth, th, ob->col1, spec->text_size, spec->text_style, " ", spec->ccol, 0);
	 return 0;
				  }
	return -1;
	     }
    
    /* line is out of visible range */
    if ((r < spec->topline) || (r >= spec->topline+nlines))
	return -1;

    if (c > tl->strlen)
	c = tl->strlen;
    rc = c - spec->leftcol;

    /* cursor is beyond the left margin */
    if (rc < 0)    {
	spec->r = r;
	spec->c = c;
	return -1; }

    onechar[0] = (c >= tl->strlen) ? ' ' : tl->buf[c];
    onechar[1] = '\0';

    /* get string width up to cursor position */
    swidth = fl_get_string_width(spec->text_style, spec->text_size, tl->buf + spec->leftcol, rc);

    /* cursor is beyond the right margin */
    if (swidth > (ob->w - 2*abs(ob->bw) - 2*TEXT_X_OFFSET)) {
	spec->r = r;
	spec->c = c;
	return -1;					    }

    tw = fl_get_string_width(spec->text_style, spec->text_size,onechar, 1);
    tx = x + swidth;
    ty = y + (r - spec->topline) * th;

    /* check if cursor is on selected text */
    /* this is kludge and the code should be cleaned up */
    invcur = 0;
    if ((r > spec->sselr) && (r < spec->eselr))
	invcur = 1;
    else
    if ((r == spec->sselr) && (r == spec->eselr)) {
	if ((spec->sselc == 0) && (spec->eselc > 0) &&
	 ((spec->eselc == -1) || (spec->eselc >= tl->strlen)))
		invcur = 1;
	else
	if ((c >= spec->sselc) && (c < spec->eselc))
	 invcur = 1;
						  }
    else
    if (r == spec->sselr) {
	if ((c >= spec->sselc) &&
		((c < tl->strlen) || (spec->sselc == 0)))
		invcur = 1;
			  }
    else
    if (r == spec->eselr) {
	if ((c < spec->eselc) || (spec->eselc == -1) ||
		(spec->eselc >= tl->strlen))
		invcur = 1;
			  }

    /* erase the current line */
    fl_rectf(tx, ty, tw, th, invcur ? tl->fgcolor : tl->bgcolor);
    /* draw the cursor */
    fl_drw_text_cursor(FL_ALIGN_LEFT, tx, ty, tw, th, invcur ? tl->bgcolor : tl->fgcolor, spec->text_size, spec->text_style, onechar, spec->ccol, 0);

    spec->r = r;
    spec->c = c;
    tb_set_current_line(tb, spec->r);

    return 0;
}

void
fl_calc_cursorpos(FL_OBJECT *ob, u_long cofft, int *cx, int *cy)
/*
 * Calculate cursor position from offset
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
int i, len;
char *line;

 tb = &spec->tb;
 *cx = 0;
 *cy = 0;
 /* scan all lines */
 for (i = 0; i < tb_get_nlines(tb);i++)	{
	tb_get_line_by_num(tb, &line, i);
	if (!line)
		break;
	len = strlen(line) + 1;
	if (len < cofft) {
		cofft -= len;
		(*cy)++; }
	 else		 {
		*cx = cofft;
		break;	 }
					}
}

void
fl_set_textedit_cursorpos(FL_OBJECT *ob, int cx, int cy, long cofft)
/*
 * set position of the cursor
 */
{
 SPEC    *spec = (SPEC *)ob->spec;
 TextBuf *tb;
 int r, c;
 char *line;
 Window wn;

    tb = &spec->tb;

    /* if offset specified then calculate the cursor position from it */
    if (cofft >= 0)
	fl_calc_cursorpos(ob, cofft, &cx, &cy);

    /* set the current line to the right one, but check if it's not
       out of range */
    if (spec->topline + cy >= tb_get_nlines(tb)) {
	return;
    }
    r = spec->topline + cy;
    tb_set_current_line(tb, spec->r);

    /* now calculate which character the cx position points to */
    /* first get the current line */
    tb_get_line(tb, &line);
    if (!line) {
	/* no line available */
	return;
    }

    c = (cx >= strlen(line)) ? strlen(line) : cx;
    spec->cpos = c;

    /* redraw the lines */
    wn = fl_winget();
    fl_winset(ob->form->window);
    fl_textedit_movecursor(ob, r, c);
    fl_winset(wn);

    return;
}

void
fl_get_textedit_cursorpos(FL_OBJECT *ob, int *cx, int *cy, u_long *cofft)
/*
 * get position of the cursor
 */
{
 SPEC    *spec = (SPEC *)ob->spec;
 char *line;
 int i;

 if (cx)
	*cx = spec->c;

 if (cy)
	*cy = spec->r;

 if (cofft) 	{
  *cofft = 0;
  if (spec->r == 0) {
	*cofft = spec->c;
	return;     }

  for (i = 0; i < spec->r; i++) {
    tb_get_line_by_num(&spec->tb, &line, i);
    if (!line)
	break;

    *cofft += strlen(line) + 1; }

  *cofft += spec->c;
		}
}

int
fl_set_textedit_wrap(FL_OBJECT *ob, int wrap, int all)
/*
 * set word/line wrap
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
int owrap;

 tb = &spec->tb;
 owrap = tb_get_linewrap(tb);
 tb_set_linewrap(tb, wrap);

  if (all && (tb_get_nlines(tb) > 0)) {
        tb_set_current_line(tb, 0);
        tb_wrap_line(tb);
        while (tb_set_next_line(tb))
	 tb_wrap_line(tb);
	fl_redraw_object(ob);
    }

 return owrap;
}

void
fl_insert_textedit(FL_OBJECT *ob, char *buf)
/*
 * insert <buf> at cursor position
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
u_long offt;

 tb = &spec->tb;

 if (!buf || !*buf)
	return;

 fl_get_textedit_cursorpos(ob, NULL, NULL, &offt);

 tb_insert_block(tb, spec->r, spec->c, buf);
 if (strchr(buf, '\n'))	{
	if (spec->c > tb_get_linelen(tb))
		spec->c = tb_get_linelen(tb);
			}
 else	{
  offt += strlen(buf);
  fl_calc_cursorpos(ob, offt, &spec->c, &spec->r);
	}
 fl_redraw_object(ob);

 fl_textedit_set_vscrollbar_wsize(ob);
 fl_textedit_set_hscrollbar_wsize(ob);
}

void
fl_insert_textedit_file(FL_OBJECT *ob, char *fname)
/*
 * insert contents of file <fname> at cursor position
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;

 tb = &spec->tb;
 if (!fname)
	return;

 tb_insert_file(tb, spec->r, spec->c, fname);
 if (spec->c > tb_get_linelen(tb))
	spec->c = tb_get_linelen(tb);
 fl_redraw_object(ob);

 fl_textedit_set_vscrollbar_wsize(ob);
 fl_textedit_set_hscrollbar_wsize(ob);
}

int
fl_textedit_readonly(FL_OBJECT *ob, int readonly)
/*
 * set/remove read-only flag
 */
{
SPEC    *spec = (SPEC *)ob->spec;
int oronly;

 oronly = spec->flags & FL_TEXTEDIT_READONLY;

 if (readonly)
	spec->flags |= FL_TEXTEDIT_READONLY;
 else
	spec->flags &= ~FL_TEXTEDIT_READONLY;

 return oronly;
}

int
fl_get_textedit_topline(FL_OBJECT *ob)
/*
 * Desc: return number of the first line displayed
 */
{
SPEC    *spec = (SPEC *)ob->spec;

 return spec->topline;
}

int
fl_get_textedit_maxline(FL_OBJECT *ob)
/*
 * Desc: return number of lines in textbuffer
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;

 tb = &spec->tb;

 return tb->n;
}

int
fl_get_textedit_screenlines(FL_OBJECT *ob)
/*
 * Desc: return number of lines in textbuffer
 */
{
SPEC    *spec = (SPEC *)ob->spec;
int nlines, lh, asc, desc;

    lh = fl_get_char_height(spec->text_style, spec->text_size,
                            &asc, &desc);
    nlines = (ob->h - 2*abs(ob->bw)) / lh;

    return nlines;
}

char *
fl_get_textedit_curline(FL_OBJECT *ob)
/*
 * Desc: returns text of the line with cursor
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;

 tb = &spec->tb;

 if (tb->currentline)
	return tb->currentline->buf;
 else
	return NULL;
}

char *
fl_get_textedit_line(FL_OBJECT *ob, int n)
/*
 * Desc: returns text of the line number <n>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
char *line;

 tb = &spec->tb;

 tb_get_line_by_num(tb, &line, n);

 return line;
}

void
fl_add_textedit_line(FL_OBJECT *ob, char *line)
/*
 * Desc: append line to text buffer
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
Window wn;

 tb = &spec->tb;
 tb_append_line(tb, line);

 wn = fl_winget();
 fl_winset(ob->form->window);
 fl_textedit_refresh_screen(ob, 0);
 fl_textedit_set_vscrollbar_max(ob);
 if (strlen(line) > (int)fl_get_scrollbar_max(spec->hsb))
	fl_textedit_set_hscrollbar_max(ob);

 fl_winset(wn);

 return;
}

void
fl_set_textedit_topline(FL_OBJECT *ob, int n)
/*
 * Desc: set topline to <n>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
Window wn;

 if ((n < 0) || (n > fl_get_textedit_maxline(ob)))
	return;

 spec->topline = n;
 spec->r = n;
 spec->c = 0;
 spec->cpos = 0;

 wn = fl_winget();
 fl_winset(ob->form->window);
 fl_redraw_object(ob);
 fl_textedit_set_vscrollbar(ob);
 fl_winset(wn);

 return;
}

void
fl_addto_textedit(FL_OBJECT *ob, char *line)
/*
 * Desc: append line to text buffer and make it visible
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
int slines;
Window wn;

 tb = &spec->tb;
 tb_append_line(tb, line);

 wn = fl_winget();
 fl_winset(ob->form->window);
 slines = fl_get_textedit_screenlines(ob);
 if (tb->n > (spec->topline + slines))
	fl_set_textedit_topline(ob, tb->n - slines);
 else
	fl_redraw_object(ob);

 fl_textedit_set_vscrollbar_max(ob);
 if (strlen(line) > (int)fl_get_scrollbar_max(spec->hsb))
	fl_set_scrollbar_max(spec->hsb, strlen(line));
 fl_winset(wn);

 return;
}

int
fl_textedit_line_visible(FL_OBJECT *ob, int line)
/*
 * Desc: check if <line> is visible
 */
{
SPEC    *spec = (SPEC *)ob->spec;

 if ((line >= spec->topline) &&
	(line < (spec->topline + fl_get_textedit_screenlines(ob))))
	return 1;

 return 0;
}

void
fl_insert_textedit_line(FL_OBJECT *ob, int where, char *line)
/*
 * Desc: insert new line after line number <where>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
TextLine *tl;
Window wn;

 tb = &spec->tb;
 if ((where < 0) || (where > tb->n))
	return;

 tl = tb->currentline;

 if (!tb_set_current_line(tb, where))
	return;

 tb_insert_line(tb, line);
 tb->currentline = tl;

 wn = fl_winget();
 fl_winset(ob->form->window);
 if (fl_textedit_line_visible(ob, where) || (where < spec->topline))
	fl_redraw_object(ob);

 fl_textedit_set_vscrollbar_max(ob);
 if (strlen(line) > (int)fl_get_scrollbar_max(spec->hsb))
	fl_set_scrollbar_max(spec->hsb, strlen(line));
 fl_winset(wn);
}

void
fl_delete_textedit_line(FL_OBJECT *ob, int line)
/*
 * Desc: delete line with number <line>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
TextLine *tl;
Window wn;

 tb = &spec->tb;
 if ((line < 0) || (line > tb->n))
	return;

 tl = tb->currentline;

 if (!tb_set_current_line(tb, line))
	return;

 tb_del_line(tb);
 tb->currentline = tl;

 if (spec->r > tb->n)
	spec->r = tb->n;

 if (spec->topline > tb->n)
	spec->topline = tb->n;

 wn = fl_winget();
 fl_winset(ob->form->window);
 if (fl_textedit_line_visible(ob, line) || (line < spec->topline))
	fl_redraw_object(ob);

 fl_textedit_set_vscrollbar_max(ob);
 fl_winset(wn);
}

void
fl_replace_textedit_line(FL_OBJECT *ob, int line, char *nline)
/*
 * Desc: replace contents of line <line> with <nline>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
Window wn;
TextBuf *tb;
TextLine *tl;
int n;

 tb = &spec->tb;
 if ((line < 0) || (line > tb->n))
	return;

 tl = tb->currentline;

 if (!tb_set_current_line(tb, line))
	return;

 n = tb->n;
 tb_del_line(tb);
 tb_insert_line(tb, nline);

 tb->currentline = tl;

 wn = fl_winget();
 fl_winset(ob->form->window);
 if (n != tb->n) {
  if (fl_textedit_line_visible(ob, line) || (line < spec->topline))
	fl_redraw_object(ob);

  fl_textedit_set_vscrollbar_max(ob);
		 }
 else
  fl_textedit_draw_line(ob, line);

 if (strlen(nline) > (int)fl_get_scrollbar_max(spec->hsb))
	fl_set_scrollbar_max(spec->hsb, strlen(nline));
 fl_winset(wn);
}

void
fl_deselect_textedit(FL_OBJECT *ob)
/*
 * Desc: deselect any selected text
 */
{
Window wn;

  wn = fl_winget();
  fl_winset(ob->form->window);
  fl_textedit_remove_selection(ob);
  fl_winset(wn);
}

int
fl_isselected_textedit_line(FL_OBJECT *ob, int line)
/*
 * Desc: check if <line> is selected (at least part of it)
 */
{
SPEC    *spec = (SPEC *)ob->spec;

 if ((line >= spec->sselr) && (line <= spec->eselr)) 	{
  if ((line == spec->sselr) && (line == spec->eselr))	{
	if (spec->sselc != spec->eselc)
		return 1;
	else
		return 0;				}

	return 1;					}

 return 0;
}

void
fl_select_textedit_line(FL_OBJECT *ob, int line)
/*
 * Select <line>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
Window wn;

 wn = fl_winget();
 fl_winset(ob->form->window);
 fl_textedit_remove_selection(ob);

 spec->sselr = spec->eselr = line;
 spec->sselc = 0;
 spec->eselc = -1;

 fl_textedit_draw_selection(ob);
 fl_winset(wn);
}

void
fl_deselect_textedit_line(FL_OBJECT *ob, int line)
/*
 * Deselect <line>
 */
{
 fl_deselect_textedit(ob);

 return;
}

void
fl_set_textedit_line_color(FL_OBJECT *ob, int line, int bgcol, int fgcol)
/*
 * Desc: set back and foreground colors of <line>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
Window wn;

 tb = &spec->tb;

 if (!tb_set_current_line(tb, line))
	return;

 if (bgcol > 0)
	tb_set_linebgcolor(tb, bgcol);

 if (fgcol > 0)
	tb_set_linefgcolor(tb, fgcol);

 wn = fl_winget();
 fl_winset(ob->form->window);
 fl_textedit_draw_line(ob, line);
 fl_winset(wn);
}

void
fl_get_textedit_line_color(FL_OBJECT *ob, int line, int *bgcol, int *fgcol)
/*
 * Desc: get back and foreground colors of <line>
 */
{
SPEC    *spec = (SPEC *)ob->spec;
TextBuf *tb;
TextLine *tl;

 tb = &spec->tb;

 if ((tl = tb_get_lineptr_by_num(tb, line)) == NULL)
	return;

 if (bgcol)
	*bgcol = tl->bgcolor;

 if (fgcol)
	*fgcol = tl->fgcolor;
}

void
fl_scroll_textedit(FL_OBJECT *ob, int type)
/*
 * Desc: scroll textedit according to type
 */
{
Window wn;

 wn = fl_winget();
 fl_winset(ob->form->window);

 switch (type) 	{
  case FL_TEXTEDITSCROLL_PGUP:
	fl_textedit_pageup(ob);
  break;

  case FL_TEXTEDITSCROLL_PGDOWN:
	fl_textedit_pagedown(ob);
  break;

  case FL_TEXTEDITSCROLL_LINEUP:
  break;

  case FL_TEXTEDITSCROLL_LINEDOWN:
  break;

  case FL_TEXTEDITSCROLL_TOP:
  break;

  case FL_TEXTEDITSCROLL_BOTTOM:
  break;
		}

 fl_winset(wn);
}

