/***********************************************************************
 *
 * Module: fl_scrollbar.c
 * Author: Marc van Kempen (marc@bowtie.nl)
 * Desc:   General scrollbar-implementation. I don't like the current
 *	   slider for scrollbar behaviour.
 *
 * Copyright (c) 1996, Marc van Kempen
 *
 * All rights reserved.
 *
 * This software may be used, modified, copied, distributed, and
 * sold, in both source and binary form provided that the above
 * copyright and these terms are retained, verbatim, as the first
 * lines of this file.  Under no circumstances is the author
 * responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with
 * its use.
 *
 ***********************************************************************/

/*       $Id: fl_scrollbar.c,v 1.15 1996/12/17 15:43:27 gena Exp $
 */

/*
 * What I need: 
 * 
 * - The range for the scrollbar must be settable. The total # of lines
 *   (or whatever) and the stepsize, as well as the window size.
 *   When the user clicks above or under the scrollbar the scrollbarbutton
 *   will be moved by a corresponding amount. The scrollbarobject will 
 *   invoke its callback function when this happens and the user
 *   can request the current lineposition.
 *
 * - the location of the scrollbar must be settable
 *
 */

#include <sys/param.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <forms.h>
#include "fl_scrollbar.h"
#include "fl_error.h"

#define MIN_SCRL_HEIGHT 30.0	/* the minimum height of the vscrollbar */
#define MIN_SCRL_WIDTH	30.0	/* the minimum width of the hscrollbar */
void fl_draw_object_label();

typedef struct {
    double	min;		/* minimum value */
    double	max;		/* maximum value */
    double	step;		/* stepsize when clicking the arrows */
    double	wsize;		/* windowsize */
    double	val;		/* current scrollbar value */
    int		coord;		/* the x/y-coordinate of the scrollbar */
} SPEC;


/***********************************************************************
 *
 * The dimensions of the scrollbar.
 *
 * get the scrollbarbutton dimensions with 
 *
 *    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh)
 *
 *  +------------------------------------------------------------+
 *  |	
 *  |     ob->x
 *  |      |
 *  |      |+--- bx = ob->x + ob->bw
 *  |      v|
 *  |       V
 *  |      +------+  <-- ob->y
 *  |      |      |  <-- ob->y + ob->bw
 *  |      |      |                                                
 *  |by -> |+----+|  ^
 *  |      ||    ||  |                                              
 *  |      ||    ||  |                                              
 *  |      ||    ||  | bh			
 *  |      ||    ||  |
 *  |      |+----+|  v
 *  |      |      |
 *  |      |      |
 *  |      |      |  <-- ob->y + ob->bw + ob->h
 *  |      +------+  <-- ob->y + 2*ob->bw + ob->h
 *  |      <------> ob->w
 *  |       <----> bw
 *  |
 *  +------------------------------------------------------------+
 *
 ***********************************************************************/

/***********************************************************************
 *
 * Misc functions
 *
 ***********************************************************************/

#ifndef MAX
#define MAX(a, b)      ((a) < (b) ? (b) : (a))
#endif
#ifndef MIN
#define MIN(a, b)      ((a) < (b) ? (a) : (b))
#endif

static void
fl_scrollbar_get_bbox(FL_OBJECT *ob, int *bx, int *by, int *bw, int *bh)
/*
 * Desc: calculate bounding box of vscrollbar button 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    *bw = ob->w - 2*abs(ob->bw);
    if (fabs(spec->max - spec->min) < DBL_MIN) {
	*bh = ob->h - 2*abs(ob->bw);
    } else {
	*bh = MAX(MIN_SCRL_HEIGHT, 
		  (double) spec->wsize/(spec->max - spec->min) * 
		  (ob->h - 2*abs(ob->bw)));
	*bh = MIN(*bh, ob->h - 2*abs(ob->bw));
    }
    *bx = ob->x + abs(ob->bw);
    *by = spec->coord + ob->y + abs(ob->bw);

    return;
} /* fl_scrollbar__get_bbox() */

static void
fl_hscrollbar_get_bbox(FL_OBJECT *ob, int *bx, int *by, int *bw, int *bh)
/*
 * Desc: calculate bounding box of hscrollbar button 
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    *bh = ob->h - 2*abs(ob->bw);
    if (fabs(spec->max - spec->min) < DBL_MIN) {
	*bw = ob->w - 2*abs(ob->bw);
    } else {
	*bw = MAX(MIN_SCRL_HEIGHT, 
		  (double) spec->wsize/(spec->max - spec->min) * 
		  (ob->w - 2*abs(ob->bw)));
	*bw = MIN(*bw, ob->w - 2*abs(ob->bw));
    }
    *by = ob->y + abs(ob->bw);
    *bx = spec->coord + ob->x + abs(ob->bw);

    return;
} /* fl_scrollbar__get_bbox() */

static void
fl_scrollbar_set_y(FL_OBJECT *ob, int y)
/*
 * Desc: set the y value of the scrollbar and calculate the 
 *	corresponding scrollbarvalue
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh, h;

    /* get the bounding box of the scrollbar button */
    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* the inner height of the scrollbar */
    h = ob->h - 2*abs(ob->bw) - bh;

    /* check if the y-value is sane and set the new coordinate */
    if (y > h) y = h;
    if (y < 0) y = 0;

    /* 0 <= y <= ob->h - bh */
    spec->coord = y;

    /* calculate the new value based upon the y-coordinate */
    if ((fabs(spec->max - spec->min) < DBL_MIN) || (h == 0)) {
	spec->val = spec->min;
    } else {
	spec->val = (double) spec->coord/h * (spec->max - spec->min)
	    + spec->min;
    }

    /* now modulate by spec->step */
    spec->val = ((int) spec->val / spec->step) * spec->step;

    return;
} /* fl_scrollbar_set_y() */

static void
fl_scrollbar_set_x(FL_OBJECT *ob, int x)
/*
 * Desc: set the x value of the hscrollbar and calculate the 
 *	corresponding scrollbarvalue
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh, w;

    /* get the bounding box of the scrollbar button */
    fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* the inner width of the scrollbar */
    w = ob->w - 2*abs(ob->bw) - bw;

    /* check if the x-value is sane and set the new coordinate */
    if (x > w) x = w;
    if (x < 0) x = 0;

    /* 0 <= x <= ob->w - bw */
    spec->coord = x;

    /* calculate the new value based upon the y-coordinate */
    if ((fabs(spec->max - spec->min) < DBL_MIN) || (w == 0)) {
	spec->val = spec->min;
    } else {
	spec->val = (double) spec->coord/w * (spec->max - spec->min)
	    + spec->min;
    }

    /* now modulate by spec->step */
    spec->val = ((int) spec->val / spec->step) * spec->step;

    return;
} /* fl_scrollbar_set_x() */

static void
fl_scrollbar_set_val(FL_OBJECT *ob, double val)
/*
 * Desc: set the value of the scrollbar and calculate the 
 *	corresponding x/y coordinate
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* get the bounding box of the scrollbar button */
    if (ob->type == FL_VERTICAL_SCROLLBAR)
     fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    else
     fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* check if the value is sane */
    if (val < spec->min) val = spec->min;
    if (val > spec->max) val = spec->max;

    /* spec->min <= val <= spec->max */
    spec->val = val;

    /* calculate the new x/y-coordinate based upon the value */
    if ((fabs(spec->max - spec->min) < DBL_MIN) || (spec->max < spec->wsize)) {
	spec->coord = 0;
    } else {
      switch (ob->type) {
       case FL_VERTICAL_SCROLLBAR:
	spec->coord = (int) ((spec->val - spec->min)/(spec->max - spec->min) *
			 (ob->h - bh - 2*abs(ob->bw)));
       break;

       case FL_HORIZONTAL_SCROLLBAR:
	spec->coord = (int) ((spec->val - spec->min)/(spec->max - spec->min) *
			 (ob->w - bw - 2*abs(ob->bw)));
       break;
			}
    }

    return;
} /* fl_scrollbar_set_val() */

/***********************************************************************
 *
 * Object drawing routines
 *
 ***********************************************************************/

void
fl_scrollbar_draw(FL_OBJECT *ob)
/*
 * Desc: draw the scrollbar in its current form
 */
{
    int bx, by, bw, bh;
    int x, y, w, h;

    if (ob->form->frozen)
	return;

    fl_drw_box(FL_DOWN_BOX, ob->x, ob->y, ob->w, ob->h, ob->col2, abs(ob->bw));

    if (ob->type == FL_VERTICAL_SCROLLBAR) {
    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));

    x = bx + abs(ob->bw) + 1;
    y = by + bh/2;
    w = bw - 3*abs(ob->bw) - 1;
    
    fl_line(x, y-5, x+w,  y-5, FL_WHITE);
    fl_line(x, y-4, x+w,  y-4, FL_BLACK);

    fl_line(x, y,   x+w,  y,   FL_WHITE);
    fl_line(x, y+1, x+w,  y+1, FL_BLACK);

    fl_line(x, y+4, x+w,  y+4, FL_WHITE);
    fl_line(x, y+5, x+w,  y+5, FL_BLACK); }
    else				  {
    fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));

    y = by + abs(ob->bw) + 1;
    x = bx + bw/2;
    h = bh - 3*abs(ob->bw) - 1;
    
    fl_line(x-5, y, x-5, y+h, FL_WHITE);
    fl_line(x-4, y, x-4, y+h, FL_BLACK);

    fl_line(x,   y, x,   y+h, FL_WHITE);
    fl_line(x+1, y, x+1, y+h, FL_BLACK);

    fl_line(x+4, y, x+4, y+h, FL_WHITE);
    fl_line(x+5, y, x+5, y+h, FL_BLACK); }
    
    return;
} /* fl_scrollbar_draw() */

static void
fl_scrollbar_fastdraw(FL_OBJECT *ob, int oldcoord)
/*
 * Desc: only redraw the scrollbarbutton
 * Pre:  0 <= oldcoord <= "scrollbarlength"
 */
{
    int bx, by, bw, bh;
    int x, y, w, h;

    if (ob->form->frozen)
	return;

    if (ob->type == FL_VERTICAL_SCROLLBAR) {
    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    if (oldcoord < 0) oldcoord = 0;
    if (oldcoord > ob->h - 2*abs(ob->bw) - bh) oldcoord = ob->h - 2*abs(ob->bw) - bh;
    oldcoord += ob->y + abs(ob->bw);

    fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));
    if (oldcoord < by) {
	fl_rectf(bx, oldcoord, bw, by-oldcoord, ob->col2);
    } else {
	/* oldcoord >= by */
	fl_rectf(bx, by+bh, bw, oldcoord-by, ob->col2);
    }

    x = bx + abs(ob->bw) + 1;
    y = by + bh/2;
    w = bw - 3*abs(ob->bw);
    
    fl_line(x, y-5, x+w,  y-5, FL_WHITE);
    fl_line(x, y-4, x+w,  y-4, FL_BLACK);

    fl_line(x, y,   x+w,  y,   FL_WHITE);
    fl_line(x, y+1, x+w,  y+1, FL_BLACK);

    fl_line(x, y+4, x+w,  y+4, FL_WHITE);
    fl_line(x, y+5, x+w,  y+5, FL_BLACK);
						}
    else	{
    fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    if (oldcoord < 0) oldcoord = 0;
    if (oldcoord > ob->w - 2*abs(ob->bw) - bw) oldcoord = ob->w - 2*abs(ob->bw) - bw;
    oldcoord += ob->x + abs(ob->bw);

    fl_drw_box(FL_UP_BOX, bx, by, bw, bh, ob->col2, abs(ob->bw));
    if (oldcoord < bx) {
	fl_rectf(oldcoord, by, bx-oldcoord, bh, ob->col2);
    } else {
	/* oldcoord >= bx */
	fl_rectf(bx+bw, by, oldcoord-bx, bh, ob->col2);
    }

    y = by + abs(ob->bw) + 1;
    x = bx + bw/2;
    h = bh - 3*abs(ob->bw) - 1;
    
    fl_line(x-5, y, x-5, y+h, FL_WHITE);
    fl_line(x-4, y, x-4, y+h, FL_BLACK);

    fl_line(x,   y, x,   y+h, FL_WHITE);
    fl_line(x+1, y, x+1, y+h, FL_BLACK);

    fl_line(x+4, y, x+4, y+h, FL_WHITE);
    fl_line(x+5, y, x+5, y+h, FL_BLACK);
		}
    
    return;
} /* fl_scrollbar_fastdraw(ob) */    
    

/***********************************************************************
 *
 * Object definition/handling routines
 *
 ***********************************************************************/

static void
fl_hscrollbar_drag(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: drag the scrollbar
 */
{
    int bx, by, bw, bh, dx, oldx;
    double oldval;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the bounding box of the scrollbar button */
    fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
    if ((mx > bx+bw) || (mx < bx))
	return;

    oldx = spec->coord;
    dx = mx - (bx + bw/2);  /* initial offset of y position */
    if (dx == 0)
	return;
    oldval = spec->val;
    /* set the new scrollbar x-coordinate and calculate the
       corresponding value */
    fl_scrollbar_set_x(ob, spec->coord + dx);
    /* check to see if spec->val was changed => callback */
    if (oldval != spec->val) {
	if (ob->object_callback)
	    ob->object_callback(ob, ob->argument);
	    }
    fl_scrollbar_fastdraw(ob, oldx);

    return;
} /* fl_hcrollbar_drag() */

static int
fl_hscrollbar_click(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: react on a click on the scrollbar somewhere
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* first find out if we click at the left, on or at the right of the
     * scrollbar button. For this we need to calculate
     * the size of the button
     */
    fl_hscrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* left of the scrollbar button ? */
    if ((my > by && my < by+bh) && (mx<bx)) {
	/* page left */
	fl_scrollbar_set_val(ob, spec->val - spec->wsize);
	fl_scrollbar_draw(ob);
	return(1);
    }
    
    /* right of the scrollbar button ? */
    if ((my > by && my < by+bh) && (mx>bx+bw)) {
	/* page right */
	fl_scrollbar_set_val(ob, spec->val + spec->wsize);
	fl_scrollbar_draw(ob);
	return(1);
    }
    
    /* we should never reach this point, but the compiler doesn't know */
    return(0);
} /* fl_hscrollbar_click() */

static void
fl_scrollbar_drag(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: drag the scrollbar
 */
{
    int bx, by, bw, bh, dy, oldy;
    double oldval;
    SPEC *spec = (SPEC *)ob->spec;

    /* get the bounding box of the scrollbar button */
    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);
/*
    if ((my > by+bh) || (my < by))
	return;
*/

    oldy = spec->coord;
    dy = my - (by + bh/2);  /* initial offset of y position */
    if (dy == 0)
	return;
    oldval = spec->val;
    /* set the new scrollbar y-coordinate and calculate the
       corresponding value */
    fl_scrollbar_set_y(ob, spec->coord + dy);
    /* check to see if spec->val was changed => callback */
    if (oldval != spec->val) {
	if (ob->object_callback)
	    ob->object_callback(ob, ob->argument);
	    }
    fl_scrollbar_fastdraw(ob, oldy);

    return;
} /* fl_scrollbar_drag() */

static int
fl_scrollbar_click(FL_OBJECT *ob, int mx, int my)
/*
 * Desc: react on a click on the scrollbar somewhere
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    int bx, by, bw, bh;

    /* first find out if we click above, on or below the
     * scrollbar button. For this we need to calculate
     * the size of the button
     */
    fl_scrollbar_get_bbox(ob, &bx, &by, &bw, &bh);

    /* above scrollbar button ? */
    if ((mx > bx && mx < bx+bw) && (my<by)) {
	/* page up */
	fl_scrollbar_set_val(ob, spec->val - spec->wsize);
	fl_scrollbar_draw(ob);
	return(1);
    }
    
    /* on scrollbar button ? */
/*
    if ((mx > bx && mx < bx+bw) && (my>=by && my<=by+bh)) {
	fl_scrollbar_drag(ob, mx, my);
	return(0);
    }
*/

    /* below the scrollbar button ? */
    if ((mx > bx && mx < bx+bw) && (my>by+bh)) {
	/* page down */
	fl_scrollbar_set_val(ob, spec->val + spec->wsize);
	fl_scrollbar_draw(ob);
	return(1);
    }
    
    /* we should never reach this point, but the compiler doesn't know */
    return(0);
} /* fl_scrollbar_click() */

static int
handle_scrollbar(FL_OBJECT *ob, int event, FL_Coord mx, FL_Coord my, 
		 int key, void *xev)
/*
 * Desc: the event handler for the scrollbar object 
 */
{

    switch(event) {
    case FL_DRAW:
	fl_scrollbar_draw(ob);
	/* fall through */
    case FL_DRAWLABEL:
	fl_draw_object_label(ob);
	return(0);
	break;
    case FL_PUSH:  /* a mouse button was pushed */
	if (ob->type == FL_VERTICAL_SCROLLBAR)
	 return(fl_scrollbar_click(ob, mx, my));
	else
	 return(fl_hscrollbar_click(ob, mx, my));
	break;
    case FL_MOUSE:
	if (ob->type == FL_VERTICAL_SCROLLBAR)
	 fl_scrollbar_drag(ob, mx, my);
	else
	 fl_hscrollbar_drag(ob, mx, my);
	break;
    case FL_KEYBOARD:
	return(0);
	break;
    case FL_FREEMEM:
	/* free the memory used by the spec member of ob */
	{
	    SPEC *spec = (SPEC *)ob->spec;
	    free(spec);
	}	
	return(0);
	break;
    }
    
    return(0);
} /* handle_scrollbar() */

static FL_OBJECT *
fl_create_scrollbar(int type, FL_Coord x, FL_Coord y, FL_Coord w, FL_Coord h,
        	        	    const char *label)
/*
 * Desc: create the scrollbar object 
 */
{
    FL_OBJECT *ob;
    SPEC *spec;

    ob = fl_make_object(FL_SCROLLBAR, type, x, y, w, h, 
			label, handle_scrollbar);

    ob->col1 = FL_BLACK;
    ob->boxtype = FL_DOWN_BOX;
    ob->resize = (type == FL_VERTICAL_SCROLLBAR) ? FL_RESIZE_Y : FL_RESIZE_X;

    /* initialize the object specific user data */
    ob->spec = (SPEC *) fl_calloc(1, sizeof(SPEC));
    spec = (SPEC *)ob->spec;
    if (!ob->spec) {
	fl_edit_error("Could not malloc space for SPEC structure in object:%s", 
		 label);
	exit(1);
    }

    spec->coord = 0;
    spec->val = 0.0;
    spec->min = 0.0;
    spec->max = 1.0;
    spec->step = 1.0;
    spec->wsize = 1.0;
    
    return(ob);
} /* fl_create_scrollbar() */

/***********************************************************************
 *
 * Object API
 *
 ***********************************************************************/

FL_OBJECT *
fl_add_scrollbar(int type, FL_Coord x, FL_Coord y, 
		 FL_Coord w, FL_Coord h, const char *label)
{
    FL_OBJECT *ob = fl_create_scrollbar(type, x, y, w, h, label);

    fl_add_object(fl_current_form, ob);

    return(ob);
} /* fl_add_scrollbar() */

void
fl_set_scrollbar_min(FL_OBJECT *ob, double min)
/*
 * Desc: set the minimum size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->min = min;
    if (spec->val < min)
	fl_scrollbar_set_val(ob, min);
    fl_redraw_object(ob);

    return;
} /* fl_set_scrollbar_min() */

double
fl_get_scrollbar_max(FL_OBJECT *ob)
/*
 * Desc: return maximum value of the scrollbar
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->max;
}

double
fl_get_scrollbar_wsize(FL_OBJECT *ob)
/*
 * Desc: return wsize value of the scrollbar
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return spec->wsize;
}

void
fl_set_scrollbar_max(FL_OBJECT *ob, double max)
/*
 * Desc: set the maximum size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->max = max;
    if (spec->val > max)
	fl_scrollbar_set_val(ob, max);
    fl_redraw_object(ob);

    return;
} /* fl_set_scrollbar_max() */

void
fl_set_scrollbar_step(FL_OBJECT *ob, double step)
/*
 * Desc: set the stepsize
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    
    spec->step= step;

    return;
} /* fl_set_scrollbar_step() */

void
fl_set_scrollbar_wsize(FL_OBJECT *ob, double wsize)
/*
 * Desc: set the window size
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->wsize = wsize;
    fl_redraw_object(ob);

    return;
} /* fl_set_scrollbar_wsize() */

void
fl_set_scrollbar_wsizeval(FL_OBJECT *ob, double wsize)
/*
 * Desc: set the window size value
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    spec->wsize = wsize;

    return;
} /* fl_set_scrollbar_wsizeval() */

void
fl_set_scrollbar_val(FL_OBJECT *ob, double val)
/*
 * Desc: set the scrollbar value'
 */
{
    int oldcoord;
    SPEC *spec = (SPEC *)ob->spec;

    oldcoord = spec->coord;
    fl_scrollbar_set_val(ob, val);
    if (spec->coord != oldcoord)
	fl_scrollbar_fastdraw(ob, oldcoord);

    return;
} /* fl_set_scrollbar_val() */

double
fl_get_scrollbar_val(FL_OBJECT *ob)
/*
 * Desc: return the current scrollbar value
 */
{
    SPEC *spec = (SPEC *)ob->spec;

    return(spec->val);
} /* fl_get_scrollbar_value() */

void
fl_set_scrollbar_parameters(FL_OBJECT *ob, double min, double max, double val, double wsize)
/*
 * Desc: set scrollbar parameters
 */
{
SPEC *spec = (SPEC *)ob->spec;

    if (max >= 0)
	spec->max = max;
    if (min >= 0)
	spec->min = min;
    if (wsize >= 0)
	spec->wsize = wsize;
    if (val >= 0)
	fl_scrollbar_set_val(ob, val);

    /* fl_redraw_object(ob); */
}

/***********************************************************************
 *
 * garbage
 *
 ***********************************************************************/

#ifdef notused

static double
fl_scrollbar_calcval(FL_OBJECT *ob, int h)
/*
 * Desc: Calculate the scrollbar value, corresponding to the
 *	 y-coordinate of the scrollbarbutton and the height
 *	 of the scrollbar 
 * Pre:  if (y==0) val = spec->min
 *	 [] (y==h) val = spec->max
 *	 fi
 */
{
    SPEC *spec = (SPEC *)ob->spec;
    double val;

    val = (double) spec->coord/h * (spec->max - spec->min) + spec->min;

    return(val);
} /* fl_scrollbar_calcval() */
#endif
